/*
 * Copyright (C) 2019-2021, Xilinx Inc - All rights reserved
 *
 * Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
 *
 * Xilinx Resource Management
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include <iostream>
#include <iomanip>
#include <vector>
#include <fstream>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/filesystem.hpp>
#include <pthread.h>

#include "xrm_system.hpp"
#include "xrm_config.hpp"

#include <jsf_xrm.h>

/*
 * All system / resource related operation should be protected by the system lock.
 *
 * The initSystem() lock enter/exit is called inside.
 *
 * All other operations will be called from command line / sigbus handler / socket close,
 * so the lock enter/exit is called from corresponding entry.
 */

/*
 * init operation should be protected by lock
 */
int32_t xrm::system::initSystem() {
    int32_t ret = XRM_SUCCESS;
    enterLock();
    /* init the configuration */
    initConfig();
    /* init the version depended lib functions */
    initLibVersionDepFunctions();

    m_devicesInited = false;
#if 0
    /*
     * Considering the FPGA devices may not ready before XRM daemon started,
     * so to do the device init later.
     */
    /* init the devices */
    if (!m_devicesInited)
        initDevices();
#endif

    /* init the devices */
    if (!m_devicesInited)
        ret = initDevices();

    /* init the xrm plugins */
    initXrmPlugins();
    /* init the user defined cu groups */
    initUdfCuGroups();

    m_clientId = 0;
    m_numConcurrentClient = 0;
    m_allocServiceId = 0;
    m_reservePoolId = 0;
    exitLock();
    return (ret);
}

/**
 * logMsg() - log system message for XRM
 *
 * @logLevel: requested log level
 * @format, ...: message to be logged
 * @return: void
 **/
void xrm::system::logMsg(xrmLogLevelType logLevel, const char* format, ...) {
    if (logLevel <= m_logLevel) {
        va_list tmpArgs;
        va_start(tmpArgs, format);
        int32_t len = std::vsnprintf(NULL, 0, format, tmpArgs);
        va_end(tmpArgs);
        if (len < 0) return;

        ++len; // To include null terminator
        std::vector<char> msg(len);
        va_list args;
        va_start(args, format);
        len = std::vsnprintf(msg.data(), len, format, args);
        va_end(args);
        syslog(logLevel, "%s", msg.data());
    }
}

int32_t xrm::system::getLogLevel() {
    return (m_logLevel);
}

/*
 * call while holding lock
 */
void xrm::system::initConfig() {
    m_logLevel = xrm::config::getVerbosity();
    if (m_logLevel > XRM_MAX_LOG_LEVEL) m_logLevel = XRM_DEFAULT_LOG_LEVEL;
    logMsg(XRM_LOG_NOTICE, "%s : logLevel = %d", __func__, m_logLevel);
    m_limitConcurrentClient = xrm::config::getLimitConcurrentClient();
    if (m_limitConcurrentClient > XRM_MAX_LIMIT_CONCURRENT_CLIENT)
        m_limitConcurrentClient = XRM_MAX_LIMIT_CONCURRENT_CLIENT;
    logMsg(XRM_LOG_NOTICE, "%s : limitConcurrentClient = %d", __func__, m_limitConcurrentClient);
#ifdef __SUPPORT_XRT__
    m_xrtVersionFileFullPathName = xrm::config::getXrtVersionFileFullPathName();
    logMsg(XRM_LOG_NOTICE, "%s : xrtVersionFileFullPathName = %s", __func__, m_xrtVersionFileFullPathName.c_str());
    m_libXrtCoreFileFullPathName = xrm::config::getLibXrtCoreFileFullPathName();
    logMsg(XRM_LOG_NOTICE, "%s : libXrtCoreFileFullPathName = %s", __func__, m_libXrtCoreFileFullPathName.c_str());
#endif
}

/*
 * open the dll library and get the functions
 */
void xrm::system::initLibVersionDepFunctions() {
#ifdef __SUPPORT_XRT__
    char* errorMsg;
    pt::ptree xrtVersionPtree;

    m_libVersionDepFuncs.libVersionDepFuncXclLockDevice = NULL;
    m_libVersionDepFuncs.libVersionDepFuncXclUnlockDevice = NULL;
    m_libVersionDepFuncs.libVersionDepFuncXclCuName2Index = NULL;
    m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index25 = NULL;
    m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index26 = NULL;

    /* load xrt library file */
    void* libXrtcoreHandle = dlopen(m_libXrtCoreFileFullPathName.c_str(), (RTLD_GLOBAL | RTLD_NOW));
    if (!libXrtcoreHandle) {
        errorMsg = dlerror();
        logMsg(XRM_LOG_ERROR, "%s: Failed to open %s\n Error msg: %s\n", __func__, m_libXrtCoreFileFullPathName.c_str(),
               errorMsg);
        return;
    }

    libVersionDepFuncXclLockDeviceType libVersionDepFuncXclLockDevice =
        (libVersionDepFuncXclLockDeviceType)dlsym(libXrtcoreHandle, "xclLockDevice");
    if ((errorMsg = dlerror()) != NULL) {
        logMsg(XRM_LOG_NOTICE, "%s: Function xclLockDevice not implemented\n Debug msg: %s\n", __func__, errorMsg);
        m_libVersionDepFuncs.libVersionDepFuncXclLockDevice = NULL;
    } else {
        m_libVersionDepFuncs.libVersionDepFuncXclLockDevice = libVersionDepFuncXclLockDevice;
    }

    libVersionDepFuncXclUnlockDeviceType libVersionDepFuncXclUnlockDevice =
        (libVersionDepFuncXclUnlockDeviceType)dlsym(libXrtcoreHandle, "xclUnlockDevice");
    if ((errorMsg = dlerror()) != NULL) {
        logMsg(XRM_LOG_NOTICE, "%s: Function xclUnlockDevice not implemented\n Debug msg: %s\n", __func__, errorMsg);
        m_libVersionDepFuncs.libVersionDepFuncXclUnlockDevice = NULL;
    } else {
        m_libVersionDepFuncs.libVersionDepFuncXclUnlockDevice = libVersionDepFuncXclUnlockDevice;
    }

    libVersionDepFuncXclCuName2IndexType libFuncXclCuName2Index =
        (libVersionDepFuncXclCuName2IndexType)dlsym(libXrtcoreHandle, "xclCuName2Index");
    if ((errorMsg = dlerror()) != NULL) {
        logMsg(XRM_LOG_NOTICE, "%s: Function xclCuName2Index not implemented\n Debug msg: %s\n", __func__, errorMsg);
        m_libVersionDepFuncs.libVersionDepFuncXclCuName2Index = NULL;
    } else {
        m_libVersionDepFuncs.libVersionDepFuncXclCuName2Index = libFuncXclCuName2Index;
    }

    /* load xrt version json file */
    try {
        pt::read_json(m_xrtVersionFileFullPathName.c_str(), xrtVersionPtree);
    } catch (const boost::property_tree::json_parser_error& e) {
        std::string eMsg = "Error: " + e.message();
        logMsg(XRM_LOG_ERROR, "%s: Failed to read %s, \n", __func__, m_xrtVersionFileFullPathName.c_str(), eMsg);
        return;
    }
    std::string xrtBuildVersion = xrtVersionPtree.get<std::string>("BUILD_VERSION", "");
    if (xrtBuildVersion.c_str()[0] == '\0') {
        logMsg(XRM_LOG_ERROR, "%s: BUILD_VERSION is not set in %s\n", __func__, m_xrtVersionFileFullPathName.c_str());
        return;
    }
    logMsg(XRM_LOG_NOTICE, "%s: XRT BUILD_VERSION: %s\n", __func__, xrtBuildVersion.c_str());
    if (Version(xrtBuildVersion.c_str()) < Version("2.6.0")) {
        // for XRT 2.5.xxx
        libVersionDepFuncXclIPName2IndexType25 libFuncXclIPName2Index25 =
            (libVersionDepFuncXclIPName2IndexType25)dlsym(libXrtcoreHandle, "xclIPName2Index");
        if ((errorMsg = dlerror()) != NULL) {
            logMsg(XRM_LOG_NOTICE, "%s: Function xclIPName2Index not implemented\n Debug msg: %s\n", __func__,
                   errorMsg);
            m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index25 = NULL;
        } else {
            m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index25 = libFuncXclIPName2Index25;
        }
    } else {
        // for XRT 2.6.xxx
        libVersionDepFuncXclIPName2IndexType26 libFuncXclIPName2Index26 =
            (libVersionDepFuncXclIPName2IndexType26)dlsym(libXrtcoreHandle, "xclIPName2Index");
        if ((errorMsg = dlerror()) != NULL) {
            logMsg(XRM_LOG_NOTICE, "%s: Function xclIPName2Index not implemented\n Debug msg: %s\n", __func__,
                   errorMsg);
            m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index26 = NULL;
        } else {
            m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index26 = libFuncXclIPName2Index26;
        }
    }
#endif
}

/*
 * Get device BDF number from string (format xxxx:xx:xx.x)
 * Domain - bit[31, 16], Bus - bit[15, 8], Device - bit[7, 3], Function - bit[2, 0]
 */
static uint32_t deviceBDFStrToNum(std::string pciDevBDF) {
    std::string pciDevDomain = pciDevBDF.substr(0,4);
    std::string pciDevBus = pciDevBDF.substr(5,2);
    std::string pciDevDevice = pciDevBDF.substr(8,2);
    std::string pciDevFunction = pciDevBDF.substr(11,1);
    uint32_t numDomain = std::stoul(pciDevDomain, NULL, 16);
    uint32_t numBus = std::stoul(pciDevBus, NULL, 16);
    uint32_t numDevice = std::stoul(pciDevDevice, NULL, 16);
    uint32_t numFunction = std::stoul(pciDevFunction, NULL, 16);
    // BDF : Domain - bit[31, 16], Bus - bit[15, 8], Device - bit[7, 3], Function - bit[2, 0]
    uint32_t devNumBDF = ((numDomain & 0xFFFF) << 16) | ((numBus & 0xFF) << 8) |
        ((numDevice & 0x1F) << 3) | ((numFunction & 0x7) << 0);

    return (devNumBDF);
}

/*
 * call while holding lock
 */
int32_t xrm::system::initDevices() {
    int32_t ret = XRM_SUCCESS;

    /* init the devices */
    if (m_devicesInited) return (ret);

#ifdef __SUPPORT_XRT__
    int32_t numDevice = xclProbe();
    m_devicesInited = true;

    if (numDevice <= 0 || numDevice > XRM_MAX_XILINX_DEVICES) {
        logMsg(XRM_LOG_ERROR, "%s : Probed %d devices, out of range.", __func__, numDevice);
        return (ret);
    }
    m_numDevice = numDevice;
    for (int32_t devId = 0; devId < numDevice; devId++) {
        flushDevData(devId);
        m_devList[devId].devId = devId;
        m_devList[devId].isDisabled = false;
        m_devList[devId].devLoadInfo.init(devId);
        openDevice(devId);
    }
    return (ret);
#endif

//#ifdef __SUPPORT_DDBI__
    jsf_xrm::init(m_DemoDDBI);
    std::string countString = jsf_xrm::execute(R"({ "path": "deviceCount"})");
    printf("%s\n", countString.c_str());
    std::stringstream countStream;
    countStream << countString;
    pt::ptree countTree;

    try {
        pt::read_json(countStream, countTree);
    } catch (const boost::property_tree::json_parser_error& e) {
        logMsg(XRM_LOG_ERROR, "%s : Devices Count Json format is not right\n", __func__);
        ret = XRM_ERROR;
        return (ret);
    }
    std::string countStatusStr = countTree.get<std::string>("status", "");
    if (countStatusStr.compare("ok") != 0) {
        logMsg(XRM_LOG_ERROR, "%s : Devices Count status error: %s\n", __func__, countStatusStr.c_str());
        ret = XRM_ERROR;
        return (ret);
    }
    int32_t numDevice = 0;
    std::string countDataStr = countTree.get<std::string>("data", "");
    if (countDataStr.empty()) {
        logMsg(XRM_LOG_ERROR, "%s : Devices Count status error: %s\n", __func__, countStatusStr.c_str());
        ret = XRM_ERROR;
        return (ret);
    } else {
        numDevice = std::stoi(countDataStr, NULL, 10);
    }

    m_devicesInited = true;

    if (numDevice <= 0 || numDevice > XRM_MAX_XILINX_DEVICES) {
        logMsg(XRM_LOG_ERROR, "%s : Probed %d devices, out of range.", __func__, numDevice);
        printf("%s : Probed %d devices, out of range.\n", __func__, numDevice);
        ret = XRM_ERROR;
        return (ret);
    }
    m_numDevice = numDevice;
    for (int32_t devId = 0; devId < numDevice; devId++) {
        flushDevData(devId);
        m_devList[devId].devId = devId;
        m_devList[devId].isDisabled = false;
        m_devList[devId].devLoadInfo.init(devId);
    }

    /*
     * Get all device cu block and type information
     */
    int32_t devIndex, cuIndex, logicGroupIndex, cuInLogicGroupIndex;
    deviceData* dev;
    cuData* cu;

    std::string devicesString = jsf_xrm::execute(R"({ "path": "devices"})");
    printf("%s\n", devicesString.c_str());
    std::stringstream devicesStream;
    devicesStream << devicesString;
    pt::ptree devicesTree;

    try {
        pt::read_json(devicesStream, devicesTree);
    } catch (const boost::property_tree::json_parser_error& e) {
        logMsg(XRM_LOG_ERROR, "%s : Devices Infor Json format is not right\n", __func__);
        printf("%s : Devices Infor Json format is not right\n", __func__);
        ret = XRM_ERROR;
        return (ret);
    }
    std::string statusStr = devicesTree.get<std::string>("status", "");
    if (statusStr.compare("ok") != 0) {
        logMsg(XRM_LOG_ERROR, "%s : Devices info status error: %s\n", __func__, statusStr.c_str());
        printf("%s : Devices info status error: %s\n", __func__, statusStr.c_str());
        ret = XRM_ERROR;
        return (ret);
    }

    /* check whether "data" is available in the json output */
    boost::optional<pt::ptree&> keyData = devicesTree.get_child_optional("data");
    if (!keyData.is_initialized()) {
        logMsg(XRM_LOG_ERROR, "%s : no devices data\n", __func__);
        printf("%s : no devices data\n", __func__);
        ret = XRM_ERROR;
        return (ret);
    }
    auto devicesData = devicesTree.get_child("data");
    for (const auto& device : devicesData) {
        /* get the device_index information */
        auto devIndexStr = device.second.get("device_index", "");
        devIndex = std::stoi(devIndexStr, NULL, 10);

        dev = &m_devList[devIndex];

        dev->devNumBDFStr = device.first;
        dev->devNumBDF = deviceBDFStrToNum(dev->devNumBDFStr);

        cuIndex = 0;
        dev->xclbinInfo.numCu = 0;
        dev->xclbinInfo.numHardwareCu = 0;
        dev->xclbinInfo.numSoftwareCu = 0;

        /* check whether "blocks" is available in the json output */
        boost::optional<const pt::ptree &> keyBlocks = device.second.get_child_optional("blocks");
        if (!keyBlocks.is_initialized()) {
            logMsg(XRM_LOG_ERROR, "%s : no blocks in device, invalid device\n", __func__);
            printf("%s : no blocks in device, invalid device\n", __func__);
            // go to next device
            dev->xclbinInfo.numCu = 0;
            dev->xclbinInfo.numHardwareCu = 0;
            dev->xclbinInfo.numSoftwareCu = 0;
            continue;
        }
        bool badIpBlock = false;
        auto cuBlocks = device.second.get_child("blocks");
        for (const auto& cuBlock : cuBlocks) {
            cu = &dev->xclbinInfo.cuList[cuIndex];
            // get and check cuId
            std::string cuIdStr= cuBlock.second.get("cuId         ","");
            if (!cuIdStr.empty()) {
                cu->cuId = std::stoi(cuIdStr, NULL, 10);
                if (cu->cuId != cuIndex) {
                    logMsg(XRM_LOG_ERROR, "%s : block [%d] id is not right: %d\n", __func__, cuIndex, cu->cuId);
                    printf("%s : block [%d] id is not right: %d\n", __func__, cuIndex, cu->cuId);
                    badIpBlock = true;
                }
            } else {
                logMsg(XRM_LOG_ERROR, "%s : block [%d] id is not set\n", __func__, cuIndex);
                printf("%s : block [%d] id is not set\n", __func__, cuIndex);
                badIpBlock = true;
            }
            if (badIpBlock)
                break;
            std::string strIpKernel ("IP Kernel");
            std::string strSoftKernel ("Soft Kernel");
            std::string strCuType = cuBlock.second.get("cuType       ", "");
            if (strCuType.compare(strIpKernel) == 0) {
                cu->cuType = CU_IPKERNEL;
                dev->xclbinInfo.numHardwareCu++;
            }
            if (strCuType.compare(strSoftKernel) == 0) {
                cu->cuType = CU_SOFTKERNEL;
                dev->xclbinInfo.numSoftwareCu++;
            }
            cu->kernelName = cuBlock.second.get("kernelName   ", "");
            cu->kernelAlias = cuBlock.second.get("kernelAlias  ", "");
            cu->instanceName = cuBlock.second.get("instanceName ", "");
            cu->cuName = cu->kernelName + ":" + cu->instanceName;
            cu->kernelPluginFileName = cuBlock.second.get("kernelPluginFileName ", "");
            std::string  maxCapacityStr = cuBlock.second.get("maxCapacity  ", "");
            if (!maxCapacityStr.empty())
                cu->maxCapacity = std::stoull(maxCapacityStr, NULL, 10);

            cu->ipLayoutIndex = cu->cuId;
            cuInitChannels(cu);
            cu->numClient = 0;
            cu->totalUsedLoadUnified = 0;
            cu->totalReservedLoadUnified = 0;
            cu->totalReservedUsedLoadUnified = 0;
            memset(cu->reserves, 0, sizeof(reserveData) * XRM_MAX_KERNEL_RESERVES);
            cu->numReserve = 0;
            cu->deviceId = devIndex;
            cu->deviceNumBDF = dev->devNumBDF;
            dev->xclbinInfo.numCu++;
            cuIndex++;
        }
        if (badIpBlock) {
            // go to next device
            dev->xclbinInfo.numCu = 0;
            dev->xclbinInfo.numHardwareCu = 0;
            dev->xclbinInfo.numSoftwareCu = 0;
            continue;
        }

        dev->xclbinInfo.numLogicGroup = 0;
        logicGroupIndex = 0;
        bool badLogicGroup = false;
        /* check whether "LogicGroups" is available in the json output */
        boost::optional<const pt::ptree &> keyLogicGroups = device.second.get_child_optional("LogicGroups");
        if (!keyLogicGroups.is_initialized()) {
            // no logic group information
            logMsg(XRM_LOG_DEBUG, "%s : no LogicGroups in device\n", __func__);
            printf("LOG:: no LogicGroups in device [%d]\n", devIndex);
        } else {
            // get the logic group information
            printf("LOG:: get LogicGroups of device [%d]\n", devIndex);
            auto logicGroups = device.second.get_child("LogicGroups");
            for (const auto& logicGroup : logicGroups) {
                // get the information of each logic group
                boost::optional<const pt::ptree &> logicGroupCuIdTree =
                    logicGroup.second.get_child_optional("cuId         ");
                if (logicGroupCuIdTree.is_initialized()) {
                    // get the cuId of current logic group
                    printf("LOG:: logic group [%d] cu id: ", logicGroupIndex);
                    cuInLogicGroupIndex = 0;
                    dev->xclbinInfo.logicGroups[logicGroupIndex].numCu = 0;
                    for (auto it : (pt::ptree&)(*logicGroupCuIdTree)) {
                        auto cuId = it.second.get_value<int32_t>();
                        printf(" %d ", cuId);
                        // invalid cuId
                        if ((cuId < 0) || (cuId >= dev->xclbinInfo.numCu)) {
                            logMsg(XRM_LOG_ERROR, "%s : invalid cu id[%d] = %d in logic group [%d]\n",
                                __func__, cuInLogicGroupIndex, cuId, logicGroupIndex);
                            printf("\nLOG:: %s : invalid cu id[%d] = %d in logic group [%d]\n",
                                __func__, cuInLogicGroupIndex, cuId, logicGroupIndex);
                            badLogicGroup = true;
                            break;
                        }
                        dev->xclbinInfo.logicGroups[logicGroupIndex].cuId[cuInLogicGroupIndex] = cuId;
                        dev->xclbinInfo.logicGroups[logicGroupIndex].numCu++;
                        cuInLogicGroupIndex++;
                    }
                    if (badLogicGroup)
                        break;
                } else {
                    logMsg(XRM_LOG_ERROR, "%s : no cu id in logic group [%d]\n", __func__, logicGroupIndex);
                    printf("LOG:: no cu id in logic group");
                    badLogicGroup = true;
                    break;
                }

                std::string logicGroupType = logicGroup.second.get("groupTypeName","");
                strncpy(dev->xclbinInfo.logicGroups[logicGroupIndex].groupType,
                    logicGroupType.c_str(), XRM_MAX_LOGIC_GROUP_TYPE_LEN -1);
                printf("\nLOG:: groupTypeName: %s\n", logicGroupType.c_str());

                dev->xclbinInfo.logicGroups[logicGroupIndex].logicGroupId = logicGroupIndex;
                logicGroupIndex++;
                dev->xclbinInfo.numLogicGroup++;
            }
        }
        if (badLogicGroup) {
            // go to next device
            dev->xclbinInfo.numCu = 0;
            dev->xclbinInfo.numHardwareCu = 0;
            dev->xclbinInfo.numSoftwareCu = 0;
            continue;
        }

        dev->dsaName = device.second.get("type", "");
        dev->isLoaded = true;
    }
    return (ret);
//#endif
}

/*
 * call while holding lock
 */
void xrm::system::initXrmPlugins() {
    /* init the plugins */
    m_numXrmPlugin = 0;
    for (uint32_t xrmPluginId = 0; xrmPluginId < XRM_MAX_PLUGIN_NUM; xrmPluginId++) {
        flushXrmPluginInfo(xrmPluginId);
    }
}

/*
 * call while holding lock
 */
void xrm::system::initUdfCuGroups() {
    /* init the user defined cu groups */
    m_numUdfCuGroup = 0;
    uint32_t udfCuGroupIdx;
    for (uint32_t udfCuGroupIdx = 0; udfCuGroupIdx < XRM_MAX_UDF_CU_GROUP_NUM; udfCuGroupIdx++) {
        flushUdfCuGroupInfo(udfCuGroupIdx);
    }
    m_numUdfCuGroupV2 = 0;
    for (udfCuGroupIdx = 0; udfCuGroupIdx < XRM_MAX_UDF_CU_GROUP_NUM_V2; udfCuGroupIdx++) {
        flushUdfCuGroupInfoV2(udfCuGroupIdx);
    }
}

/*
 * It's used to assign new client id for new connection during context creating.
 *
 * Need to be proected by lock.
 */
uint64_t xrm::system::getNewClientId() {
    uint64_t clientId;
    if (m_clientId == 0xFFFFFFFFFFFFFFFF)
        m_clientId = 1;
    else
        m_clientId++;
    clientId = m_clientId;
    return (clientId);
}

void* xrm::system::listDevice(int32_t devId) {
#ifdef __SUPPORT_XRT__
    /* if the device is not opened yet, then need to open it at first */
    if (m_devList[devId].deviceHandle == 0) openDevice(devId);
#endif
#if 1
    /* For test and debug */
    deviceDumpResource(devId);
#endif
    return ((void*)&m_devList[devId]);
}

int32_t xrm::system::getDeviceNumber() {
    /*
     * XRM daemon may start before FPGA device ready during boot,
     * so try to init devices at the first time using devices.
     */
    if (!m_devicesInited) initDevices();

    return (m_numDevice);
}

/*
 * call while holding lock
 */
int32_t xrm::system::openDevice(int32_t devId) {
    int32_t ret = 0;
    if (devId < 0 || devId >= m_numDevice) {
        logMsg(XRM_LOG_ERROR, "%s : device id %d out of range, device num = %d\n", __func__, devId, m_numDevice);
        ret = XRM_ERROR_INVALID;
        return (ret);
    }
#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];

    dev->deviceHandle = xclOpen(devId, NULL, XCL_INFO);
    ret = xclGetDeviceInfo2(dev->deviceHandle, &dev->deviceInfo);
    if (ret != 0) {
        logMsg(XRM_LOG_ERROR, "%s : Could not get device %d info", __func__, devId);
        return (ret);
    }

    dev->dsaName = dev->deviceInfo.mName;
#if 0
    logMsg(LOG_NOTICE, "%s : device %d dsaName = %s", __func__, devId, dev->dsaName.c_str());
#endif
#endif
    return (ret);
}

#ifdef __SUPPORT_XRT__
bool static isFileExist(const std::string& fileName) {
    struct stat buf;
    return (stat(fileName.c_str(), &buf) == 0);
}
#endif

/*
 * call while holding lock
 */
int32_t xrm::system::isDeviceOffline(int32_t devId) {
    int32_t status = XRM_ERROR_INVALID;
    if (devId < 0 || devId >= m_numDevice) {
        logMsg(XRM_LOG_ERROR, "%s : device id %d out of range, device num = %d\n", __func__, devId, m_numDevice);
        return (status);
    }
#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];
    if (m_devList[devId].deviceHandle == 0) {
        logMsg(XRM_LOG_ERROR, "%s : device %d is not opened\n", __func__, devId);
        return (status);
    }

    /* mPciSlot is "(mDev->domain<<16) + (mDev->bus<<8) + (mDev->dev<<3) + mDev->func" */
    int32_t pciSlot = dev->deviceInfo.mPciSlot;
    int32_t domain, bus, device, func;
    std::string domainStr, busStr, deviceStr, funcStr, offlineFileName;
    std::stringstream domainSs, busSs, deviceSs, funcSs;
    func = pciSlot & 0x7;
    funcSs << std::setw(1) << std::setfill('0') << std::hex << func;
    funcStr = funcSs.str();
    device = (pciSlot >> 3) & 0x1F;
    deviceSs << std::setw(2) << std::setfill('0') << std::hex << device;
    deviceStr = deviceSs.str();
    bus = (pciSlot >> 8) & 0xFF;
    busSs << std::setw(2) << std::setfill('0') << std::hex << bus;
    busStr = busSs.str();
    domain = (pciSlot >> 16) & 0xFFFF;
    domainSs << std::setw(4) << std::setfill('0') << std::hex << domain;
    domainStr = domainSs.str();

    offlineFileName =
        "/sys/bus/pci/devices/" + domainStr + ":" + busStr + ":" + deviceStr + "." + funcStr + "/dev_offline";
    if (!isFileExist(offlineFileName)) {
        logMsg(XRM_LOG_ERROR, "%s : %s is not existing\n", __func__, offlineFileName.c_str());
        return (status);
    }
    std::string offlineResultStr;
    std::ifstream offlineIfs(offlineFileName);
    if (offlineIfs.is_open()) {
        std::getline(offlineIfs, offlineResultStr);
        offlineIfs.close();
        status = std::stoi(offlineResultStr, nullptr, 0);
        logMsg(XRM_LOG_DEBUG, "%s : device [%d] offline status is %d\n", __func__, devId, status);
    }
#endif
    return (status);
}

/*
 * Should be called while holding lock
 *
 * To close the device, clean device handler
 */
int32_t xrm::system::closeDevice(int32_t devId) {
    int32_t ret = XRM_SUCCESS;
    if (devId < 0 || devId >= m_numDevice) {
        logMsg(XRM_LOG_ERROR, "%s : device id %d out of range, device num = %d\n", __func__, devId, m_numDevice);
        ret = XRM_ERROR_INVALID;
        return (ret);
    }
#ifdef __SUPPORT_XRT__
    if (m_devList[devId].deviceHandle) {
        logMsg(XRM_LOG_DEBUG, "%s : close device %d\n", __func__, devId);
        xclClose(m_devList[devId].deviceHandle);
        memset(&(m_devList[devId].deviceHandle), 0, sizeof(xclDeviceHandle));
        memset(&(m_devList[devId].deviceInfo), 0, sizeof(xclDeviceInfo2));
    } else {
        logMsg(XRM_LOG_ERROR, "%s : device handle is not valid\n", __func__);
        ret = XRM_ERROR;
    }
#endif
    return (ret);
}

/*
 * Should be called while holding lock
 *
 * To reset device:
 *    1) recycle all the resource used by all clients on this device.
 *    2) unload xclbin from this device.
 *    3) close this device.
 */
int32_t xrm::system::resetDevice(int32_t devId) {
    int32_t ret = XRM_SUCCESS;
    if (devId < 0 || devId >= m_numDevice) {
        logMsg(XRM_LOG_ERROR, "%s : device id %d out of range, device num = %d\n", __func__, devId, m_numDevice);
        ret = XRM_ERROR_INVALID;
        return (ret);
    }
#ifdef __SUPPORT_XRT__
    uint32_t i;
    if (m_devList[devId].deviceHandle) {
        logMsg(XRM_LOG_DEBUG, "%s : reset device %d\n", __func__, devId);

        /* clean all the clients which are using the device */
        std::vector<uint64_t> clientIdVect;
        std::vector<pid_t> clientProcessIdVect;
        if (m_devList[devId].isExcl) {
            if (m_devList[devId].clientProcs[0].clientId != 0) {
                clientIdVect.push_back(m_devList[devId].clientProcs[0].clientId);
                clientProcessIdVect.push_back(m_devList[devId].clientProcs[0].clientProcessId);
            }
        } else {
            for (i = 0; i < XRM_MAX_DEV_CLIENTS; i++) {
                if (m_devList[devId].clientProcs[i].clientId != 0) {
                    clientIdVect.push_back(m_devList[devId].clientProcs[i].clientId);
                    clientProcessIdVect.push_back(m_devList[devId].clientProcs[i].clientProcessId);
                }
            }
        }
        /* recycle all the resource used by the clients */
        for (i = 0; i < clientIdVect.size(); i++) {
            recycleResource(clientIdVect[i]);
        }
        /* send SIGBUS to kill the client processes */
        for (i = 0; i < clientProcessIdVect.size(); i++) {
            kill(clientProcessIdVect[i], SIGBUS);
        }

        /* unload xclbin from the device */
        std::string unloadErrmsg;
        unloadOneDevice(devId, unloadErrmsg);

        /* close the device, clean device handler */
        closeDevice(devId);
    } else {
        logMsg(XRM_LOG_ERROR, "%s : device handle is not valid\n", __func__);
        ret = XRM_ERROR;
    }
#endif
    return (ret);
}

/*
 * the load process need to be protected by lock
 * device id: 0 - (numDevice -1): the target device; -2: all devices on the system.
 */
int32_t xrm::system::loadDevices(pt::ptree& loadTree, std::string& errmsg) {
    int32_t ret = XRM_ERROR;
    std::string strDevId, xclbin;
    int32_t devId;
    std::string loadErrmsg;

    /*
     * XRM daemon may start before FPGA device ready during boot,
     * so try to init devices at the first time using devices.
     */
    if (!m_devicesInited) initDevices();

    for (auto it : loadTree) {
        auto kv = it.second;
        strDevId = kv.get<std::string>("device", "");
        if (strDevId.c_str()[0] == '\0') {
            errmsg += "request parameters device is not provided";
            ret = XRM_ERROR_INVALID;
            break;
        }
        devId = std::stoi(strDevId, nullptr, 0);
        if (devId < -2 || devId == -1 || devId >= m_numDevice) {
            errmsg += "device " + std::to_string(devId) + " not found";
            ret = XRM_ERROR_INVALID;
            break;
        }

        xclbin = kv.get<std::string>("xclbin", "");
        if (xclbin.c_str()[0] == '\0') {
            errmsg += "request parameters xclbin is not provided";
            ret = XRM_ERROR_INVALID;
            break;
        }

        if (devId == -2) {
            for (int32_t i = 0; i < m_numDevice; i++) {
                ret = deviceLoadXclbin(i, xclbin, loadErrmsg);
                if (ret != XRM_SUCCESS) {
                    errmsg += loadErrmsg;
                    break;
                }
            }
        } else {
            ret = deviceLoadXclbin(devId, xclbin, loadErrmsg);
            if (ret != XRM_SUCCESS) {
                errmsg += loadErrmsg;
                break;
            }
        }
    }
    return (ret);
}

/*
 * Load xclbin to any possible device
 *
 * success: return the device id loaded with new xclbin
 * fail: return the error code
 *
 * the operation should already within lock protection
 */
int32_t xrm::system::loadAnyDevice(int32_t* devId, std::string& xclbin, std::string& errmsg) {
    int32_t ret = XRM_ERROR;
    std::string loadErrmsg;

    /*
     * XRM daemon may start before FPGA device ready during boot,
     * so try to init devices at the first time using devices.
     */
    if (!m_devicesInited) initDevices();

    if (xclbin.c_str()[0] == '\0') {
        errmsg += "request parameters xclbin is not provided";
        ret = XRM_ERROR_INVALID;
        return (ret);
    }

    /* load xclbin to one device not loaded */
    for (int32_t i = 0; i < m_numDevice; i++) {
        if (!isDeviceLoaded(i)) {
            ret = deviceLoadXclbin(i, xclbin, loadErrmsg);
            if (ret == XRM_SUCCESS) {
                *devId = i;
                return (ret);
            }
        }
    }
    if (ret != XRM_SUCCESS) {
        /* load xclbin to one device not in use */
        for (int32_t i = 0; i < m_numDevice; i++) {
            if (!isDeviceBusy(i)) {
                if (unloadOneDevice(i, loadErrmsg) != XRM_SUCCESS) continue;
                ret = deviceLoadXclbin(i, xclbin, loadErrmsg);
                if (ret == XRM_SUCCESS) {
                    *devId = i;
                    return (ret);
                }
            }
        }
    }
    if (ret != XRM_SUCCESS) {
        errmsg += loadErrmsg;
    }
    return (ret);
}

/*
 * Load xclbin to one device
 *
 * success: return the device id loaded with new xclbin
 * fail: return the error code
 *
 * the load process need to be protected by lock
 */
int32_t xrm::system::loadOneDevice(pt::ptree& loadTree, std::string& errmsg) {
    int32_t ret = XRM_ERROR;
    std::string loadErrmsg;

    /*
     * XRM daemon may start before FPGA device ready during boot,
     * so try to init devices at the first time using devices.
     */
    if (!m_devicesInited) initDevices();

    auto xclbin = loadTree.get<std::string>("xclbinFileName");
    auto devId = loadTree.get<int32_t>("deviceId");
    logMsg(XRM_LOG_NOTICE, "%s(): devId is %d, xclbin file is %s\n", __func__, devId, xclbin.c_str());

    if (xclbin.c_str()[0] == '\0') {
        errmsg += "request parameters xclbin is not provided";
        ret = XRM_ERROR_INVALID;
        goto load_exit;
    }
    if (devId < -1 || devId >= m_numDevice) {
        errmsg += "device id is out of range";
        ret = XRM_ERROR_INVALID;
        goto load_exit;
    }

    if (devId != -1) {
        /* load xclbin to one specified device */
        ret = deviceLoadXclbin(devId, xclbin, loadErrmsg);
    } else {
        /* load xclbin to any possible device */
        ret = loadAnyDevice(&devId, xclbin, loadErrmsg);
    }
    if (ret != XRM_SUCCESS) {
        errmsg += loadErrmsg;
        goto load_exit;
    }

load_exit:
    if (ret == XRM_SUCCESS) /* return the device id if load is successful */
        return (devId);
    else
        /* return the error code if load fail */
        return (ret);
}

int32_t xrm::system::xclbinFileReadUuid(std::string& name, std::string& uuidStr, std::string& errmsg) {
    std::ifstream file(name.c_str(), std::ios::binary | std::ios::ate);
    if (!file.good()) {
        errmsg = "Failed to open xclbin file " + name;
        return (XRM_ERROR);
    }
    std::streamsize size = file.tellg();
    file.seekg(0, std::ios::beg);

    char* memBuffer = (char*)malloc(size);
    if (memBuffer == NULL) {
        errmsg = "Failed to alloce buffer for xclbin file " + name;
        return (XRM_ERROR);
    }
    if (!file.read(memBuffer, size)) {
        errmsg = "Failed to read xclbin file " + name;
        free(memBuffer);
        return (XRM_ERROR);
    }

    uuid_t uuid;
#ifdef __SUPPORT_XRT__
    axlf* xclbin = reinterpret_cast<axlf*>(memBuffer);
    uuid_copy(uuid, xclbin->m_header.uuid);
    free(memBuffer);
#endif
    /* convert uuid to string */
    binToHexstr((unsigned char*)&uuid, sizeof(uuid_t), uuidStr);
    return (XRM_SUCCESS);
}

int32_t xrm::system::deviceLoadXclbin(int32_t devId, std::string& xclbin, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        errmsg += ", failed to load " + xclbin + " to device " + std::to_string(devId);
        return (XRM_ERROR_INVALID);
    }
    if (m_devList[devId].isDisabled) {
        errmsg = "device " + std::to_string(devId) + " is disabled";
        errmsg += ", failed to load " + xclbin + " to device " + std::to_string(devId);
        return (XRM_ERROR);
    }
    if (m_devList[devId].isLoaded) {
        std::string uuidStr;
        if (xclbinFileReadUuid(xclbin, uuidStr, errmsg) != XRM_SUCCESS) {
            errmsg += ", failed to load " + xclbin + " to device " + std::to_string(devId);
            return (XRM_ERROR);
        }
        if (uuidStr.compare(m_devList[devId].xclbinInfo.uuidStr) == 0) {
            /* target xclbin file has same uuid as the one already loaded to device */
            return (XRM_SUCCESS);
        } else {
            errmsg = "device " + std::to_string(devId) + " is already loaded with xclbin (" +
                     m_devList[devId].xclbinName + " uuid: " + m_devList[devId].xclbinInfo.uuidStr + ")";
            errmsg += ", failed to load " + xclbin + " (uuid: " + uuidStr + ") to device " + std::to_string(devId);
            return (XRM_ERROR);
        }
    }

#ifdef __SUPPORT_XRT__
    logMsg(XRM_LOG_NOTICE, "%s load xclbin file %s to device %d", __func__, xclbin.c_str(), devId);

    /* if the device is not opened yet, then need to open it at first */
    if (m_devList[devId].deviceHandle == 0) {
        openDevice(devId);
    }

    /* Read xclbin file */
    if (xclbinReadFile(devId, xclbin, errmsg) != XRM_SUCCESS) {
        errmsg += ", failed to load " + xclbin + " to device " + std::to_string(devId);
        deviceClearInfo(devId);
        return (XRM_ERROR);
    }

    /* Load xclbin to device */
    if (xclbinLoadToDevice(devId, errmsg) != XRM_SUCCESS) {
        errmsg += ", failed to load " + xclbin + " to device " + std::to_string(devId);
        deviceClearInfo(devId);
        return (XRM_ERROR);
    }

    /*
     * Get the information of xclbin on this device.
     * This function should be called after load because it will retrieve cuId which
     * depends on xclbin loading sucessful.
     */
    if (xclbinGetInfo(devId, errmsg) != XRM_SUCCESS) {
        errmsg += ", failed to load " + xclbin + " to device " + std::to_string(devId);
        deviceClearInfo(devId);
        return (XRM_ERROR);
    }

    /* Lock the loaded xclbin to the device so current downloaded xclbin will NOT be
     * replaced by other xclbin download process.
     */
    if (deviceLockXclbin(devId) != XRM_SUCCESS) {
        errmsg += "Unable to lock device, failed to load " + xclbin + " to device " + std::to_string(devId);
        deviceClearInfo(devId);
        return (XRM_ERROR);
    }

    /* Free the memory buffer */
    deviceFreeBuffer(devId);

    /* Update state */
    m_devList[devId].xclbinName = xclbin;
    m_devList[devId].isLoaded = true;

    return (XRM_SUCCESS);
#else
    // NOT __SUPPORT_XRT__
    // xclbin load will not work
    return (XRM_ERROR);
#endif
}

void xrm::system::flushDevData(int32_t devId) {
    int i;

    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];
        xclbinInformation* xclbinInfo = &dev->xclbinInfo;
        cuData* cu;

        // fresh the dev struct
        dev->dsaName = "";
        dev->xclbinName = "";
        dev->devId = 0;

        dev->isLoaded = 0;
        dev->isExcl = 0;
        dev->memBuffer = NULL;
#ifdef __SUPPORT_XRT__
        memset(&dev->deviceHandle, 0, sizeof(xclDeviceHandle));
        memset(&dev->deviceInfo, 0, sizeof(xclDeviceInfo2));
#endif
        memset(dev->clientProcs, 0, sizeof(clientData) * XRM_MAX_DEV_CLIENTS);

        // fresh the xclbin info
        memset(&xclbinInfo->uuid, 0, sizeof(uuid_t));
        xclbinInfo->uuidStr = "";
        xclbinInfo->numCu = 0;
        xclbinInfo->numHardwareCu = 0;
        xclbinInfo->numSoftwareCu = 0;
        xclbinInfo->numMemBank = 0;
        xclbinInfo->numConnect = 0;
        memset(xclbinInfo->memTopologyList, 0, sizeof(memTopologyData) * XRM_MAX_DDR_MAP);
        memset(xclbinInfo->connectList, 0, sizeof(connectData) * XRM_MAX_CONNECTION_ENTRIES);
        xclbinInfo->numLogicGroup = 0;
        memset(xclbinInfo->logicGroups, 0, sizeof(logicGroupData) * XRM_MAX_LOGIC_GROUPS);

        // fresh the cuList
        for (i = 0; i < XRM_MAX_XILINX_KERNELS; i++) {
            cu = &xclbinInfo->cuList[i];
            cu->cuType = CU_NULL;
            cu->ipLayoutIndex = 0;
            cu->kernelName = "";
            cu->kernelAlias = "";
            cu->cuName = "";
            cu->instanceName = "";
            cu->kernelPluginFileName = "";
            cu->maxCapacity = 0;
            cu->baseAddr = 0;
            cu->membankId = 0;
            cu->membankType = 0;
            cu->membankSize = 0;
            cu->membankBaseAddr = 0;
            memset(cu->channels, 0, sizeof(channelData) * XRM_MAX_KERNEL_CHANNELS);
            cu->numChanInuse = 0;
            memset(cu->clients, 0, sizeof(uint64_t) * XRM_MAX_KERNEL_CHANNELS);
            cu->numClient = 0;
            cu->totalUsedLoadUnified = 0; // will set dev load when numCu is set
            cu->totalReservedLoadUnified = 0;
            cu->totalReservedUsedLoadUnified = 0;
            memset(cu->reserves, 0, sizeof(reserveData) * XRM_MAX_KERNEL_RESERVES);
            cu->numReserve = 0;
            cu->deviceId = devId;
        }
    }
}

bool xrm::system::deviceFreeBuffer(int32_t devId) {
    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];

        if (dev->memBuffer) {
            free(dev->memBuffer);
            dev->memBuffer = NULL;
        }
        return (true);
    } else {
        return (false);
    }
}

bool xrm::system::deviceClearInfo(int32_t devId) {
    std::string dsaName;
#ifdef __SUPPORT_XRT__
    xclDeviceHandle deviceHandle;
    xclDeviceInfo2 deviceInfo;
#endif

    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];

        // device is busy
        if (isDeviceBusy(devId)) {
            return (false);
        }

        if (dev->memBuffer) {
            free(dev->memBuffer);
            dev->memBuffer = NULL;
        }
        /* save the required info */
        dsaName = dev->dsaName;
#ifdef __SUPPORT_XRT__
        deviceHandle = dev->deviceHandle;
        deviceInfo = dev->deviceInfo;
#endif
        flushDevData(devId);

        /* restore the info */
        dev->devId = devId;
        dev->dsaName = dsaName;
#ifdef __SUPPORT_XRT__
        dev->deviceHandle = deviceHandle;
        dev->deviceInfo = deviceInfo;
#endif
        return (true);
    }
    return (false);
}

int32_t xrm::system::xclbinReadFile(int32_t devId, std::string& name, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        return (XRM_ERROR_INVALID);
    }
    std::ifstream file(name.c_str(), std::ios::binary | std::ios::ate);
    if (!file.good()) {
        errmsg = "Failed to open xclbin file " + name;
        return (XRM_ERROR);
    }
    std::streamsize size = file.tellg();
    file.seekg(0, std::ios::beg);

    m_devList[devId].memBuffer = (char*)malloc(size);
    if (m_devList[devId].memBuffer == NULL) {
        errmsg = "Failed to alloce buffer for xclbin file " + name;
        return (XRM_ERROR);
    }
    if (!file.read(m_devList[devId].memBuffer, size)) {
        errmsg = "Failed to read xclbin file " + name;
        free(m_devList[devId].memBuffer);
        return (XRM_ERROR);
    }

    return (XRM_SUCCESS);
}

int32_t xrm::system::xclbinGetInfo(int32_t devId, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        return (XRM_ERROR_INVALID);
    }

    if (xclbinGetLayout(devId, errmsg) != XRM_SUCCESS) return (XRM_ERROR);
    if (xclbinGetUuid(devId, errmsg) != XRM_SUCCESS) return (XRM_ERROR);
    if (xclbinGetMemTopology(devId, errmsg) != XRM_SUCCESS) return (XRM_ERROR);
    if (xclbinGetConnectivity(devId, errmsg) != XRM_SUCCESS) return (XRM_ERROR);
    if (xclbinGetKeyvalues(devId, errmsg) != XRM_SUCCESS) return (XRM_ERROR);
    xclbinInformation* xclbinInfo = &m_devList[devId].xclbinInfo;
    for (int32_t cuIdx = 0; cuIdx < xclbinInfo->numCu; cuIdx++) {
        for (int32_t connectIdx = 0; connectIdx < xclbinInfo->numConnect; connectIdx++) {
            connectData* conn = &xclbinInfo->connectList[connectIdx];
            cuData* cu = &xclbinInfo->cuList[cuIdx];
            if (conn->ipLayoutIndex == cu->ipLayoutIndex) {
                memTopologyData* topology = &xclbinInfo->memTopologyList[conn->memDataIndex];
                cu->membankId = conn->memDataIndex;
                cu->membankType = topology->memType;
                cu->membankSize = topology->memSize;
                cu->membankBaseAddr = topology->memBaseAddress;
            }
        }
    }

    return (XRM_SUCCESS);
}

int32_t xrm::system::xclbinGetLayout(int32_t devId, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        return (XRM_ERROR_INVALID);
    }

    bool hardwareKernelExisting = false;
#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];
    char* buffer = m_devList[devId].memBuffer;
    xclbinInformation* xclbinInfo = &dev->xclbinInfo;
    axlf* xclbin = reinterpret_cast<axlf*>(buffer);
    xclbinInfo->numCu = 0;
    xclbinInfo->numHardwareCu = 0;
    xclbinInfo->numSoftwareCu = 0;
    std::string cuName;
    std::string instanceName;

    const axlf_section_header* ipHdr = xclbin::get_axlf_section(xclbin, IP_LAYOUT);
    if (ipHdr) {
        /* Handle the IP Hardware Kernel */
        char* data = &buffer[ipHdr->m_sectionOffset];
        const ip_layout* ipl = reinterpret_cast<ip_layout*>(data);
        for (int32_t i = 0; i < ipl->m_count; i++) {
            if (ipl->m_ip_data[i].m_type != IP_KERNEL) continue;
            xclbinInfo->numCu++;
            xclbinInfo->numHardwareCu++;
            if (xclbinInfo->numCu > XRM_MAX_XILINX_KERNELS) {
                errmsg = "cu number in xclbin is out range of " + std::to_string(XRM_MAX_XILINX_KERNELS);
                return (XRM_ERROR);
            }
            /* For hardware kernel, here should use the XRT interface to retrive cuId. */
            cuName = std::string((char*)ipl->m_ip_data[i].m_name);
            logMsg(XRM_LOG_NOTICE, "%s : cuName is %s\n", __func__, cuName.c_str());

            int32_t cuId = wrapIPName2Index(dev->deviceHandle, cuName.c_str());
            cuData* cu;
            if (cuId < 0) {
                if (cuId == XRM_ERROR_INVALID) {
                    // the function in XRT lib is NOT implemented, handle the cuId by XRM itself
                    cu = &xclbinInfo->cuList[xclbinInfo->numCu - 1];
                    cu->cuId = xclbinInfo->numCu - 1;
                } else {
                    // the function in XRT lib is implemented, return result is error
                    logMsg(XRM_LOG_ERROR, "%s : failed to get cu id of %s, ret = %d\n", __func__, cuName.c_str(), cuId);
                    return (XRM_ERROR);
                }
            } else {
                // the function in XRT lib is implemented, return result is good
                cu = &xclbinInfo->cuList[cuId];
                logMsg(XRM_LOG_NOTICE, "%s : cu id of %s is %d\n", __func__, cuName.c_str(), cuId);
                cu->cuId = cuId;
            }
            cu->ipLayoutIndex = xclbinInfo->numCu - 1; // the static index will be index of m_ip_data in xclbin file

            cu->cuType = CU_IPKERNEL;
            // cuName (m_name in m_ip_data) is "kerenlName:instanceName"
            cu->cuName = cuName;
            std::vector<std::string> results;
            // retrieve the kernelName
            boost::split(results, cu->cuName, [](char c) { return c == ':'; });
            cu->kernelName = results[0];
            instanceName = cuName.erase(0, cu->kernelName.length() + 1);
            cu->instanceName = instanceName;
            cu->baseAddr = ipl->m_ip_data[i].m_base_address;
            cuInitChannels(cu);
            cu->numClient = 0;
            cu->totalUsedLoadUnified = 0; // update dev load at end of loop
            cu->totalReservedLoadUnified = 0;
            cu->totalReservedUsedLoadUnified = 0;
            memset(cu->reserves, 0, sizeof(reserveData) * XRM_MAX_KERNEL_RESERVES);
            cu->numReserve = 0;
            cu->deviceId = devId;
            cu->deviceNumBDF = dev->devNumBDF;
            hardwareKernelExisting = true;
        } // end of hardware kernel handling
        /*
         * After all hardware kernel handled, it's assuming that all the previous cuId
         * is in range from 0 - current (xclbinInfo->numCu - 1).
         */

        /* Handle the Soft Kernel */
        for (const axlf_section_header* softKernelHdr = xclbin::get_axlf_section(xclbin, SOFT_KERNEL);
             softKernelHdr != NULL; softKernelHdr = xclbin::get_axlf_section_next(xclbin, softKernelHdr, SOFT_KERNEL)) {
            char* data = &buffer[softKernelHdr->m_sectionOffset];
            const soft_kernel* sk = reinterpret_cast<soft_kernel*>(data);
            if ((xclbinInfo->numCu + sk->m_num_instances) > XRM_MAX_XILINX_KERNELS) {
                errmsg = "soft kernel number in xclbin is out range of " + std::to_string(XRM_MAX_XILINX_KERNELS);
                return (XRM_ERROR);
            }

            for (uint32_t i = 0; i < sk->m_num_instances; i++) {
                xclbinInfo->numCu++;
                xclbinInfo->numSoftwareCu++;
                cuData* cu = &xclbinInfo->cuList[xclbinInfo->numCu - 1];
                cu->cuId = xclbinInfo->numCu - 1;
                cu->ipLayoutIndex = xclbinInfo->numCu - 1; // the static index will be index of m_ip_data in xclbin file
                cu->cuType = CU_SOFTKERNEL;
                /*
                 * mpo_name in soft_kernel is just "kernelName", but not "kernelName:instanceName"
                 * since instanceName of soft kernel is not descripted in xclbin, so cuName is same as kernelName.
                 */
                cu->kernelName = std::string((char*)&buffer[softKernelHdr->m_sectionOffset + sk->mpo_name]);
                cu->cuName = cu->kernelName;
                logMsg(XRM_LOG_NOTICE, "%s soft kernel name: %s\n", __func__, cu->kernelName.c_str());
                cuInitChannels(cu);
                cu->numClient = 0;
                cu->totalUsedLoadUnified = 0;
                cu->totalReservedLoadUnified = 0;
                cu->totalReservedUsedLoadUnified = 0;
                memset(cu->reserves, 0, sizeof(reserveData) * XRM_MAX_KERNEL_RESERVES);
                cu->numReserve = 0;
                cu->deviceId = devId;
                cu->deviceNumBDF = dev->devNumBDF;
            }
        } // end of soft kernel handling
    }
    updateDeviceLoad(devId, 0, 0);
#endif
    /* the xclbin should contain at least one hardware kernel */
    if (hardwareKernelExisting) {
        return (XRM_SUCCESS);
    } else {
        errmsg = "No hardware kernel existing in xclbin";
        return (XRM_ERROR);
    }
}

void xrm::system::cuInitChannels(cuData* cu) {
    int32_t i;

    if (cu == NULL) {
        logMsg(XRM_LOG_ERROR, "%s : init channels, the cu is NULL\n", __func__);
        return;
    }

    for (i = 0; i < XRM_MAX_KERNEL_CHANNELS; i++) {
        cu->channels[i].channelId = i;
        cu->channels[i].allocServiceId = 0;
        cu->channels[i].poolId = 0;
        cu->channels[i].clientId = 0;
        cu->channels[i].clientProcessId = 0;
        cu->channels[i].channelLoadUnified = 0;
        cu->channels[i].channelLoadOriginal = 0;
    }
    cu->numChanInuse = 0;
}

/*
 * ret: first free channel id,
 *    : -1, no free channel
 */
int32_t xrm::system::cuFindFreeChannelId(cuData* cu) {
    int32_t i;
    int32_t ret = -1;

    if (cu == NULL) return (ret);

    if (cu->numChanInuse >= XRM_MAX_KERNEL_CHANNELS) return (ret);

    for (i = 0; i < XRM_MAX_KERNEL_CHANNELS; i++) {
        /* channelLoadUnified is 0, clientId are also 0 */
        if (cu->channels[i].channelLoadUnified == 0) {
            ret = i;
            break;
        }
    }
    return (ret);
}

/*
 * To convert bin array to hex string.
 */
void xrm::system::binToHexstr(unsigned char* in, int32_t insz, std::string& outStr) {
    unsigned char* pin = in;
    const char* hex = "0123456789abcdef";
    for (; pin < (in + insz); pin++) {
        outStr = outStr + hex[(*pin >> 4) & 0xF];
        outStr = outStr + hex[*pin & 0xF];
    }
}

/*
 * To conver hex string like "0123456789abcdef" to bin 0123456789abcdef
 */
void xrm::system::hexstrToBin(std::string& inStr, int32_t insz, unsigned char* out) {
    char in[insz];
    unsigned char* pout = out;
    unsigned char hex_0, hex_1;
    strncpy(in, inStr.c_str(), insz);
    for (int32_t i = 0; i < insz; i += 2, pout++) {
        if (in[i] < 'a')
            hex_0 = (in[i] - '0');
        else
            hex_0 = (in[i] - 'a') + 10;
        if (in[i + 1] < 'a')
            hex_1 = (in[i + 1] - '0');
        else
            hex_1 = (in[i + 1] - 'a') + 10;
        *pout = ((hex_0 << 4) & 0xF0) | (hex_1 & 0xF);
    }
}

int32_t xrm::system::xclbinGetUuid(int32_t devId, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        return (XRM_ERROR_INVALID);
    }

#ifdef __SUPPORT_XRT__
    axlf* xclbin = reinterpret_cast<axlf*>(m_devList[devId].memBuffer);
    uuid_copy(m_devList[devId].xclbinInfo.uuid, xclbin->m_header.uuid);
    /* convert uuid to string */
    binToHexstr((unsigned char*)&m_devList[devId].xclbinInfo.uuid, sizeof(uuid_t), m_devList[devId].xclbinInfo.uuidStr);
#endif
    return (XRM_SUCCESS);
}

int32_t xrm::system::xclbinGetKeyvalues(int32_t devId, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        return (XRM_ERROR_INVALID);
    }

#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];
    char* buffer = m_devList[devId].memBuffer;
    axlf* xclbin = reinterpret_cast<axlf*>(buffer);
    const axlf_section_header* kvHdr = xclbin::get_axlf_section(xclbin, KEYVALUE_METADATA);
    if (kvHdr) {
        auto size = kvHdr->m_sectionSize;
        char* data = &buffer[kvHdr->m_sectionOffset];
        if (size == 0 || data == NULL) {
            logMsg(XRM_LOG_NOTICE, "%s : There is no Keyvalues metadata in xclbin file\n", __func__);
            return (XRM_SUCCESS);
        }
        std::string kvData(data, size);

        // Fix-up the metadata string
        boost::erase_all(kvData, "\\");
        boost::replace_all(kvData, "\"{", "{");
        boost::replace_all(kvData, "}\"", "}");

        std::stringstream kvStr;
        kvStr << kvData;

        pt::ptree kvTree;
        try {
            pt::read_json(kvStr, kvTree);
        } catch (const boost::property_tree::json_parser_error& e) {
            errmsg = "Keyvalues metadata Json format is not right: " + e.message();
            return (XRM_ERROR_INVALID);
        }

        /*
         * The kernel name is unique in system, so xrm can identify the kernel with the name.
         */
        auto keyValues = kvTree.get_child("key_values");
        for (const auto& kviter : keyValues) {
            std::string kernelName = kviter.second.get<std::string>("key", "");
            if (kernelName.c_str()[0] == '\0') continue;
            logMsg(XRM_LOG_NOTICE, "%s : kernel name in metadata is %s\n", __func__, kernelName.c_str());
            auto value = kviter.second.get_child("value");
            std::string alias = value.get("alias", "");
            logMsg(XRM_LOG_NOTICE, "%s : alias in metadata is %s\n", __func__, alias.c_str());
            std::string kernelPluginFileName = value.get("plugin", "");
            logMsg(XRM_LOG_NOTICE, "%s : kernelPluginFileName in metadata is %s\n", __func__,
                   kernelPluginFileName.c_str());
            std::string maxCapacity = value.get("maxCapacity", "");
            logMsg(XRM_LOG_NOTICE, "%s : maxCapacity in metadata is %s\n", __func__, maxCapacity.c_str());
            for (int32_t i = 0; i < dev->xclbinInfo.numCu; i++) {
                cuData* cu = &dev->xclbinInfo.cuList[i];
                logMsg(XRM_LOG_NOTICE, "%s : kernel name in IP Layout is %s\n", __func__, cu->kernelName.c_str());
                if (kernelName.compare(cu->kernelName) == 0) {
                    logMsg(XRM_LOG_NOTICE, "%s : matching the kernel name\n", __func__);
                    if (alias.c_str()[0] != '\0') cu->kernelAlias = alias;
                    if (kernelPluginFileName.c_str()[0] != '\0') cu->kernelPluginFileName = kernelPluginFileName;
                    if (maxCapacity.c_str()[0] != '\0') cu->maxCapacity = std::strtoull(maxCapacity.c_str(), NULL, 0);
                }
            }
        }
        logMsg(XRM_LOG_NOTICE, "%s : successed handling key value\n", __func__);
        return (XRM_SUCCESS);
    } else {
        logMsg(XRM_LOG_NOTICE, "%s : There is no metadata session in the xclbin file.\n", __func__);
        return (XRM_SUCCESS);
    }
#endif
    return (XRM_SUCCESS);
}

int32_t xrm::system::xclbinGetMemTopology(int32_t devId, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        return (XRM_ERROR_INVALID);
    }

#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];
    char* buffer = m_devList[devId].memBuffer;
    axlf* xclbin = reinterpret_cast<axlf*>(buffer);

    const axlf_section_header* ipHdr = xclbin::get_axlf_section(xclbin, MEM_TOPOLOGY);
    if (ipHdr) {
        char* data = &buffer[ipHdr->m_sectionOffset];
        const mem_topology* memTopo = reinterpret_cast<mem_topology*>(data);
        dev->xclbinInfo.numMemBank = memTopo->m_count;
        logMsg(XRM_LOG_NOTICE, "MEM TOPOLOGY - %d banks\n", dev->xclbinInfo.numMemBank);
        for (int i = 0; i < memTopo->m_count; i++) {
            memTopologyData* topology = &dev->xclbinInfo.memTopologyList[i];
            topology->memType = memTopo->m_mem_data[i].m_type;
            topology->memUsed = memTopo->m_mem_data[i].m_used;
            topology->memSize = memTopo->m_mem_data[i].m_size;
            topology->memBaseAddress = memTopo->m_mem_data[i].m_base_address;
            memcpy(topology->memTag, memTopo->m_mem_data[i].m_tag, XRM_MAX_DDR_MAP * sizeof(unsigned char));
            logMsg(XRM_LOG_NOTICE, "index=%d, tag=%s, type = %d, used = %d, size = %lx, base = %lx\n", i,
                   topology->memTag, topology->memType, topology->memUsed, topology->memSize, topology->memBaseAddress);
        }
    } else {
        logMsg(XRM_LOG_NOTICE, "Could not find MEM_TOPOLOGY in xclbin");
    }
#endif

    return (XRM_SUCCESS);
}

int32_t xrm::system::xclbinGetConnectivity(int32_t devId, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        return (XRM_ERROR_INVALID);
    }

#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];
    char* buffer = m_devList[devId].memBuffer;
    axlf* xclbin = reinterpret_cast<axlf*>(buffer);

    const axlf_section_header* ipHdr = xclbin::get_axlf_section(xclbin, CONNECTIVITY);
    if (ipHdr) {
        char* data = &buffer[ipHdr->m_sectionOffset];
        const connectivity* axlfConn = reinterpret_cast<connectivity*>(data);
        dev->xclbinInfo.numConnect = axlfConn->m_count;
        logMsg(XRM_LOG_NOTICE, "CONNECTIVITY - %d connections\n", dev->xclbinInfo.numConnect);
        for (int i = 0; i < axlfConn->m_count; i++) {
            connectData* conn = &dev->xclbinInfo.connectList[i];
            conn->argIndex = axlfConn->m_connection[i].arg_index;
            conn->ipLayoutIndex = axlfConn->m_connection[i].m_ip_layout_index;
            conn->memDataIndex = axlfConn->m_connection[i].mem_data_index;
            logMsg(XRM_LOG_NOTICE, "index = %d, arg_idx = %d, ip_idx = %d, mem_idx = %d\n", i, conn->argIndex,
                   conn->ipLayoutIndex, conn->memDataIndex);
        }
    } else {
        logMsg(XRM_LOG_NOTICE, "Could not find CONNECTIVITY in xclbin");
    }
#endif

    return (XRM_SUCCESS);
}

int32_t xrm::system::xclbinLoadToDevice(int32_t devId, std::string& errmsg) {
    if (devId < 0 || devId >= m_numDevice) {
        errmsg = "device id " + std::to_string(devId) + " is not found";
        return (XRM_ERROR_INVALID);
    }

    logMsg(XRM_LOG_NOTICE, "%s load xclbin to device %d", __func__, devId);

    int32_t rc = 0;
#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];
    char* buffer = m_devList[devId].memBuffer;
    xclDeviceHandle deviceHandle = dev->deviceHandle;
    rc = wrapLockDevice(deviceHandle);
    if (rc != 0) {
        errmsg = "Failed to lock device " + std::to_string(devId);
        return (rc);
    }
    rc = xclLoadXclBin(deviceHandle, (const xclBin*)buffer);
    if (rc != 0) errmsg = "xclLoadXclBin failed, rc = " + std::to_string(rc);
    wrapUnlockDevice(deviceHandle);
#endif
    return (rc);
}

int32_t xrm::system::deviceLockXclbin(int32_t devId) {
    int32_t ret = XRM_SUCCESS;
    if (devId < 0 || devId >= m_numDevice) {
        return (XRM_ERROR_INVALID);
    }

#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];
    xclbinInformation* xclbinInfo = &m_devList[devId].xclbinInfo;

    /* To use xclOpenContext() with ipIndex as -1 to ocupie the resouce so that xclLoadXclBin()
     * can NOT load new xclbin to replace current loaded xclbin.
     */
    ret = xclOpenContext(dev->deviceHandle, xclbinInfo->uuid, -1, true);
#endif
    if (ret != 0) {
        logMsg(XRM_LOG_ERROR, "%s Failed to lock down xclbin to device %d", __func__, devId);
        return (ret);
    }
    return (XRM_SUCCESS);
}

int32_t xrm::system::deviceUnlockXclbin(int32_t devId) {
    int32_t ret = XRM_SUCCESS;
    if (devId < 0 || devId >= m_numDevice) {
        return (XRM_ERROR_INVALID);
    }

#ifdef __SUPPORT_XRT__
    deviceData* dev = &m_devList[devId];
    if (!dev->isLoaded) return (XRM_SUCCESS);

    xclbinInformation* xclbinInfo = &m_devList[devId].xclbinInfo;

    /* To use xclCloseContext() with ipIndex as -1 to release the resouce so that xclLoadXclBin()
     * can load new xclbin to replace current loaded xclbin.
     */
    ret = xclCloseContext(dev->deviceHandle, xclbinInfo->uuid, -1);
#endif
    if (ret != 0) {
        logMsg(XRM_LOG_ERROR, "%s Failed to unlock xclbin from device %d", __func__, devId);
        return (ret);
    }
    return (XRM_SUCCESS);
}

void xrm::system::deviceDumpResource(int32_t devId) {
    int32_t i, j;
    if (devId < 0 || devId >= m_numDevice) return;

    deviceData* dev = &m_devList[devId];
    xclbinInformation* xclbinInfo = &m_devList[devId].xclbinInfo;

    logMsg(XRM_LOG_NOTICE, "%s(): deviceId : %d", __func__, devId);
    logMsg(XRM_LOG_NOTICE, "%s(): deviceNumBDF : %d", __func__, dev->devNumBDF);
    logMsg(XRM_LOG_NOTICE, "%s(): isDisabled : %d", __func__, dev->isDisabled);
    logMsg(XRM_LOG_NOTICE, "%s(): isLoaded : %d", __func__, dev->isLoaded);
    if (!dev->isLoaded) return;
    logMsg(XRM_LOG_NOTICE, "%s(): uuid : %s", __func__, xclbinInfo->uuidStr.c_str());
#if 0
    for (i = 0; i < XRM_MAX_DEV_CLIENTS; i++)
    {
        logMsg(XRM_LOG_NOTICE, "%s() dev[%d] clientProcs[%d] : clientId: %lu, ref: %d",
               __func__, devId, i, dev->clientProcs[i].clientId,
               dev->clientProcs[i].ref);
    }
#endif
    logMsg(XRM_LOG_NOTICE, "%s(): numCu : %d", __func__, xclbinInfo->numCu);
    logMsg(XRM_LOG_NOTICE, "%s(): numHardwareCu : %d", __func__, xclbinInfo->numHardwareCu);
    logMsg(XRM_LOG_NOTICE, "%s(): numSoftwareCu : %d", __func__, xclbinInfo->numSoftwareCu);
    for (i = 0; i < xclbinInfo->numCu; i++) {
        cuData* cu = &xclbinInfo->cuList[i];
        logMsg(XRM_LOG_NOTICE, "%s() cu : cuId :%d", __func__, cu->cuId);
        logMsg(XRM_LOG_NOTICE, "%s() cu : ipLayoutIndex :%d", __func__, cu->ipLayoutIndex);
        logMsg(XRM_LOG_NOTICE, "%s() cu : cuType :%d", __func__, cu->cuType);
        logMsg(XRM_LOG_NOTICE, "%s() cu : kernelName :%s", __func__, cu->kernelName.c_str());
        logMsg(XRM_LOG_NOTICE, "%s() cu : kernelAlias :%s", __func__, cu->kernelAlias.c_str());
        logMsg(XRM_LOG_NOTICE, "%s() cu : cuName :%s", __func__, cu->cuName.c_str());
        logMsg(XRM_LOG_NOTICE, "%s() cu : instanceName :%s", __func__, cu->instanceName.c_str());
        logMsg(XRM_LOG_NOTICE, "%s() cu : baseAddr :%lu", __func__, cu->baseAddr);
        logMsg(XRM_LOG_NOTICE, "%s() cu : membankId :%d", __func__, cu->membankId);
        logMsg(XRM_LOG_NOTICE, "%s() cu : membankType :%d", __func__, cu->membankId);
        logMsg(XRM_LOG_NOTICE, "%s() cu : membankSize :%lu", __func__, cu->membankSize);
        logMsg(XRM_LOG_NOTICE, "%s() cu : membankBaseAddr :%lu", __func__, cu->membankBaseAddr);
        logMsg(XRM_LOG_NOTICE, "%s() cu : kernelPluginFileName :%s", __func__, cu->kernelPluginFileName.c_str());
        logMsg(XRM_LOG_NOTICE, "%s() cu : maxCapacity :%lu", __func__, cu->maxCapacity);
        logMsg(XRM_LOG_NOTICE, "%s() cu : numChanInuse :%d", __func__, cu->numChanInuse);
        logMsg(XRM_LOG_NOTICE, "%s() cu : numReserve :%d", __func__, cu->numReserve);
        logMsg(XRM_LOG_NOTICE, "%s() cu : totalUsedLoadUnified :%d", __func__, cu->totalUsedLoadUnified);
        logMsg(XRM_LOG_NOTICE, "%s() cu : totalReservedLoadUnified :%d", __func__, cu->totalReservedLoadUnified);
        for (j = 0; j < XRM_MAX_KERNEL_CHANNELS; j++) {
            channelData* chan = &cu->channels[j];
            logMsg(XRM_LOG_NOTICE, "%s() cu : clients[%d] : %d", __func__, j, cu->clients[j]);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] channelId: %d", __func__, j, chan->channelId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] allocServiceId: %lu", __func__, j, chan->allocServiceId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] poolId: %lu", __func__, j, chan->poolId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] clientId: %lu", __func__, j, chan->clientId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] clientProcessId: %lu", __func__, j, chan->clientProcessId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] channelLoadUnified: %d", __func__, j,
                   chan->channelLoadUnified);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] channelLoadOriginal: %d", __func__, j,
                   chan->channelLoadOriginal);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] logic group: indicator: %d", __func__, j, chan->logicGroupInfo.indicator);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] logic group: group id: %d", __func__, j, chan->logicGroupInfo.groupId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : channels[%d] logic group: group type: %s", __func__, j, chan->logicGroupInfo.groupType);
        }
        cu->totalReservedUsedLoadUnified = 0;
        for (j = 0; j < XRM_MAX_KERNEL_RESERVES; j++) {
            reserveData* reserve = &cu->reserves[j];
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] reservePoolId: %lu", __func__, j, reserve->reservePoolId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] reserveLoadUnified: %d", __func__, j,
                   reserve->reserveLoadUnified);
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] reserveUsedLoadUnified: %d", __func__, j,
                   reserve->reserveUsedLoadUnified);
            cu->totalReservedUsedLoadUnified += reserve->reserveUsedLoadUnified;
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] clientIsActive: %d", __func__, j, reserve->clientIsActive);
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] clientId: %lu", __func__, j, reserve->clientId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] clientProcessId: %lu", __func__, j,
                   reserve->clientProcessId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] logic group: indicator: %d", __func__, j, reserve->logicGroupInfo.indicator);
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] logic group: group id: %d", __func__, j, reserve->logicGroupInfo.groupId);
            logMsg(XRM_LOG_NOTICE, "%s() cu : reserves[%d] logic group: group type: %s", __func__, j, reserve->logicGroupInfo.groupType);
        }
        logMsg(XRM_LOG_NOTICE, "%s() cu : totalReservedUsedLoadUnified :%d", __func__,
               cu->totalReservedUsedLoadUnified);
    }
    logMsg(XRM_LOG_NOTICE, "%s(): numLogicGroup : %d", __func__, xclbinInfo->numLogicGroup);
    for ( i= 0; i < xclbinInfo->numLogicGroup; i++) {
        logicGroupData* lg = &xclbinInfo->logicGroups[i];
        logMsg(XRM_LOG_NOTICE, "%s() logic group : logicGroupId :%d", __func__, lg->logicGroupId);
        logMsg(XRM_LOG_NOTICE, "%s() logic group : numCu :%d", __func__, lg->numCu);
        for (j = 0; j < lg->numCu; j++) {
            logMsg(XRM_LOG_NOTICE, "%s() cuId [%d] : %d", __func__, j, lg->cuId[j]);
        }
    }
}

/*
 * To get the cu max capacity based on kernel name or alias or both
 *
 * return:
 *        0: cu not specified or not found
 *        > 0: max capacity of the cu
 */
uint64_t xrm::system::getCuMaxCapacity(cuProperty* cuProp) {
    if (cuProp == NULL) return (0);

    uint64_t maxCapacity = 0;
    uint64_t maxCapacityWithAlias = 0;
    std::string name = cuProp->kernelName;
    std::string alias = cuProp->kernelAlias;
    deviceData* dev;
    cuData* cu;
    int32_t devId, cuId;
    bool found;

    if ((name[0] == '\0') && (alias[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "%s : neither name nor alias are presented", __func__);
        return (0);
    }

    if (name[0] != '\0') {
        /* kernel name is presented */
        found = false;
        for (devId = 0; devId < m_numDevice; devId++) {
            dev = &m_devList[devId];
            if (dev->isLoaded) {
                for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
                    cu = &dev->xclbinInfo.cuList[cuId];
                    if (name.compare(cu->kernelName.c_str()) == 0) {
                        found = true;
                        maxCapacity = cu->maxCapacity;
                        break;
                    }
                }
            }
            if (found) break;
        }
        if (found) {
            logMsg(XRM_LOG_NOTICE, "%s : maxCapacity with name is %lu", __func__, maxCapacity);
        } else {
            logMsg(XRM_LOG_NOTICE, "%s : cu name%s not found", __func__, name.c_str());
        }

        /* kernel alias is also presented */
        if (alias[0] != '\0') {
            found = false;
            for (devId = 0; devId < m_numDevice; devId++) {
                dev = &m_devList[devId];
                if (dev->isLoaded) {
                    for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
                        cu = &dev->xclbinInfo.cuList[cuId];
                        if (alias.compare(cu->kernelAlias.c_str()) == 0) {
                            found = true;
                            maxCapacityWithAlias = cu->maxCapacity;
                            break;
                        }
                    }
                }
                if (found) break;
            }
            if (found) {
                logMsg(XRM_LOG_NOTICE, "%s : maxCapacity with alias is %lu", __func__, maxCapacityWithAlias);
            } else {
                logMsg(XRM_LOG_NOTICE, "%s : cu alias%s not found", __func__, alias.c_str());
            }

            if (maxCapacity != maxCapacityWithAlias) {
                logMsg(XRM_LOG_NOTICE, "%s : kernel name and alias mismatch", __func__);
                maxCapacity = 0;
            }
        }
    } else // alias[0] != '\0'
    {
        /* kernel alias is presented */
        found = false;
        for (devId = 0; devId < m_numDevice; devId++) {
            dev = &m_devList[devId];
            if (dev->isLoaded) {
                for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
                    cu = &dev->xclbinInfo.cuList[cuId];
                    if (alias.compare(cu->kernelAlias.c_str()) == 0) {
                        found = true;
                        maxCapacityWithAlias = cu->maxCapacity;
                        break;
                    }
                }
            }
            if (found) break;
        }
        if (found) {
            logMsg(XRM_LOG_NOTICE, "%s : maxCapacity with alias is %lu", __func__, maxCapacityWithAlias);
        } else {
            logMsg(XRM_LOG_NOTICE, "%s : cu alias%s not found", __func__, alias.c_str());
        }
        maxCapacity = maxCapacityWithAlias;
    }

    return (maxCapacity);
}

void xrm::system::initLock() {
    pthread_mutex_init(&m_lock, NULL);
}

/*
 * Considering the xrmadm cli and xrm lib are using asio to do the IPC with xrmd deamon,
 * So the daemon is running as a single process there is no need to use lock in xrmd
 * daemon to protect the resource pool data.
 */
void xrm::system::enterLock() {
#if 1
    pthread_mutex_lock(&m_lock);
#endif
}

void xrm::system::exitLock() {
#if 1
    pthread_mutex_unlock(&m_lock);
#endif
}

inline void xrm::system::save() {
    std::string fname = "/dev/shm/xrm.data";
    /* create and open the character archive for output */
    std::ofstream ofs(fname.c_str());
    /* write class instance to archive */
    boost::archive::text_oarchive ar(ofs);
    ar&* this;
}

/*
 * call while holding lock
 */
bool xrm::system::restore() {
    bool rc = false;

    std::string fname = "/dev/shm/xrm.data";
    /* create and open the archive for input */
    std::ifstream ifs(fname.c_str());
    if (ifs.is_open()) {
        logMsg(XRM_LOG_NOTICE, "Database found, reloading daemon");
        /* read class from archive */
        boost::archive::text_iarchive ar(ifs);
        ar&* this;
        initConfig();
        initLibVersionDepFunctions();
        for (int32_t devId = 0; devId < m_numDevice; devId++) {
            if (!m_devList[devId].isDisabled) openDevice(devId);
        }
        rc = true;
    } else
        logMsg(XRM_LOG_NOTICE, "No database found, starting from fresh state");

    return (rc);
}

/* input parameters valid check need to be done before the calling */
bool xrm::system::isDeviceLoaded(int32_t devId) {
    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];
        return (dev->isLoaded);
    } else {
        return (false);
    }
}

/* input parameters valid check need to be done before the calling */
bool xrm::system::isDeviceLoadedXclbinWithUuid(int32_t devId, uuid_t uuid) {
    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];
        bool sameUuid = true;
        size_t size = sizeof(uuid_t);
        unsigned char* uuidOfDev = dev->xclbinInfo.uuid;

        for (size_t i = 0; i < size; i++) {
            if ((uint8_t)(uuidOfDev[i]) != (uint8_t)(uuid[i])) {
                sameUuid = false;
                break;
            }
        }
        return (sameUuid);
    } else {
        return (false);
    }
}

/* input parameters valid check need to be done before the calling */
bool xrm::system::isDeviceBusy(int32_t devId) {
    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];

        for (int32_t cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            if (dev->xclbinInfo.cuList[cuId].totalUsedLoadUnified) return (true);
        }
        return (false);
    } else {
        return (false);
    }
}

/*
 * the unload process should be protected by system lock
 */
int32_t xrm::system::unloadOneDevice(const int32_t& devId, std::string& errmsg) {
    /*
     * XRM daemon may start before FPGA device ready during boot,
     * so try to init devices at the first time using devices.
     */
    if (!m_devicesInited) initDevices();

    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];

        if (!dev->isLoaded) {
            /*
             * if device is not loaded with xclbin, then the device will
             * still keep unload state.
             */
            deviceClearInfo(devId);
            return (XRM_SUCCESS);
        }

        // device is busy
        if (isDeviceBusy(devId)) {
            errmsg = "Device  " + std::to_string(devId) + " is busy";
            return (XRM_ERROR_DEVICE_IS_BUSY);
        }

        // unlock loaded xclbin from device
        if (deviceUnlockXclbin(devId)) {
            errmsg = "Fail to unlock xclbin from device  " + std::to_string(devId);
            return (XRM_ERROR_DEVICE_IS_LOCKED);
        }

        deviceClearInfo(devId);
        return (XRM_SUCCESS);
    }
    errmsg = "Invalid device id [" + std::to_string(devId) + "] passed in";
    return (XRM_ERROR_INVALID);
}

/*
 * the device enable process should be protected by system lock
 */
int32_t xrm::system::enableOneDevice(const int32_t& devId, std::string& errmsg) {
    /*
     * XRM daemon may start before FPGA device ready during boot,
     * so try to init devices at the first time using devices.
     */
    if (!m_devicesInited) initDevices();

    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];

        dev->isDisabled = false;
        return (XRM_SUCCESS);
    }
    errmsg = "Invalid device id [" + std::to_string(devId) + "] passed in";
    return (XRM_ERROR_INVALID);
}

/*
 * the device disable process should be protected by system lock
 */
int32_t xrm::system::disableOneDevice(const int32_t& devId, std::string& errmsg) {
    /*
     * XRM daemon may start before FPGA device ready during boot,
     * so try to init devices at the first time using devices.
     */
    if (!m_devicesInited) initDevices();

    if (devId >= 0 && devId < m_numDevice) {
        deviceData* dev = &m_devList[devId];

        if (dev->isDisabled) {
            return (XRM_SUCCESS);
        }
        if (!dev->isLoaded) {
            dev->isDisabled = true;
            return (XRM_SUCCESS);
        }

        // device is busy
        if (isDeviceBusy(devId)) {
            errmsg = "Device  " + std::to_string(devId) + " is busy";
            return (XRM_ERROR_DEVICE_IS_BUSY);
        }

        // unlock loaded xclbin from device
        if (deviceUnlockXclbin(devId)) {
            errmsg = "Fail to unlock xclbin from device  " + std::to_string(devId);
            return (XRM_ERROR_DEVICE_IS_LOCKED);
        }

        deviceClearInfo(devId);
        dev->isDisabled = true;
        return (XRM_SUCCESS);
    }
    errmsg = "Invalid device id [" + std::to_string(devId) + "] passed in";
    return (XRM_ERROR_INVALID);
}

/*
 * Check whether the cu (compute unit) existing on specified device
 *
 * true: cu is existing on device
 * false: cu is NOT existing on device
 *
 * Lock: should enter lock before calling the function
 */
bool xrm::system::resIsCuExistingOnDev(int32_t devId, cuProperty* cuProp) {
    deviceData* dev;
    cuData* cu;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    int32_t cuId;
    bool cuFound = false;

    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0') && (cuProp->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (cuFound);
    }

    dev = &m_devList[devId];
    /*
     * Now check whether matching cu is on allocated device
     * if not, free device, increment dev count, re-loop
     */
    for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuFound; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];

        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
        cuFound = true;
    }
    return (cuFound);
}

/*
 * Check whether the cu (compute unit) existing in system
 *
 * true: cu is existing in system
 * false: cu is NOT existing in system
 *
 * Lock: should enter lock during the checking so that the result is accurate.
 */
bool xrm::system::resIsCuExisting(cuProperty* cuProp) {
    deviceData* dev;
    int32_t devId;
    bool cuFound = false;

    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0') && (cuProp->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (cuFound);
    }

    for (devId = 0; !cuFound && (devId < m_numDevice); devId++) {
        dev = &m_devList[devId];
        if (!dev->isLoaded) continue;
        cuFound = resIsCuExistingOnDev(devId, cuProp);
    }
    return (cuFound);
}

/*
 * Check whether the cu list existing in system
 *
 * true: cu list is existing in system
 * false: cu list is NOT existing in system
 *
 * Lock: should enter lock during the checking so that the result is accurate.
 */
bool xrm::system::resIsCuListExisting(cuListProperty* cuListProp) {
    deviceData* dev;
    int32_t i, devId;
    cuProperty* cuProp = NULL;
    bool cuListFound = false;
    bool allCuFound = false;
    bool cuFound = false;

    if (cuListProp->cuNum <= 0 || cuListProp->cuNum > XRM_MAX_LIST_CU_NUM) {
        return (cuListFound);
    }

    if (cuListProp->sameDevice) {
        /*
         * To find the cu list from same device
         */
        for (devId = 0; (devId < m_numDevice) && !cuListFound; devId++) {
            dev = &m_devList[devId];
            if (!dev->isLoaded) continue;
            /* find the cu list from this device */
            allCuFound = true;
            for (i = 0; i < cuListProp->cuNum; i++) {
                cuProp = &cuListProp->cuProps[i];
                if (!resIsCuExistingOnDev(devId, cuProp)) {
                    allCuFound = false;
                    break;
                }
            }
            cuListFound = allCuFound;
        }
    } else {
        /*
         * To find the cu list from any device
         */
        for (i = 0; i < cuListProp->cuNum; i++) {
            cuProp = &cuListProp->cuProps[i];
            /* find the cu from all device */
            for (devId = 0; (devId < m_numDevice) && !cuFound; devId++) {
                dev = &m_devList[devId];
                if (!dev->isLoaded) continue;
                /* find the cu from this device */
                if (resIsCuExistingOnDev(devId, cuProp)) {
                    cuFound = true;
                    break;
                }
            }
            if (!cuFound) break;
            if (i == (cuListProp->cuNum - 1)) cuListFound = true;
        }
    }
    return (cuListFound);
}

/*
 * Check whether the cu group existing in system
 *
 * true: cu group is existing in system
 * false: cu group is NOT existing in system
 *
 * Lock: will not use lock since it will call resIsCuListExisting() to check the resource.
 */
bool xrm::system::resIsCuGroupExisting(cuGroupProperty* cuGroupProp) {
    cuListProperty cuListProp;
    cuListProperty* udfCuListProp;
    cuProperty* cuProp;
    cuProperty* udfCuProp;
    bool cuGroupFound = false;
    int32_t udfCuGroupIdx = -1;

    if (cuGroupProp == NULL) {
        return (cuGroupFound);
    }
    for (uint32_t cuGroupIdx = 0; cuGroupIdx < m_numUdfCuGroup; cuGroupIdx++) {
        /* compare, 0: equal */
        if (!m_udfCuGroups[cuGroupIdx].udfCuGroupName.compare(cuGroupProp->udfCuGroupName.c_str())) {
            udfCuGroupIdx = cuGroupIdx;
            break;
        }
    }
    if (udfCuGroupIdx == -1) {
        logMsg(XRM_LOG_ERROR, "%s : user defined cu group is not declared\n", __func__);
        return (cuGroupFound);
    }
    for (int32_t cuListIdx = 0; cuListIdx < m_udfCuGroups[udfCuGroupIdx].optionUdfCuListNum; cuListIdx++) {
        memset(&cuListProp, 0, sizeof(cuListProperty));
        udfCuListProp = &m_udfCuGroups[udfCuGroupIdx].optionUdfCuListProps[cuListIdx];
        cuListProp.cuNum = udfCuListProp->cuNum;
        cuListProp.sameDevice = udfCuListProp->sameDevice;
        for (int32_t i = 0; i < cuListProp.cuNum; i++) {
            cuProp = &cuListProp.cuProps[i];
            udfCuProp = &udfCuListProp->cuProps[i];
            strncpy(cuProp->cuName, udfCuProp->cuName, XRM_MAX_NAME_LEN - 1);
        }
        cuGroupFound = resIsCuListExisting(&cuListProp);
        if (cuGroupFound) break;
    }
    return (cuGroupFound);
}

/*
 * Alloc one cu (compute unit) based on the request property.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocCu(cuProperty* cuProp, cuResource* cuRes, bool updateId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuProp->clientId;
    int32_t devId, cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    if ((cuProp == NULL) || (cuRes == NULL)) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0') && (cuProp->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (XRM_ERROR_INVALID);
    }

cu_alloc_loop:
    for (devId = -1; !cuAcquired && (devId < m_numDevice);) {
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            break;
        }

        ret = allocClientFromDev(devId, cuProp);
        if (ret < 0) {
            continue;
        }
        dev = &m_devList[devId];
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to re-use existing kernels; else, use a new kernel */
            if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            /* cu name is presented, compare it; otherwise no need to compare */
            cuNameEqual = true;
            if (cuProp->cuName[0] != '\0') {
                if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
            /* alloc channel and register client id */
            ret = allocChanClientFromCu(cu, cuProp, cuRes);
            if (ret != XRM_SUCCESS) {
                continue;
            }

            cuRes->deviceId = devId;
            cuRes->deviceNumBDF = dev->devNumBDF;
            cuRes->cuId = cuId;
            strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
            strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
            cuAcquired = true;
        }

        if (!cuAcquired) {
            releaseClientOnDev(devId, clientId);
        }
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_alloc_loop;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Alloc one cu (compute unit) based on the request property version 2.
 *
 * it's only used for single CU allocation, it's not called by multiple cu allocation.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocCuV2(cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    if ((cuPropV2 == NULL) || (cuRes == NULL)) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuPropV2->kernelName[0] == '\0') && (cuPropV2->kernelAlias[0] == '\0') && (cuPropV2->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (XRM_ERROR_INVALID);
    }
    uint64_t deviceInfoConstraintType =
        (cuPropV2->deviceInfo >> XRM_DEVICE_INFO_CONSTRAINT_TYPE_SHIFT) & XRM_DEVICE_INFO_CONSTRAINT_TYPE_MASK;
    uint64_t deviceInfoDeviceIndex =
        (cuPropV2->deviceInfo >> XRM_DEVICE_INFO_DEVICE_INDEX_SHIFT) & XRM_DEVICE_INFO_DEVICE_INDEX_MASK;

    /*
     * 1) device constraint is handled here.
     * 2) allocation policy is handled here.
     * 3) login group constraint is handled in different cu allocation sub functions.
     */
    switch (deviceInfoConstraintType) {
        case XRM_DEVICE_INFO_CONSTRAINT_TYPE_NULL: {
            if (cuPropV2->policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_DEV_MOST_USED_FIRST ||
                cuPropV2->policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_DEV_LEAST_USED_FIRST)
                return (resAllocCuByDevLoad(cuPropV2, cuRes, updateId));
            if (cuPropV2->policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_CU_MOST_USED_FIRST ||
                cuPropV2->policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_CU_LEAST_USED_FIRST)
                return (resAllocCuByCuLoad(cuPropV2, cuRes, updateId));

            return (resAllocCuWithoutPolicy(cuPropV2, cuRes, updateId));
        }
        case XRM_DEVICE_INFO_CONSTRAINT_TYPE_HARDWARE_DEVICE_INDEX: {
            return (resAllocCuFromDevByCuLoad((uint32_t)deviceInfoDeviceIndex, cuPropV2, cuRes, updateId));
        }
        default: {
            logMsg(XRM_LOG_ERROR, "invalid device info constraint type\n");
            return (XRM_ERROR_INVALID);
        }
    }
}

/*
 * Alloc one cu (compute unit) from target device based on the request property.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocCuFromDev(int32_t deviceId, cuProperty* cuProp, cuResource* cuRes, bool updateId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuProp->clientId;
    int32_t cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    if ((deviceId < 0) || (deviceId > (m_numDevice - 1))) {
        logMsg(XRM_LOG_ERROR, "deviceId (%d) is out of range\n", deviceId);
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp == NULL) || (cuRes == NULL)) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0') && (cuProp->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (XRM_ERROR_INVALID);
    }

    dev = &m_devList[deviceId];
    if (!dev->isLoaded) {
        return (XRM_ERROR_NO_DEV);
    }
    if ((dev->isExcl) && (dev->clientProcs[0].clientId != clientId)) {
        /* the device is used by other client exclusively */
        return (XRM_ERROR_NO_DEV);
    }
    ret = allocClientFromDev(deviceId, cuProp);
    if (ret < 0) {
        return (XRM_ERROR_NO_DEV);
    }

cu_alloc_from_dev_loop:
    /*
     * Now check whether matching cu is on allocated device
     * and try to allocate requested cu.
     */
    for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];

        /* first attempt to re-use existing kernels; else, use a new kernel */
        if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
        /* alloc channel and register client id */
        ret = allocChanClientFromCu(cu, cuProp, cuRes);
        if (ret != XRM_SUCCESS) {
            continue;
        }

        cuRes->deviceId = deviceId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_alloc_from_dev_loop;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    } else {
        releaseClientOnDev(deviceId, clientId);
        return (XRM_ERROR_NO_KERNEL);
    }
}

/*
 * Alloc one cu (compute unit) from target device based on the request property.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocLeastUsedCuFromDev(int32_t deviceId,
                                                cuProperty* cuProp,
                                                cuResource* cuRes,
                                                bool updateId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuProp->clientId;
    int32_t cuId;
    std::vector<std::tuple<int32_t, int32_t> > leastUsedCus;
    bool cuAcquired = false;

    if ((deviceId < 0) || (deviceId > (m_numDevice - 1))) {
        logMsg(XRM_LOG_ERROR, "deviceId (%d) is out of range\n", deviceId);
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp == NULL) || (cuRes == NULL)) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0') && (cuProp->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (XRM_ERROR_INVALID);
    }

    dev = &m_devList[deviceId];
    if (!dev->isLoaded) {
        return (XRM_ERROR_NO_DEV);
    }
    if ((dev->isExcl) && (dev->clientProcs[0].clientId != clientId)) {
        /* the device is used by other client exclusively */
        return (XRM_ERROR_NO_DEV);
    }
    ret = allocClientFromDev(deviceId, cuProp);
    if (ret < 0) {
        return (XRM_ERROR_NO_DEV);
    }

    /*
     * Now check whether matching cu is on allocated device
     * and try to allocate requested cu.
     */
    for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];

        /* first attempt to use a new kernel */
        if (cu->numClient > 0) continue;
        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
        /* alloc channel and register client id */
        ret = allocChanClientFromCu(cu, cuProp, cuRes);
        if (ret != XRM_SUCCESS) {
            continue;
        }

        cuRes->deviceId = deviceId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    }

    if (!cuAcquired) {
        goto cu_alloc_from_dev_loop;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    } else {
        releaseClientOnDev(deviceId, clientId);
        return (XRM_ERROR_NO_KERNEL);
    }

cu_alloc_from_dev_loop:
    for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];

        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;

        int64_t resultLoad = cuProp->requestLoadUnified + cu->totalUsedLoadUnified;
        if (resultLoad <= XRM_MAX_CHAN_LOAD_GRANULARITY_1000000) {
            // We've found a potential CU, lets note down its load, and cuId
            leastUsedCus.emplace_back(cu->totalUsedLoadUnified, cuId);
        }
    }

    std::sort(leastUsedCus.begin(), leastUsedCus.end());

    ret = XRM_ERROR_NO_KERNEL; // If no candidates, we fail

    for (const auto& candidate : leastUsedCus) {
        cuId = std::get<1>(candidate);

        cu = &dev->xclbinInfo.cuList[cuId];

        ret = allocChanClientFromCu(cu, cuProp, cuRes);
        if (ret != XRM_SUCCESS) {
            releaseClientOnDev(deviceId, clientId);
            continue;
        }
        cuRes->deviceId = deviceId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
        if (updateId) updateAllocServiceId();

        return (XRM_SUCCESS);
    }
    return (XRM_ERROR_NO_KERNEL);
}
/*
 * Allocate one cu (compute unit) based on the request property. if fail to allocate cu, then try
 * to load the xclbin to one device and allocate cu from this device.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu acquiration
 */
int32_t xrm::system::resAllocCuWithLoad(cuProperty* cuProp, std::string xclbin, cuResource* cuRes, bool updateId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuProp->clientId;
    int32_t devId, cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    if ((cuProp == NULL) || (cuRes == NULL)) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0') && (cuProp->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (XRM_ERROR_INVALID);
    }

    /*
     * try to allocate the cu from existing resource pool
     */
cu_acquire_loop:
    for (devId = -1; !cuAcquired && (devId < m_numDevice);) {
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            break;
        }

        ret = allocClientFromDev(devId, cuProp);
        if (ret < 0) {
            continue;
        }
        dev = &m_devList[devId];
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to re-use existing kernels; else, use a new kernel */
            if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            /* cu name is presented, compare it; otherwise no need to compare */
            cuNameEqual = true;
            if (cuProp->cuName[0] != '\0') {
                if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
            /* alloc channel and register client id */
            ret = allocChanClientFromCu(cu, cuProp, cuRes);
            if (ret != XRM_SUCCESS) {
                continue;
            }

            cuRes->deviceId = devId;
            cuRes->deviceNumBDF = dev->devNumBDF;
            cuRes->cuId = cuId;
            strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
            strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
            cuAcquired = true;
        }

        if (!cuAcquired) {
            releaseClientOnDev(devId, clientId);
        }
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_acquire_loop;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    }

    /*
     * try to load the xclbin to one possible device
     */
    devId = -1;
    std::string loadErrmsg;
    ret = loadAnyDevice(&devId, xclbin, loadErrmsg);
    if (ret != XRM_SUCCESS) {
        logMsg(XRM_LOG_NOTICE, "%s fail to load xclbin to any device\n", __func__);
        goto cu_acquire_exit;
    }

    /*
     * try to allocate the cu from the device loaded with xclbin file
     */
    ret = allocClientFromDev(devId, cuProp);
    if (ret < 0) {
        goto cu_acquire_exit;
    }
    dev = &m_devList[devId];
    /* Now check whether matching cu is on allocated device, if not, free device. */
    for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];

        /* first attempt to re-use existing kernels; else, use a new kernel */
        if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
        /* alloc channel and register client id */
        ret = allocChanClientFromCu(cu, cuProp, cuRes);
        if (ret != XRM_SUCCESS) {
            continue;
        }

        cuRes->deviceId = devId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        ret = XRM_SUCCESS;
    } else {
        releaseClientOnDev(devId, clientId);
        ret = XRM_ERROR_NO_KERNEL;
    }

cu_acquire_exit:
    return (ret);
}

/*
 * Allocate the least used cu (compute unit) based on the request property.
 * if an unused cu cannot be allocated, then try
 * to load the xclbin to one device and allocate cu from that device.
 * if there are no available devices, try to share the least used cu.
 * Additionally, it strictly enforces that the allocated CU comes
 * from a device that was loaded with the user specified xclbinFileName.
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu acquiration
 */
int32_t xrm::system::resAllocCuLeastUsedWithLoad(cuProperty* cuProp,
                                                 std::string xclbin,
                                                 cuResource* cuRes,
                                                 bool updateId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuProp->clientId;
    int32_t devId, cuId;
    bool cuAcquired = false;

    // Vector of tuples (cuData::totalUsedLoadUnified, devId, cuId)
    std::vector<std::tuple<int32_t, int32_t, int32_t> > leastUsedCus;

    if ((cuProp == NULL) || (cuRes == NULL)) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0') && (cuProp->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (XRM_ERROR_INVALID);
    }

    /*
     * try to allocate the cu from existing resource pool
     */
    for (devId = -1; !cuAcquired && (devId < m_numDevice);) {
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            break;
        }

        // If user's requested xclbin does not match, keep searching
        if (m_devList[devId].xclbinName != xclbin) continue;

        ret = allocClientFromDev(devId, cuProp);
        if (ret < 0) {
            continue;
        }
        dev = &m_devList[devId];
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to use a new kernel */
            if (cu->numClient > 0) continue;
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            /* cu name is presented, compare it; otherwise no need to compare */
            cuNameEqual = true;
            if (cuProp->cuName[0] != '\0') {
                if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
            /* alloc channel and register client id */
            ret = allocChanClientFromCu(cu, cuProp, cuRes);
            if (ret != XRM_SUCCESS) {
                continue;
            }

            cuRes->deviceId = devId;
            cuRes->deviceNumBDF = dev->devNumBDF;
            cuRes->cuId = cuId;
            strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
            strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
            cuAcquired = true;
        }

        if (!cuAcquired) {
            releaseClientOnDev(devId, clientId);
        }
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    }

    /*
     * try to load the xclbin to one possible device
     */
    devId = -1;
    std::string loadErrmsg;
    ret = loadAnyDevice(&devId, xclbin, loadErrmsg);
    if (ret != XRM_SUCCESS) {
        logMsg(XRM_LOG_NOTICE, "%s fail to load xclbin to any device\n", __func__);
        goto cu_acquire_least_used_loop;
    }

    /*
     * try to allocate the cu from the device loaded with xclbin file
     */
    ret = allocClientFromDev(devId, cuProp);
    if (ret < 0) {
        goto cu_acquire_least_used_loop;
    }
    dev = &m_devList[devId];
    /* Now check whether matching cu is on allocated device, if not, free device. */
    for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];

        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
        /* alloc channel and register client id */
        ret = allocChanClientFromCu(cu, cuProp, cuRes);
        if (ret != XRM_SUCCESS) {
            continue;
        }

        cuRes->deviceId = devId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        ret = XRM_SUCCESS;
        goto cu_acquire_exit;
    } else {
        releaseClientOnDev(devId, clientId);
        ret = XRM_ERROR_NO_KERNEL;
    }

cu_acquire_least_used_loop:

    // Search all possible devices
    for (devId = -1; devId < m_numDevice;) {
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            break;
        }

        // If user's requested xclbin does not match, keep searching
        if (m_devList[devId].xclbinName != xclbin) continue;

        ret = allocClientFromDev(devId, cuProp);
        if (ret < 0) {
            continue;
        }
        dev = &m_devList[devId];
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        // Search all possible cus
        for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            /* cu name is presented, compare it; otherwise no need to compare */
            cuNameEqual = true;
            if (cuProp->cuName[0] != '\0') {
                if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;

            int64_t resultLoad = cuProp->requestLoadUnified + cu->totalUsedLoadUnified;
            if (resultLoad <= XRM_MAX_CHAN_LOAD_GRANULARITY_1000000) {
                // We've found a potential CU, lets note down its load, devId, and cuId
                leastUsedCus.emplace_back(cu->totalUsedLoadUnified, devId, cuId);
            }
        }

        // Release this device
        releaseClientOnDev(devId, clientId);
    }

    std::sort(leastUsedCus.begin(), leastUsedCus.end());

    ret = XRM_ERROR_NO_KERNEL; // If no candidates, we fail

    for (const auto& candidate : leastUsedCus) {
        devId = std::get<1>(candidate);
        cuId = std::get<2>(candidate);

        ret = allocClientFromDev(devId, cuProp);
        if (ret < 0) {
            continue;
        }

        dev = &m_devList[devId];
        cu = &dev->xclbinInfo.cuList[cuId];

        ret = allocChanClientFromCu(cu, cuProp, cuRes);
        if (ret != XRM_SUCCESS) {
            releaseClientOnDev(devId, clientId);
            continue;
        }

        cuRes->deviceId = devId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    }

cu_acquire_exit:
    return (ret);
}

/*
 * Alloc one cu list based on the request property.
 *
 * XRM_SUCCESS: allocated resouce is filled into cuListRes
 * Otherwise: failed to allocate the cu list
 *
 * Lock: should enter lock during the cu list allocation
 */
int32_t xrm::system::resAllocCuList(cuListProperty* cuListProp, cuListResource* cuListRes) {
    int32_t i, devId, ret;
    cuProperty* cuProp = NULL;
    cuResource* cuRes = NULL;

    if (cuListProp == NULL || cuListRes == NULL) {
        return (XRM_ERROR_INVALID);
    }
    if (cuListProp->cuNum <= 0 || cuListProp->cuNum > XRM_MAX_LIST_CU_NUM) {
        return (XRM_ERROR_INVALID);
    }

    cuListRes->cuNum = 0;
    ret = XRM_ERROR;
    if (cuListProp->sameDevice) {
        /*
         * To alloc the cu list from same device
         */
        for (devId = -1; devId < m_numDevice;) {
            /* find one available device for first cu*/
            cuProp = &cuListProp->cuProps[0];
            devId = allocDevForClient(&devId, cuProp);
            if (devId < 0) {
                return (XRM_ERROR_NO_KERNEL);
            }
            /* alloc the cu list from this device */
            for (i = 0; i < cuListProp->cuNum; i++) {
                cuProp = &cuListProp->cuProps[i];
                cuRes = &cuListRes->cuResources[i];
                ret = allocClientFromDev(devId, cuProp);
                if (ret != XRM_SUCCESS) {
                    resReleaseCuList(cuListRes);
                    memset(cuListRes, 0, sizeof(cuListResource));
                    break;
                }
                ret = allocCuFromDev(devId, cuProp, cuRes);
                if (ret != XRM_SUCCESS) {
                    releaseClientOnDev(devId, cuProp->clientId);
                    resReleaseCuList(cuListRes);
                    memset(cuListRes, 0, sizeof(cuListResource));
                    break;
                }
                cuListRes->cuNum++;
            }
            if (ret == XRM_SUCCESS) {
                updateAllocServiceId();
                return (XRM_SUCCESS);
            }
        }
        if (ret == XRM_SUCCESS)
            return (XRM_SUCCESS);
        else
            return (XRM_ERROR_NO_KERNEL);
    } else {
        /*
         * To alloc the cu list from any device
         */
        for (i = 0; i < cuListProp->cuNum; i++) {
            cuProp = &cuListProp->cuProps[i];
            cuRes = &cuListRes->cuResources[i];
            bool updateId = false;
            ret = resAllocCu(cuProp, cuRes, updateId);
            if (ret != XRM_SUCCESS) {
                resReleaseCuList(cuListRes);
                memset(cuListRes, 0, sizeof(cuListResource));
                return (XRM_ERROR_NO_KERNEL);
            }
            cuListRes->cuNum++;
        }
        updateAllocServiceId();
        return (XRM_SUCCESS);
    }
}

/*
 * copy cu property from src prop (V2) to des prop.
 */
void xrm::system::cuPropertyCopyFromV2(cuProperty* desCuProp, cuPropertyV2* srcCuPropV2) {
    if (desCuProp != NULL && srcCuPropV2 != NULL) {
        strncpy(desCuProp->kernelName, srcCuPropV2->kernelName, XRM_MAX_NAME_LEN - 1);
        desCuProp->kernelName[XRM_MAX_NAME_LEN - 1] = '\0';
        strncpy(desCuProp->kernelAlias, srcCuPropV2->kernelAlias, XRM_MAX_NAME_LEN - 1);
        desCuProp->kernelAlias[XRM_MAX_NAME_LEN - 1] = '\0';
        strcpy(desCuProp->cuName, "");
        desCuProp->devExcl = srcCuPropV2->devExcl;
        desCuProp->requestLoadUnified = srcCuPropV2->requestLoadUnified;
        desCuProp->requestLoadOriginal = srcCuPropV2->requestLoadOriginal;
        desCuProp->clientId = srcCuPropV2->clientId;
        desCuProp->clientProcessId = srcCuPropV2->clientProcessId;
        desCuProp->poolId = srcCuPropV2->poolId;
    }
}

/*
 * copy cu resource from src to des.
 */
void xrm::system::cuResourceCopy(cuResource* desCuRes, cuResource* srcCuRes) {
    if (desCuRes != NULL && srcCuRes != NULL) {
        strncpy(desCuRes->xclbinFileName, srcCuRes->xclbinFileName, XRM_MAX_NAME_LEN - 1);
        desCuRes->xclbinFileName[XRM_MAX_NAME_LEN - 1] = '\0';
        strncpy(desCuRes->kernelPluginFileName, srcCuRes->kernelPluginFileName, XRM_MAX_NAME_LEN - 1);
        desCuRes->kernelPluginFileName[XRM_MAX_NAME_LEN - 1] = '\0';
        strncpy(desCuRes->kernelName, srcCuRes->kernelName, XRM_MAX_NAME_LEN - 1);
        desCuRes->kernelName[XRM_MAX_NAME_LEN - 1] = '\0';
        strncpy(desCuRes->kernelAlias, srcCuRes->kernelAlias, XRM_MAX_NAME_LEN - 1);
        desCuRes->kernelAlias[XRM_MAX_NAME_LEN - 1] = '\0';
        strncpy(desCuRes->instanceName, srcCuRes->instanceName, XRM_MAX_NAME_LEN - 1);
        desCuRes->instanceName[XRM_MAX_NAME_LEN - 1] = '\0';
        strncpy(desCuRes->cuName, srcCuRes->cuName, XRM_MAX_NAME_LEN - 1);
        desCuRes->cuName[XRM_MAX_NAME_LEN - 1] = '\0';
        strncpy(desCuRes->uuidStr, srcCuRes->uuidStr, XRM_MAX_NAME_LEN - 1);
        desCuRes->uuidStr[XRM_MAX_NAME_LEN - 1] = '\0';
        desCuRes->deviceId = srcCuRes->deviceId;
        desCuRes->deviceNumBDF = srcCuRes->deviceNumBDF;
        desCuRes->cuId = srcCuRes->cuId;
        desCuRes->channelId = srcCuRes->channelId;
        desCuRes->cuType = srcCuRes->cuType;
        desCuRes->baseAddr = srcCuRes->baseAddr;
        desCuRes->membankId = srcCuRes->membankId;
        desCuRes->membankType = srcCuRes->membankType;
        desCuRes->membankSize = srcCuRes->membankSize;
        desCuRes->membankBaseAddr = srcCuRes->membankBaseAddr;
        desCuRes->allocServiceId = srcCuRes->allocServiceId;
        desCuRes->clientId = srcCuRes->clientId;
        desCuRes->channelLoadUnified = srcCuRes->channelLoadUnified;
        desCuRes->channelLoadOriginal = srcCuRes->channelLoadOriginal;
        desCuRes->logicGroupInfo.indicator = srcCuRes->logicGroupInfo.indicator;
        desCuRes->logicGroupInfo.groupId = srcCuRes->logicGroupInfo.groupId;
        strncpy(desCuRes->logicGroupInfo.groupType, srcCuRes->logicGroupInfo.groupType,
            XRM_MAX_LOGIC_GROUP_TYPE_LEN -1);
        desCuRes->logicGroupInfo.groupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1] = '\0';
        desCuRes->poolId = srcCuRes->poolId;
    }
}

/* check whether devId in device id list */
bool xrm::system::isInDeviceIdList(deviceIdListPropertyV2* devIdList, int32_t devId) {
    bool ret = false;
    if (devIdList->deviceNum < 0 || devIdList->deviceNum > XRM_MAX_XILINX_DEVICES) return (ret);
    if (devId < 0 || devId > (XRM_MAX_XILINX_DEVICES - 1)) return (ret);
    for (int32_t i = 0; i < devIdList->deviceNum; i++) {
        if (devIdList->deviceIds[i] == (uint64_t)devId) {
            ret = true;
            break;
        }
    }
    return (ret);
}

/* insert one devId into device id list, device id in list is unique */
uint32_t xrm::system::insertDeviceIdList(deviceIdListPropertyV2* devIdList, int32_t devId) {
    uint32_t ret = XRM_ERROR;
    if (devIdList->deviceNum < 0 || devIdList->deviceNum > XRM_MAX_XILINX_DEVICES) return (ret);
    if (devId < 0 || devId > (XRM_MAX_XILINX_DEVICES - 1)) return (ret);
    for (int32_t i = 0; i < devIdList->deviceNum; i++) {
        if (devIdList->deviceIds[i] == (uint64_t)devId) {
            ret = XRM_SUCCESS;
            return (ret);
        }
    }
    if (devIdList->deviceNum < XRM_MAX_XILINX_DEVICES) {
        devIdList->deviceIds[devIdList->deviceNum] = (uint64_t)devId;
        devIdList->deviceNum++;
        ret = XRM_SUCCESS;
    } else {
        ret = XRM_ERROR;
    }
    return (ret);
}

/*
 * To handle the device constraint information, CU list will be put into differernt
 *    sub-list:
 *    a) one or multiple sub-lists with same hw device index, different index into
 *       different sub-list. CU in such sub-list will be allocated from target device later.
 *    b) one or multiple sub-lists with same virtual device index, different index
 *       into different sub-list. cu in same sub list will be allocated from same device.
 *       different sub-list will be allocated from different device later.
 *    c) one sub-list for cu without device constraint. cu in this sub-list will be
 *       allocated from any device later.
 *
 * This function is to put cu into the corresponding sub-list.
 */
int32_t xrm::system::putCuIntoDeviceConstraintSubLists(cuPropertyV2* cuPropV2, int32_t index,
    int32_t* numSubList, deviceConstraintSubListTypeV2* deviceConstraintSubLists) {
    int32_t ret = XRM_SUCCESS;
    uint64_t deviceInfoConstraintType;
    uint64_t deviceInfoDeviceIndex;
    int32_t subListIndex, numExistingSubList, numCu;
    bool subListExisting = false;

    deviceInfoConstraintType =
        (((cuPropV2->deviceInfo) >> XRM_DEVICE_INFO_CONSTRAINT_TYPE_SHIFT) & XRM_DEVICE_INFO_CONSTRAINT_TYPE_MASK);
    deviceInfoDeviceIndex =
        (((cuPropV2->deviceInfo) >> XRM_DEVICE_INFO_DEVICE_INDEX_SHIFT) & XRM_DEVICE_INFO_DEVICE_INDEX_MASK);
    numExistingSubList = *numSubList;
    switch (deviceInfoConstraintType) {
        case XRM_DEVICE_INFO_CONSTRAINT_TYPE_NULL:
            for (subListIndex = 0; !subListExisting && (subListIndex < numExistingSubList); subListIndex++) {
                if (deviceConstraintSubLists[subListIndex].type == DEV_CONSTRAINT_SUBLIST_TYPE_NULL) {
                    /* find the subList: match type */
                    numCu = deviceConstraintSubLists[subListIndex].numCu;
                    deviceConstraintSubLists[subListIndex].cuId[numCu] = index;
                    deviceConstraintSubLists[subListIndex].numCu++;
                    subListExisting = true;
                }
            }
            if (!subListExisting) {
                /* such subList is not existing, set next list as new type of it */
                deviceConstraintSubLists[numExistingSubList].type = DEV_CONSTRAINT_SUBLIST_TYPE_NULL;
                deviceConstraintSubLists[numExistingSubList].numCu = 1;
                deviceConstraintSubLists[numExistingSubList].cuId[0] = index;
                (*numSubList)++;
            }
            break;
        case XRM_DEVICE_INFO_CONSTRAINT_TYPE_HARDWARE_DEVICE_INDEX:
            for (subListIndex = 0; !subListExisting && (subListIndex < numExistingSubList); subListIndex++) {
                if ((deviceConstraintSubLists[subListIndex].type == DEV_CONSTRAINT_SUBLIST_TYPE_HARDWARE_DEVICE_ID) &&
                    (deviceConstraintSubLists[subListIndex].deviceId == (int32_t)deviceInfoDeviceIndex)) {
                    /* find the subList: match type and device id */
                    numCu = deviceConstraintSubLists[subListIndex].numCu;
                    deviceConstraintSubLists[subListIndex].cuId[numCu] = index;
                    deviceConstraintSubLists[subListIndex].numCu++;
                    subListExisting = true;
                }
            }
            if (!subListExisting) {
                /* such subList is not existing, set next list as new type of it */
                deviceConstraintSubLists[numExistingSubList].type = DEV_CONSTRAINT_SUBLIST_TYPE_HARDWARE_DEVICE_ID;
                deviceConstraintSubLists[numExistingSubList].deviceId = (int32_t)deviceInfoDeviceIndex;
                deviceConstraintSubLists[numExistingSubList].numCu = 1;
                deviceConstraintSubLists[numExistingSubList].cuId[0] = index;
                (*numSubList)++;
            }
            break;
        case XRM_DEVICE_INFO_CONSTRAINT_TYPE_VIRTUAL_DEVICE_INDEX:
            for (subListIndex = 0; !subListExisting && (subListIndex < numExistingSubList); subListIndex++) {
                if ((deviceConstraintSubLists[subListIndex].type == DEV_CONSTRAINT_SUBLIST_TYPE_VIRTUAL_DEVICE_ID) &&
                    (deviceConstraintSubLists[subListIndex].deviceId == (int32_t)deviceInfoDeviceIndex)) {
                    /* find the subList: match type and device id */
                    numCu = deviceConstraintSubLists[subListIndex].numCu;
                    deviceConstraintSubLists[subListIndex].cuId[numCu] = index;
                    deviceConstraintSubLists[subListIndex].numCu++;
                    subListExisting = true;
                }
            }
            if (!subListExisting) {
                /* such subList is not existing, set next list as new type of it */
                deviceConstraintSubLists[numExistingSubList].type = DEV_CONSTRAINT_SUBLIST_TYPE_VIRTUAL_DEVICE_ID;
                deviceConstraintSubLists[numExistingSubList].deviceId = (int32_t)deviceInfoDeviceIndex;
                deviceConstraintSubLists[numExistingSubList].numCu = 1;
                deviceConstraintSubLists[numExistingSubList].cuId[0] = index;
                (*numSubList)++;
            }
            break;
        default:
            logMsg(XRM_LOG_ERROR, "invalid device info [%d] constraint type\n", index);
            ret = XRM_ERROR_INVALID;
            break;
    }
    return (ret);
}

/*
 * Device constraint based sub-lists will be further put into logic group
 *    constraint based small sub-lists:
 *    a) one or multiple sub-list with same real group id.
 *    b) one or multiple sub-list with same group type.
 *    c) one or multiple sub-list with same pseudo group id, different list
 *       from different logic group.
 *    d) one or multiple sub-list with same pseudo group id and type, different list
 *       from different logic group.
 *    e) one sub-list for cu without logic group constraint.
 *
 * This function is to put cu into the corresponding logic group base sub-list.
 */
int32_t xrm::system::putCuIntoLgConstraintSubLists(cuPropertyV2* cuPropV2, int32_t index,
    int32_t* numSubList, logicGroupConstraintSubListTypeV2* lgConstraintSubLists) {
    int32_t ret = XRM_SUCCESS;
    int32_t groupInfoConstraintType;
    int32_t groupInfoGroupId;
    char groupInfoGroupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN];
    int32_t subListIndex, numExistingSubList, numCu;
    bool subListExisting = false;

    groupInfoConstraintType =
        (((cuPropV2->logicGroupInfo.constraintInfo) >> XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_SHIFT) & XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_MASK);
    groupInfoGroupId = cuPropV2->logicGroupInfo.groupId;
    strncpy(groupInfoGroupType, cuPropV2->logicGroupInfo.groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1);
    groupInfoGroupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1] = '\0';

    numExistingSubList = *numSubList;
    switch (groupInfoConstraintType) {
        case XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_NULL:
            for (subListIndex = 0; !subListExisting && (subListIndex < numExistingSubList); subListIndex++) {
                if (lgConstraintSubLists[subListIndex].type == DEV_CONSTRAINT_SUBLIST_TYPE_NULL) {
                    /* find the subList: match type */
                    numCu = lgConstraintSubLists[subListIndex].numCu;
                    lgConstraintSubLists[subListIndex].cuId[numCu] = index;
                    lgConstraintSubLists[subListIndex].numCu++;
                    subListExisting = true;
                }
            }
            if (!subListExisting) {
                /* such subList is not existing, set next list as new type of it */
                lgConstraintSubLists[numExistingSubList].type = DEV_CONSTRAINT_SUBLIST_TYPE_NULL;
                lgConstraintSubLists[numExistingSubList].numCu = 1;
                lgConstraintSubLists[numExistingSubList].cuId[0] = index;
                (*numSubList)++;
            }
            break;
        case XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID:
            for (subListIndex = 0; !subListExisting && (subListIndex < numExistingSubList); subListIndex++) {
                if ((lgConstraintSubLists[subListIndex].type == LG_CONSTRAINT_SUBLIST_TYPE_REAL_GROUP_ID) &&
                    (lgConstraintSubLists[subListIndex].groupId == groupInfoGroupId)) {
                    /* find the subList: match type and device id */
                    numCu = lgConstraintSubLists[subListIndex].numCu;
                    lgConstraintSubLists[subListIndex].cuId[numCu] = index;
                    lgConstraintSubLists[subListIndex].numCu++;
                    subListExisting = true;
                }
            }
            if (!subListExisting) {
                /* such subList is not existing, set next list as new type of it */
                lgConstraintSubLists[numExistingSubList].type = LG_CONSTRAINT_SUBLIST_TYPE_REAL_GROUP_ID;
                lgConstraintSubLists[numExistingSubList].groupId = groupInfoGroupId;
                lgConstraintSubLists[numExistingSubList].numCu = 1;
                lgConstraintSubLists[numExistingSubList].cuId[0] = index;
                (*numSubList)++;
            }
            break;
        case XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE:
            for (subListIndex = 0; !subListExisting && (subListIndex < numExistingSubList); subListIndex++) {
                if ((lgConstraintSubLists[subListIndex].type == LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE) &&
                    (strncmp(lgConstraintSubLists[subListIndex].groupType,
                    groupInfoGroupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) == 0 )) {
                    /* find the subList: match type and device id */
                    numCu = lgConstraintSubLists[subListIndex].numCu;
                    lgConstraintSubLists[subListIndex].cuId[numCu] = index;
                    lgConstraintSubLists[subListIndex].numCu++;
                    subListExisting = true;
                }
            }
            if (!subListExisting) {
                /* such subList is not existing, set next list as new type of it */
                lgConstraintSubLists[numExistingSubList].type = LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE;
                strncpy(lgConstraintSubLists[numExistingSubList].groupType,
                    groupInfoGroupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1);
                lgConstraintSubLists[numExistingSubList].groupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1] = '\0';
                lgConstraintSubLists[numExistingSubList].numCu = 1;
                lgConstraintSubLists[numExistingSubList].cuId[0] = index;
                (*numSubList)++;
            }
            break;
        case XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_PSEUDO_LOGIC_GROUP_ID:
            for (subListIndex = 0; !subListExisting && (subListIndex < numExistingSubList); subListIndex++) {
                if ((lgConstraintSubLists[subListIndex].type == LG_CONSTRAINT_SUBLIST_TYPE_PSEUDO_GROUP_ID) &&
                    (lgConstraintSubLists[subListIndex].groupId == groupInfoGroupId)) {
                    /* find the subList: match type and device id */
                    numCu = lgConstraintSubLists[subListIndex].numCu;
                    lgConstraintSubLists[subListIndex].cuId[numCu] = index;
                    lgConstraintSubLists[subListIndex].numCu++;
                    subListExisting = true;
                }
            }
            if (!subListExisting) {
                /* such subList is not existing, set next list as new type of it */
                lgConstraintSubLists[numExistingSubList].type = LG_CONSTRAINT_SUBLIST_TYPE_PSEUDO_GROUP_ID;
                lgConstraintSubLists[numExistingSubList].groupId = groupInfoGroupId;
                lgConstraintSubLists[numExistingSubList].numCu = 1;
                lgConstraintSubLists[numExistingSubList].cuId[0] = index;
                (*numSubList)++;
            }
            break;
        case XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE_AND_PSEUDO_LOGIC_GROUP_ID:
            for (subListIndex = 0; !subListExisting && (subListIndex < numExistingSubList); subListIndex++) {
                if ((lgConstraintSubLists[subListIndex].type == LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE_AND_PSEUDO_GROUP_ID) &&
                    (lgConstraintSubLists[subListIndex].groupId == groupInfoGroupId)) {
                    if (strncmp(lgConstraintSubLists[subListIndex].groupType,
                        groupInfoGroupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) != 0 ) {
                        /*
                         * Wrong constraint logic group information:
                         * pseudo group is matching, but group type is NOT matching
                         */
                        logMsg(XRM_LOG_ERROR, "invalid group info [%d], same pseudo group id, group type not same\n", index);
                        return (XRM_ERROR_INVALID);
                    }
                    /* find the subList: match type and group id, group type */
                    numCu = lgConstraintSubLists[subListIndex].numCu;
                    lgConstraintSubLists[subListIndex].cuId[numCu] = index;
                    lgConstraintSubLists[subListIndex].numCu++;
                    subListExisting = true;

                }
            }
            if (!subListExisting) {
                /* such subList is not existing, set next list as new type of it */
                lgConstraintSubLists[numExistingSubList].type = LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE_AND_PSEUDO_GROUP_ID;
                strncpy(lgConstraintSubLists[numExistingSubList].groupType,
                    groupInfoGroupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1);
                lgConstraintSubLists[numExistingSubList].groupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1] = '\0';
                lgConstraintSubLists[numExistingSubList].groupId = groupInfoGroupId;
                lgConstraintSubLists[numExistingSubList].numCu = 1;
                lgConstraintSubLists[numExistingSubList].cuId[0] = index;
                (*numSubList)++;
            }
            break;
        default:
            logMsg(XRM_LOG_ERROR, "invalid group info [%d] constraint type\n", index);
            ret = XRM_ERROR_INVALID;
            break;
    }
    return (ret);
}

/*
 * Alloc one cu (compute unit) from target device's target logic group.
 * It's used for cu list allocation, no allocate policy applied.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes, including logic group infor.
 * Otherwise: failed to allocate the cu resource
 *
 * logic group infor in returned resource is valid.
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocCuFromTargetDeviceTargetLogicGroup(int32_t devId, int32_t lgId,
    cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuPropV2->clientId;
    int32_t cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;

    // no need to check cuPropV2 or cuRes as it should be checked already
    cuPropertyCopyFromV2(cuProp, cuPropV2);

cu_alloc_loop:
    ret = allocClientFromDev(devId, cuProp);
    if (ret < 0) {
        return (ret);
    }
    dev = &m_devList[devId];
    /*
     * Now check whether matching cu is in target device target logic group
     */
    if (lgId >= dev->xclbinInfo.numLogicGroup) {
        return (XRM_ERROR);
    }
    logicGroupData* logicGroup = &(dev->xclbinInfo.logicGroups[lgId]);
    for (int32_t lgCuIndex = 0; lgCuIndex < logicGroup->numCu && !cuAcquired; lgCuIndex++) {
        cuId = logicGroup->cuId[lgCuIndex];
        cu = &dev->xclbinInfo.cuList[cuId];

        /* first attempt to re-use existing kernels; else, use a new kernel */
        if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
        /* set the logic group infor for resource alloction */
        cuResourceLogicGroupInfo lgInfo;
        lgInfo.indicator = XRM_LOGIC_GROUP_INFO_ID_TYPE_VALID;
        lgInfo.groupId = lgId;
        strncpy(lgInfo.groupType, logicGroup->groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1);
        lgInfo.groupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1] = '\0';
        /* alloc channel and register client id */
        // logic group information will be set in allocChanClientFromCu call based on the input.
        ret = allocChanClientFromCu(cu, cuProp, cuRes, &lgInfo);
        if (ret != XRM_SUCCESS) {
            continue;
        }

        /*
         * XRM_SUCCESS, channel allocation is done. To set allocated resource information.
         */
        cuRes->deviceId = devId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    } // cu in the logic group

    if (!cuAcquired) {
        releaseClientOnDev(devId, clientId);
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_alloc_loop;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Allocate cu sub list from:
 *          Target device
 *          Target logic group (same group)
 */
int32_t xrm::system::resAllocCuSubListFromTargetDeviceTargetLogicGroup(int32_t devId,
    int32_t lgId, logicGroupConstraintSubListTypeV2* lgSubList,
    cuListPropertyV2* cuListPropV2, cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;
    cuResource* cuRes = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        cuRes = &cuListResV2->cuResources[cuId];
        ret = resAllocCuFromTargetDeviceTargetLogicGroup(devId, lgId, cuPropV2, cuRes, false);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /* fail to allocate cu of this sub list, release the allocated resource for this sub-list */
        resReleaseCuLgSubListV2(lgSubList, cuListResV2, itemFlags);
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Any device
 *          Target logic group
 * So the cu in this sub list can from group on differernt device.
 */
int32_t xrm::system::resAllocCuSubListFromAnyDeviceTargetLogicGroup(
    int32_t lgId, logicGroupConstraintSubListTypeV2* lgSubList,
    cuListPropertyV2* cuListPropV2, cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;
    cuResource* cuRes = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        cuRes = &cuListResV2->cuResources[cuId];
        ret = resAllocCuFromAnyDeviceTargetLogicGroup(lgId, cuPropV2, cuRes, false);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /* fail to allocate cu of this sub list, release the allocated resource for this sub-list */
        resReleaseCuLgSubListV2(lgSubList, cuListResV2, itemFlags);
    }
    return (ret);
}

/*
 * Allocate cu from:
 *          Any device
 *          Target logic group
 * So the cu can from group on differernt device.
 */
int32_t xrm::system::resAllocCuFromAnyDeviceTargetLogicGroup(int32_t lgId,
    cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    int32_t ret = XRM_ERROR;
    int32_t devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }

        ret = resAllocCuFromTargetDeviceTargetLogicGroup(devId, lgId, cuPropV2, cuRes, updateId);
        if (ret == XRM_SUCCESS) {
            break;
        }
    }
    return (ret);
}

/*
 * Allocate single cu from:
 *          Target device
 *          Target type logic group
 */
int32_t xrm::system::resAllocCuFromTargetDeviceTargetTypeLogicGroup(int32_t devId,
    cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // find logic group index with target type match
        if (strncmp(cuPropV2->logicGroupInfo.groupType, dev->xclbinInfo.logicGroups[lgId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) == 0 )
            ret = resAllocCuFromTargetDeviceTargetLogicGroup(devId, lgId, cuPropV2, cuRes, updateId);
        if (ret == XRM_SUCCESS)
            break;
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Target device
 *          Target type logic group (can be different group)
 */
int32_t xrm::system::resAllocCuSubListFromTargetDeviceTargetTypeLogicGroup(int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;
    cuResource* cuRes = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        cuRes = &cuListResV2->cuResources[cuId];

        ret = resAllocCuFromTargetDeviceTargetTypeLogicGroup(devId, cuPropV2, cuRes, false);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /* fail to allocate cu of this sub list, release the allocated resource for this sub-list */
        resReleaseCuLgSubListV2(lgSubList, cuListResV2, itemFlags);
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Any device
 *          Target type logic group (can be different group), can also be
 *          group from different device.
 */
int32_t xrm::system::resAllocCuSubListFromAnyDeviceTargetTypeLogicGroup(
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;
    cuResource* cuRes = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        cuRes = &cuListResV2->cuResources[cuId];

        ret = resAllocCuFromAnyDeviceTargetTypeLogicGroup(cuPropV2, cuRes, false);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /* fail to allocate cu of this sub list, release the allocated resource for this sub-list */
        resReleaseCuLgSubListV2(lgSubList, cuListResV2, itemFlags);
    }
    return (ret);
}

/*
 * Allocate cu from:
 *          Any device
 *          Target type logic group
 * So the cu can from group on differernt device.
 */
int32_t xrm::system::resAllocCuFromAnyDeviceTargetTypeLogicGroup(
    cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    int32_t ret = XRM_ERROR;
    int32_t devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }

        ret = resAllocCuFromTargetDeviceTargetTypeLogicGroup(devId, cuPropV2, cuRes, updateId);
        if (ret == XRM_SUCCESS) {
            break;
        }
    }
    return (ret);
}

/* check whether logic group id / device id pair in the list */
bool xrm::system::isInLogicGroupIdDeviceIdPairList(
    logicGroupIdDeviceIdPairListPropertyV2* lgIdPairList, int32_t lgId, int32_t devId) {
    bool ret = false;
    for (int32_t i = 0; i < lgIdPairList->pairNum; i++) {
        if ((lgIdPairList->pairGroupIds[i] == lgId) && (lgIdPairList->pairDevIds[i] == devId)) {
            ret = true;
            break;
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Target device
 *          Same logic group, and the group is different from that in used group list
 */
int32_t xrm::system::resAllocCuSubListFromTargetDeviceSameLogicGroupNotUsed(int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags,
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // find logic group index (with device id pair) not in used list
        if (isInLogicGroupIdDeviceIdPairList(usedLgIdDevIdPairList, lgId, devId))
            // logic group is used, go to next one
            continue;
        ret = resAllocCuSubListFromTargetDeviceTargetLogicGroup(devId, lgId, lgSubList,
            cuListPropV2, cuListResV2, itemFlags);
        if (ret == XRM_SUCCESS) {
            // update new logic group id / device id pair into used list
            usedLgIdDevIdPairList->pairGroupIds[usedLgIdDevIdPairList->pairNum] = lgId;
            usedLgIdDevIdPairList->pairDevIds[usedLgIdDevIdPairList->pairNum] = devId;
            usedLgIdDevIdPairList->pairNum++;
            break;
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Target device
 *          Same logic group, There is no limit the group is different from that in used group list
 */
int32_t xrm::system::resAllocCuSubListFromTargetDeviceSameLogicGroupNoLimitUsedOrNot(int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // Can be any logic group
        ret = resAllocCuSubListFromTargetDeviceTargetLogicGroup(devId, lgId, lgSubList,
            cuListPropV2, cuListResV2, itemFlags);
        if (ret == XRM_SUCCESS) {
            // no need to record the logic group id / device id pair into used list
            break;
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Any device
 *          Same logic group
 *               groupId >= 0: the group is different from that in used group list
 */
int32_t xrm::system::resAllocCuSubListFromAnyDeviceSameLogicGroupNotUsed(
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags,
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList) {
    int32_t ret, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = &(cuListPropV2->cuProps[lgSubList->cuId[0]]);
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    ret = XRM_ERROR;
    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }
        ret = resAllocCuSubListFromTargetDeviceSameLogicGroupNotUsed(devId, lgSubList,
                cuListPropV2, cuListResV2, itemFlags, usedLgIdDevIdPairList);
        if (ret == XRM_SUCCESS) {
            return (ret);
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Any device
 *          Same logic group
 *               groupId < 0 : no limit the group is different from that in used group list
 */
int32_t xrm::system::resAllocCuSubListFromAnyDeviceSameLogicGroupNoLimitUsedOrNot(
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = &(cuListPropV2->cuProps[lgSubList->cuId[0]]);
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    ret = XRM_ERROR;
    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }
        ret = resAllocCuSubListFromTargetDeviceSameLogicGroupNoLimitUsedOrNot(devId, lgSubList,
                cuListPropV2, cuListResV2, itemFlags);
        if (ret == XRM_SUCCESS) {
            return (ret);
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Target device
 *          Target type logic group
 *          Same logic group, and the group is different from that in used group list
 */
int32_t xrm::system::resAllocCuSubListFromTargetDeviceTargetTypeSameLogicGroupNotUsed(int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags,
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // find logic group id / device id pair not in used list
        if (isInLogicGroupIdDeviceIdPairList(usedLgIdDevIdPairList, lgId, devId))
            // logic group is used, go to next one
            continue;
        if (strncmp(lgSubList->groupType,
            dev->xclbinInfo.logicGroups[lgId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) != 0 )
            // type not match, go to next one
            continue;
        ret = resAllocCuSubListFromTargetDeviceTargetLogicGroup(devId, lgId, lgSubList,
            cuListPropV2, cuListResV2, itemFlags);
        if (ret == XRM_SUCCESS) {
            // update new logic group id / device id pair into used list
            usedLgIdDevIdPairList->pairGroupIds[usedLgIdDevIdPairList->pairNum] = lgId;
            usedLgIdDevIdPairList->pairDevIds[usedLgIdDevIdPairList->pairNum] = devId;
            usedLgIdDevIdPairList->pairNum++;
            break;
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Target device
 *          Target type logic group
 *          Same logic group, There is no limit whether the logic group used or not
 */
int32_t xrm::system::resAllocCuSubListFromTargetDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // can be any logic group
        if (strncmp(lgSubList->groupType,
            dev->xclbinInfo.logicGroups[lgId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) != 0 )
            // type not match, go to next one
            continue;
        ret = resAllocCuSubListFromTargetDeviceTargetLogicGroup(devId, lgId, lgSubList,
            cuListPropV2, cuListResV2, itemFlags);
        if (ret == XRM_SUCCESS) {
            // no need to record the logic group in used list
            break;
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Any device
 *          Target type logic grou
 *          Same logic group
 *               groupId >= 0: the group is different from that in used group list
 */
int32_t xrm::system::resAllocCuSubListFromAnyDeviceTargetTypeSameLogicGroupNotUsed(
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags,
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList) {
    int32_t ret, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = &(cuListPropV2->cuProps[lgSubList->cuId[0]]);
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    ret = XRM_ERROR;
    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }
        ret = resAllocCuSubListFromTargetDeviceTargetTypeSameLogicGroupNotUsed(devId, lgSubList,
                cuListPropV2, cuListResV2, itemFlags, usedLgIdDevIdPairList);
        if (ret == XRM_SUCCESS) {
            return (ret);
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Any device
 *          Target type logic grou
 *          Same logic group
 *               groupId < 0: no limit the group is different from that in used group list
 */
int32_t xrm::system::resAllocCuSubListFromAnyDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = &(cuListPropV2->cuProps[lgSubList->cuId[0]]);
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    ret = XRM_ERROR;
    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }
        ret = resAllocCuSubListFromTargetDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(devId, lgSubList,
                cuListPropV2, cuListResV2, itemFlags);
        if (ret == XRM_SUCCESS) {
            return (ret);
        }
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Target device
 *          Any logic group
 * NOTE: since there is no logic group constraint, the logic group infor in
 *       return resource will be set invalid.
 */
int32_t xrm::system::resAllocCuSubListFromTargetDeviceAnyLogicGroup(int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;
    cuResource* cuRes = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        cuRes = &cuListResV2->cuResources[cuId];
        ret = resAllocCuFromTargetDeviceAnyLogicGroup(devId, cuPropV2, cuRes, false);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /* fail to allocate cu of this sub list, release the allocated resource for this sub-list */
        resReleaseCuLgSubListV2(lgSubList, cuListResV2, itemFlags);
    }
    return (ret);
}

/*
 * Allocate cu sub list from:
 *          Any device
 *          Any logic group
 * NOTE: since there is no logic group constraint, the logic group infor in
 *       return resource will be set invalid.
 */
int32_t xrm::system::resAllocCuSubListFromAnyDeviceAnyLogicGroup(
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;
    cuResource* cuRes = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        cuRes = &cuListResV2->cuResources[cuId];
        ret = resAllocCuFromAnyDeviceAnyLogicGroup(cuPropV2, cuRes, false);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /* fail to allocate cu of this sub list, release the allocated resource for this sub-list */
        resReleaseCuLgSubListV2(lgSubList, cuListResV2, itemFlags);
    }
    return (ret);
}

/*
 * Allocate cu from:
 *          Any device
 *          Any logic group
 * So the cu can from group on differernt device.
 */
int32_t xrm::system::resAllocCuFromAnyDeviceAnyLogicGroup(
    cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    int32_t ret = XRM_ERROR;
    int32_t devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }

        ret = resAllocCuFromTargetDeviceAnyLogicGroup(devId, cuPropV2, cuRes, updateId);
        if (ret == XRM_SUCCESS) {
            break;
        }
    }
    return (ret);
}

/*
 * Release the allocated resource for the logic group based sub-list,
 * and update the record in itemFlags which indicate resource allocated or not.
 */
void xrm::system::resReleaseCuLgSubListV2(logicGroupConstraintSubListTypeV2* lgSubList,
    cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t lgCuIndex, cuId;
    cuResource* cuRes = NULL;

    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request logic group based sub-list
        cuId = lgSubList->cuId[lgCuIndex];
        if (itemFlags[cuId].allocated) {
            cuRes = &cuListResV2->cuResources[cuId];
            resReleaseCuV2(cuRes);
            itemFlags[cuId].allocated = false;
        }
    }
}

/*
 * Alloc one cu (compute unit) from target device's any logic group.
 * It's used for cu list allocation, no allocate policy applied.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes.
 * Otherwise: failed to allocate the cu resource.
 *
 * NOTE: since there is no logic group constraint, the logic group infor in
 *       return resource will be set as invalid.
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocCuFromTargetDeviceAnyLogicGroup(int32_t devId, cuPropertyV2* cuPropV2,
    cuResource* cuRes, bool updateId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuPropV2->clientId;
    int32_t cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;

    // no need to check cuPropV2 or cuRes as it should be checked already
    cuPropertyCopyFromV2(cuProp, cuPropV2);

cu_alloc_loop:
    ret = allocClientFromDev(devId, cuProp);
    if (ret < 0) {
        return (ret);
    }
    dev = &m_devList[devId];
    /*
     * Now check whether matching cu is in target device, no logic group constraint
     */
    for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];

        /* first attempt to re-use existing kernels; else, use a new kernel */
        if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
        /* set the logic group infor as invalid for resource alloction */
        cuResourceLogicGroupInfo lgInfo;
        setLogicGroupInforInvalid(&lgInfo);
        /* alloc channel and register client id */
        /*
         * logic group information will be set in allocChanClientFromCu call:
         * if it's allocated from reservation pool and the pool resource logic group
         * information is valid, then allocated resource will carry the group
         * information of pool resource. Otherwise the allocated resource logic
         * group information will be set as invalid.
         */
        ret = allocChanClientFromCu(cu, cuProp, cuRes, &lgInfo);
        if (ret != XRM_SUCCESS) {
            continue;
        }

        /*
         * XRM_SUCCESS, channel allocation is done. To set allocated resource information.
         */
        cuRes->deviceId = devId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    } // loop cu in the device

    if (!cuAcquired) {
        releaseClientOnDev(devId, clientId);
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_alloc_loop;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Allocate multiple cu sub lists from target device:
 *
 * These different cu sub lists are generated based on logic group constraint information.
 */
int32_t xrm::system::resAllocCuSubListsFromTargetDevice(int32_t devId, int32_t numLgConstraintSubList,
    logicGroupConstraintSubListTypeV2* lgConstraintSubLists,
    cuListPropertyV2* cuListPropV2, cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret;
    logicGroupConstraintSubListTypeV2* lgSubList = NULL;
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList = NULL;
    int32_t lgSubListIndex, lgSubListType, lgId;

    /*
     * use to record used logic group id/device id pair for sub-list which request from
     * different pseudo logic group index (different pair means different logic group)
     */
    usedLgIdDevIdPairList = (logicGroupIdDeviceIdPairListPropertyV2*)malloc(
        sizeof(logicGroupIdDeviceIdPairListPropertyV2));
    memset(usedLgIdDevIdPairList, 0, sizeof(logicGroupIdDeviceIdPairListPropertyV2));

    // loop all the Logic group constraint based sub-Lists
    for (lgSubListIndex = 0; lgSubListIndex < numLgConstraintSubList; lgSubListIndex++) {
        lgSubList = &(lgConstraintSubLists[lgSubListIndex]);
        lgSubListType = lgSubList->type;
        switch (lgSubListType) {
            case LG_CONSTRAINT_SUBLIST_TYPE_REAL_GROUP_ID:
                /* Allocate this logic group sub-list from target logic group of this device */
                // get the target logic group id for this sub-list
                lgId = lgSubList->groupId;
                ret = resAllocCuSubListFromTargetDeviceTargetLogicGroup(devId, lgId, lgSubList,
                    cuListPropV2, cuListResV2, itemFlags);
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE:
                /*
                 * Allocate this logic group sub-list from target type logic group of this device.
                 * Can from different logic group.
                 */
                ret = resAllocCuSubListFromTargetDeviceTargetTypeLogicGroup(devId, lgSubList,
                    cuListPropV2, cuListResV2, itemFlags);
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_PSEUDO_GROUP_ID:
                if (lgSubList->groupId >= 0) {
                    /*
                     * Allocate this logic group sub-list from same logic group of this device.
                     * This logic group need to be different from previous used groups.
                     */
                    ret = resAllocCuSubListFromTargetDeviceSameLogicGroupNotUsed(devId, lgSubList,
                        cuListPropV2, cuListResV2, itemFlags, usedLgIdDevIdPairList);
                } else {
                    /*
                     * groupId < 0, which means just care the this sub list cu from same group,
                     * But not require this sub list must from a different logic group with other
                     * sub list.
                     */
                    ret = resAllocCuSubListFromTargetDeviceSameLogicGroupNoLimitUsedOrNot(devId, lgSubList,
                        cuListPropV2, cuListResV2, itemFlags);
                }
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE_AND_PSEUDO_GROUP_ID:
                if (lgSubList->groupId >= 0) {
                    /*
                     * Allocate this logic group sub-list from target type same logic group of this device.
                     * This logic group need to be different from previous used groups.
                     */
                    ret = resAllocCuSubListFromTargetDeviceTargetTypeSameLogicGroupNotUsed(devId, lgSubList,
                        cuListPropV2, cuListResV2, itemFlags, usedLgIdDevIdPairList);
                } else {
                    /*
                     * groupId < 0, which means just care the this sub list cu from same group,
                     * But not require this sub list must from a different logic group with other
                     * sub list.
                     */
                    ret = resAllocCuSubListFromTargetDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(devId,
                        lgSubList, cuListPropV2, cuListResV2, itemFlags);
                }
                break;
            case DEV_CONSTRAINT_SUBLIST_TYPE_NULL:
                ret = resAllocCuSubListFromTargetDeviceAnyLogicGroup(devId, lgSubList,
                    cuListPropV2, cuListResV2, itemFlags);
                break;
            default:
                ret = XRM_ERROR_INVALID;
                break;
        }
        if (ret != XRM_SUCCESS) {
            break;
        }
    } // loop lgConstraintSubLists

    if (ret != XRM_SUCCESS) {
        /* fail to allocate cu of this sub list, release the allocated resource for all sub-lists */
        for (lgSubListIndex = 0; lgSubListIndex < numLgConstraintSubList; lgSubListIndex++) {
            lgSubList = &(lgConstraintSubLists[lgSubListIndex]);
            resReleaseCuLgSubListV2(lgSubList, cuListResV2, itemFlags);
        }
    }
    if (usedLgIdDevIdPairList != NULL)
        free(usedLgIdDevIdPairList);
    return (ret);
}

/*
 * Allocate multiple cu sub lists from any device:
 *
 * These different cu sub lists are generated based on logic group constraint information.
 */
int32_t xrm::system::resAllocCuSubListsFromAnyDevice(int32_t numLgConstraintSubList,
    logicGroupConstraintSubListTypeV2* lgConstraintSubLists,
    cuListPropertyV2* cuListPropV2, cuListResourceV2* cuListResV2, itemFlag* itemFlags) {
    int32_t ret;
    logicGroupConstraintSubListTypeV2* lgSubList = NULL;
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList = NULL;
    int32_t lgSubListIndex, lgSubListType, lgId;

    /*
     * use to record used logic group id/device id pair for sub-list which request from
     * different pseudo logic group index (different pair means different logic group)
     */
    usedLgIdDevIdPairList = (logicGroupIdDeviceIdPairListPropertyV2*)malloc(
        sizeof(logicGroupIdDeviceIdPairListPropertyV2));
    memset(usedLgIdDevIdPairList, 0, sizeof(logicGroupIdDeviceIdPairListPropertyV2));

    // loop all the Logic group constraint based sub-Lists
    for (lgSubListIndex = 0; lgSubListIndex < numLgConstraintSubList; lgSubListIndex++) {
        lgSubList = &(lgConstraintSubLists[lgSubListIndex]);
        lgSubListType = lgSubList->type;
        switch (lgSubListType) {
            case LG_CONSTRAINT_SUBLIST_TYPE_REAL_GROUP_ID:
                /* 
                 * Allocate this logic group sub-list from target logic group of any device.
                 * CU can from different device.
                 */
                lgId = lgSubList->groupId;
                ret = resAllocCuSubListFromAnyDeviceTargetLogicGroup(lgId, lgSubList,
                    cuListPropV2, cuListResV2, itemFlags);
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE:
                /*
                 * Allocate this logic group sub-list from target type logic group of any device.
                 * CU can from different logic group.
                 */
                ret = resAllocCuSubListFromAnyDeviceTargetTypeLogicGroup(lgSubList,
                    cuListPropV2, cuListResV2, itemFlags);
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_PSEUDO_GROUP_ID:
                if (lgSubList->groupId >= 0) {
                    /*
                     * Allocate this logic group sub-list from same logic group of any device.
                     * groupId >= 0, This logic group need to be different from previous used groups.
                     */
                    ret = resAllocCuSubListFromAnyDeviceSameLogicGroupNotUsed(lgSubList,
                        cuListPropV2, cuListResV2, itemFlags, usedLgIdDevIdPairList);
                } else {
                    /*
                     * Allocate this logic group sub-list from same logic group of any device.
                     * groupId <0, no limit the logic group to be different from previous used groups.
                     */
                    ret = resAllocCuSubListFromAnyDeviceSameLogicGroupNoLimitUsedOrNot(lgSubList,
                        cuListPropV2, cuListResV2, itemFlags);
                }
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE_AND_PSEUDO_GROUP_ID:
                if (lgSubList->groupId >= 0) {
                    /*
                     * Allocate this logic group sub-list from target type same logic group of any device.
                     * groupId >= 0, This logic group need to be different from previous used groups.
                     */
                    ret = resAllocCuSubListFromAnyDeviceTargetTypeSameLogicGroupNotUsed(lgSubList,
                        cuListPropV2, cuListResV2, itemFlags, usedLgIdDevIdPairList);
                } else {
                    /*
                     * Allocate this logic group sub-list from target type same logic group of any device.
                     * groupId <0, no limit the logic group to be different from previous used groups.
                     */
                    ret = resAllocCuSubListFromAnyDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(lgSubList,
                        cuListPropV2, cuListResV2, itemFlags);
                }
                break;
            case DEV_CONSTRAINT_SUBLIST_TYPE_NULL:
                ret = resAllocCuSubListFromAnyDeviceAnyLogicGroup(lgSubList,
                    cuListPropV2, cuListResV2, itemFlags);
                break;
            default:
                ret = XRM_ERROR_INVALID;
                break;
        }
        if (ret != XRM_SUCCESS) {
            break;
        }
    } // loop lgConstraintSubLists

    if (ret != XRM_SUCCESS) {
        /* fail to allocate cu of this sub list, release the allocated resource for all sub-lists */
        for (lgSubListIndex = 0; lgSubListIndex < numLgConstraintSubList; lgSubListIndex++) {
            lgSubList = &(lgConstraintSubLists[lgSubListIndex]);
            resReleaseCuLgSubListV2(lgSubList, cuListResV2, itemFlags);
        }
    }
    if (usedLgIdDevIdPairList != NULL)
        free(usedLgIdDevIdPairList);
    return (ret);
}


/*
 * Alloc one cu list based on the request property.
 *
 * 0. Allocation policy is not applied for cu list allocation, so no need to handle it.
 *
 * 1. Device constraint is priority 1, it's handled first.
 *
 *    To handle the device constraint information, CU list will be put into differernt
 *    sub-list:
 *    a) one or multiple sub-lists with same hw device index, different index into
 *       different sub-list. CU in such sub-list will be allocated from target device later.
 *    b) one or multiple sub-lists with same virtual device index, different index
 *       into different sub-list. cu in same sub list will be allocated from same device.
 *       different sub-list will be allocated from different device later.
 *    c) one sub-list for cu without device constraint. cu in this sub-list will be
 *       allocated from any device later.
 *
 * 2. Logic group constraint is priority 2, it's handled based on device constraint
 *    process result.
 *
 *    Above device constraint based sub-lists willbe further put into logic group
 *    constraint based small sub-lists:
 *    a) one or multiple sub-list with same real group id.
 *    b) one or multiple sub-list with same group type.
 *    c) one or multiple sub-list with same pseudo group id, different list
 *       from different logic group.
 *    d) one or multiple sub-list with same pseudo group id and type, different list
 *       from different logic group.
 *    e) one sub-list for cu without logic group constraint.
 *
 *    Then these sub logic group based on sub-lists will be allocated accordingly:
 *    1) if logic group constraint not set: allocate it from any logic group.
 *    2) if logic group constraint set with real group id: allocate it from target logic group.
 *    3) if logic group constraint set with group type: allocate it from logic group with the type.
 *    4) if logic group constraint set with pseudo group id: find all others with same
 *       pseudo group id, put them into one sub-list, and allocate requested CU
 *       from same logic group, the logic group id should be different from that
 *       for previous sub-lists request from pseudo group id.
 *    5) if logic group constraint set with pseudo group id and type: find all others with same
 *       pseudo group id, put them into one sub-list, and check the group type information. If
 *       group type of same sub-list is different, whole allocation will fail. Then allocate
 *       requested CU from same logic group, the logic group id should be different from that
 *       for previous sub-lists request from pseudo group id.
 *
 * XRM_SUCCESS: allocated resouce is filled into cuListRes
 * Otherwise: failed to allocate the cu list
 *
 * Lock: should enter lock during the cu list allocation
 */
int32_t xrm::system::resAllocCuListV2(cuListPropertyV2* cuListPropV2, cuListResourceV2* cuListResV2) {
    int32_t index, ret;
    deviceData* dev = NULL;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = NULL;
    cuResource* cuRes = NULL;
    uint64_t clientId;

    itemFlag* itemFlags = NULL;
    deviceIdListPropertyV2* usedDevIdList = NULL;

    deviceConstraintSubListTypeV2* deviceConstraintSubLists = NULL;
    logicGroupConstraintSubListTypeV2* lgConstraintSubLists = NULL;

    deviceConstraintSubListTypeV2* devSubList = NULL;
    int32_t devSubListIndex, devSubListNumCu, devCuIndex, cuId, devSubListType, devId;
    int32_t numLgConstraintSubList;

    if (cuListPropV2 == NULL || cuListResV2 == NULL) {
        logMsg(XRM_LOG_ERROR, "%s : cuListPropV2 or cuListResV2 is null\n", __func__);
        return (XRM_ERROR_INVALID);
    }
    if (cuListPropV2->cuNum <= 0 || cuListPropV2->cuNum > XRM_MAX_LIST_CU_NUM_V2) {
        logMsg(XRM_LOG_ERROR, "%s : cuNum out of rage %d\n", __func__, cuListPropV2->cuNum);
        return (XRM_ERROR_INVALID);
    }
    // client id of same list is same
    clientId = cuListPropV2->cuProps[0].clientId;

    /* use to set the flag to record whether the item is handled */
    itemFlags = (itemFlag*)malloc(sizeof(itemFlag) * XRM_MAX_LIST_CU_NUM_V2);
    memset(itemFlags, 0, sizeof(itemFlag) * XRM_MAX_LIST_CU_NUM_V2);

    /* use to record used device id for sub-list which request from different virtual device index */
    usedDevIdList = (deviceIdListPropertyV2*)malloc(sizeof(deviceIdListPropertyV2));
    memset(usedDevIdList, 0, sizeof(deviceIdListPropertyV2));

    ret = XRM_ERROR;

    /*
     * 1. handle device constraint information
     *
     * put cu into differernt type deviceConstraintSubList:
     *     DEV_CONSTRAINT_SUBLIST_TYPE_HARDWARE_DEVICE_ID
     *     DEV_CONSTRAINT_SUBLIST_TYPE_VIRTUAL_DEVICE_ID
     *     DEV_CONSTRAINT_SUBLIST_TYPE_NULL
     */
    deviceConstraintSubLists = (deviceConstraintSubListTypeV2*)malloc(
        sizeof(deviceConstraintSubListTypeV2) * XRM_MAX_LIST_CU_NUM_V2);
    memset(deviceConstraintSubLists, 0, sizeof(deviceConstraintSubListTypeV2) * XRM_MAX_LIST_CU_NUM_V2);
    int32_t numDeviceConstraintSubList = 0;
    for (index = 0; index < cuListPropV2->cuNum; index++) {
        cuPropV2 = &cuListPropV2->cuProps[index];
        if ((cuPropV2->policyInfo & XRM_POLICY_INFO_CONSTRAINT_TYPE_MASK) != XRM_POLICY_INFO_CONSTRAINT_TYPE_NULL) {
            logMsg(XRM_LOG_ERROR, "%s : cu [%d] policy should not set for CU List\n", __func__, index);
            ret = XRM_ERROR_INVALID;
            break;
        }
        ret = putCuIntoDeviceConstraintSubLists(cuPropV2, index, &numDeviceConstraintSubList,
            deviceConstraintSubLists);
        if (ret != XRM_SUCCESS)
            break;
    }
    if (ret != XRM_SUCCESS) {
        goto resAllocCuListV2Exit;
    }

    /*
     * 2. handle logic group constraint information
     *
     * For each deviceConstraintSubList, put cu into differernt type logicGroupConstraintSubList:
     *     LG_CONSTRAINT_SUBLIST_TYPE_REAL_GROUP_ID
     *     LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE
     *     LG_CONSTRAINT_SUBLIST_TYPE_PSEUDO_GROUP_ID
     *     LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE_AND_PSEUDO_GROUP_ID
     *     DEV_CONSTRAINT_SUBLIST_TYPE_NULL
     * Then allocate cu of all logicGroupConstraintSubList (from this deviceConstraintSubList).
     */
    lgConstraintSubLists = (logicGroupConstraintSubListTypeV2*)malloc(sizeof
        (logicGroupConstraintSubListTypeV2) * XRM_MAX_LIST_CU_NUM_V2);

    for (devSubListIndex = 0; devSubListIndex < numDeviceConstraintSubList; devSubListIndex++) {
        /* handle each devSubList */

        /* init logic group related data for each devSubList */
        memset(lgConstraintSubLists, 0, sizeof(logicGroupConstraintSubListTypeV2) * XRM_MAX_LIST_CU_NUM_V2);
        numLgConstraintSubList = 0;

        devSubList = &(deviceConstraintSubLists[devSubListIndex]);
        devSubListNumCu = devSubList->numCu;

        /* put cu into differernt type logicGroupConstraintSubList */
        for (devCuIndex = 0; devCuIndex < devSubListNumCu; devCuIndex++) {
            /* get the cuId in cuListPropV2 */
            cuId = devSubList->cuId[devCuIndex];
            cuPropV2 = &cuListPropV2->cuProps[cuId];
            ret = putCuIntoLgConstraintSubLists(cuPropV2, cuId, &numLgConstraintSubList, lgConstraintSubLists);
            if (ret != XRM_SUCCESS) {
                goto resAllocCuListV2Exit;
            }
        } // put cu into differernt type logicGroupConstraintSubList

        /* allocate cu of each lgConstraintSubList */
        devSubListType = devSubList->type;
        switch (devSubListType) {
            case DEV_CONSTRAINT_SUBLIST_TYPE_HARDWARE_DEVICE_ID:
                /* handle the sub list with hardware device id */
                devId = devSubList->deviceId;

                dev = &m_devList[devId];
                if (!dev->isLoaded) {
                    ret = XRM_ERROR_NO_DEV;
                    goto resAllocCuListV2Exit;
                }
                if ((dev->isExcl) && (dev->clientProcs[0].clientId != clientId)) {
                    /* the device is used by other client exclusively */
                    ret = XRM_ERROR_NO_DEV;
                    goto resAllocCuListV2Exit;
                }

                // allocate resource for the logic based sub-lists (device id is set)
                ret = resAllocCuSubListsFromTargetDevice(devId, numLgConstraintSubList,
                    lgConstraintSubLists, cuListPropV2, cuListResV2, itemFlags);
                if (ret != XRM_SUCCESS) {
                    goto resAllocCuListV2Exit;
                }
                break;
            case DEV_CONSTRAINT_SUBLIST_TYPE_VIRTUAL_DEVICE_ID:
                /*
                 * To handle the sub list with virtual device id, Different sub list with differernt
                 * virtual device id.
                 *
                 * Allocate resource for the logic group based sub-lists (virtual device id is set):
                 * CU in one sub-list from same device, the device is not in used list to make sure
                 * CU in different sub-list will come from different device.
                 */
                for (devId = -1; devId < m_numDevice;) {
                    /* find one available device for first cu */
                    cuPropV2 = &(cuListPropV2->cuProps[lgConstraintSubLists[0].cuId[0]]);
                    cuPropertyCopyFromV2(cuProp, cuPropV2);
                alloc_device:
                    devId = allocDevForClient(&devId, cuProp);
                    if (devId < 0) {
                        ret = XRM_ERROR_NO_DEV;
                        goto resAllocCuListV2Exit;
                    }
                    /* check whether the device is in used devId list, if yes, then find another one */
                    if (isInDeviceIdList(usedDevIdList, devId)) {
                        goto alloc_device;
                    }

                    ret = resAllocCuSubListsFromTargetDevice(devId, numLgConstraintSubList,
                        lgConstraintSubLists, cuListPropV2, cuListResV2, itemFlags);
                    if (ret == XRM_SUCCESS) {
                        // update used device id list: add this new one in
                        usedDevIdList->deviceIds[usedDevIdList->deviceNum] = devId;
                        usedDevIdList->deviceNum++;
                        break;
                    }
                } // loop all the devices
                if (ret != XRM_SUCCESS) {
                    goto resAllocCuListV2Exit;
                }
                break;
            case DEV_CONSTRAINT_SUBLIST_TYPE_NULL:
                /*
                 * To handle the sub list without device constraint: can from any device:
                 * To allocate resource for the logic based sub-lists (device id is NOT set)
                 */
                ret = resAllocCuSubListsFromAnyDevice(numLgConstraintSubList, lgConstraintSubLists,
                    cuListPropV2, cuListResV2, itemFlags);
                if (ret != XRM_SUCCESS) {
                    goto resAllocCuListV2Exit;
                }
                break;
            default:
                logMsg(XRM_LOG_ERROR, "invalid devSubList [%d] constraint type\n", devSubListIndex);
                ret = XRM_ERROR_INVALID;
                break;
        } // allocate cu of each lgConstraintSubList in deviceConstraintSubList
    } // handle each devSubList

resAllocCuListV2Exit:
    if (ret == XRM_SUCCESS) {
        cuListResV2->cuNum = cuListPropV2->cuNum;
        updateAllocServiceId();
    } else {
        // fail to allocate whole list, release the allocated resource
        for (index = 0; index < cuListPropV2->cuNum; index++) {
            if (itemFlags[index].allocated) {
                cuRes = &cuListResV2->cuResources[index];
                resReleaseCuV2(cuRes);
            }
        }
        cuListResV2->cuNum = 0;
    }
    if (deviceConstraintSubLists != NULL)
        free(deviceConstraintSubLists);
    if (lgConstraintSubLists != NULL)
        free(lgConstraintSubLists);
    if (itemFlags != NULL)
        free(itemFlags);
    if (usedDevIdList != NULL)
        free(usedDevIdList);
    return (ret);
}

/*
 * Load xclbin to one device and alloc all cu from this device.
 *
 * XRM_SUCCESS: allocated resouce is filled into cuListRes
 * Otherwise: failed to load xclbin and allocate all cu from this device
 *
 * Lock: should enter lock during the operation
 */
int32_t xrm::system::resLoadAndAllocAllCu(std::string xclbin,
                                          uint64_t clientId,
                                          pid_t clientProcessId,
                                          cuListResource* cuListRes) {
    int32_t i, devId, ret;

    if (cuListRes == NULL) {
        return (XRM_ERROR_INVALID);
    }

    /*
     * try to load the xclbin to one possible device
     */
    devId = -1;
    std::string loadErrmsg;
    ret = loadAnyDevice(&devId, xclbin, loadErrmsg);
    if (ret != XRM_SUCCESS) {
        return (ret);
    }

    /*
     * allocate all the cu from this device
     */
    deviceData* dev = &m_devList[devId];
    dev->clientProcs[0].clientId = clientId;
    dev->clientProcs[0].clientProcessId = clientProcessId;
    dev->clientProcs[0].ref = 1;
    if (dev->xclbinInfo.numCu < XRM_MAX_LIST_CU_NUM)
        cuListRes->cuNum = dev->xclbinInfo.numCu;
    else
        cuListRes->cuNum = XRM_MAX_LIST_CU_NUM;
    cuData* cu = NULL;
    channelData* channel = NULL;
    cuResource* cuRes = NULL;
    uint64_t allocServiceId = getNextAllocServiceId();
    for (i = 0; i < cuListRes->cuNum; i++) {
        cuRes = &cuListRes->cuResources[i];
        cu = &dev->xclbinInfo.cuList[i];

        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->kernelName, cu->kernelName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->instanceName, cu->instanceName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->kernelAlias, cu->kernelAlias.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->cuName, cu->cuName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->kernelPluginFileName, cu->kernelPluginFileName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuRes->deviceId = devId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = i;
        cuRes->cuType = cu->cuType;
        cuRes->channelId = 0;
        cuRes->clientId = clientId;
        cuRes->channelLoadUnified = XRM_MAX_CHAN_LOAD_GRANULARITY_1000000;
        cuRes->channelLoadOriginal = XRM_MAX_CHAN_LOAD_GRANULARITY_100;
        cuRes->allocServiceId = allocServiceId;
        cuRes->baseAddr = cu->baseAddr;
        cuRes->membankId = cu->membankId;
        cuRes->membankType = cu->membankType;
        cuRes->membankSize = cu->membankSize;
        cuRes->membankBaseAddr = cu->membankBaseAddr;
        cuRes->poolId = 0;

        cu->totalUsedLoadUnified = XRM_MAX_CU_LOAD_GRANULARITY_1000000;
        cu->numChanInuse = 1;
        cu->clients[0] = clientId;
        cu->numClient = 1;
        channel = &cu->channels[0];
        channel->channelLoadUnified = XRM_MAX_CHAN_LOAD_GRANULARITY_1000000;
        channel->channelLoadOriginal = XRM_MAX_CHAN_LOAD_GRANULARITY_100;
        channel->allocServiceId = allocServiceId;
        channel->clientId = clientId;
        channel->clientProcessId = clientProcessId;
    }
    updateDeviceLoad(devId, 0, XRM_MAX_CU_LOAD_GRANULARITY_1000000 * cuListRes->cuNum);
    updateAllocServiceId();
    return (XRM_SUCCESS);
}

/*
 * flush the information data of user define cu group
 */
void xrm::system::flushUdfCuGroupInfo(uint32_t udfCuGroupIdx) {
    m_udfCuGroups[udfCuGroupIdx].udfCuGroupName = "";
    m_udfCuGroups[udfCuGroupIdx].optionUdfCuListNum = 0;
    for (int32_t cuListIdx = 0; cuListIdx < m_udfCuGroups[udfCuGroupIdx].optionUdfCuListNum; cuListIdx++) {
        memset(&m_udfCuGroups[udfCuGroupIdx].optionUdfCuListProps[cuListIdx], 0, sizeof(cuListProperty));
    }
}

/*
 * flush the information data of user define cu group V2
 */
void xrm::system::flushUdfCuGroupInfoV2(uint32_t udfCuGroupIdx) {
    m_udfCuGroupsV2[udfCuGroupIdx].udfCuGroupName = "";
    m_udfCuGroupsV2[udfCuGroupIdx].optionUdfCuListNum = 0;
    for (int32_t cuListIdx = 0; cuListIdx < m_udfCuGroupsV2[udfCuGroupIdx].optionUdfCuListNum; cuListIdx++) {
        memset(&m_udfCuGroupsV2[udfCuGroupIdx].optionUdfCuListProps[cuListIdx], 0, sizeof(cuListPropertyV2));
    }
}

/*
 * declare one user defined cu group type based on the input information.
 *
 * XRM_SUCCESS: the user defined cu group is added into end of m_udfCuGroups[], no hole
 * Otherwise: failed to declare the user defined cu group
 *
 * Lock: should enter lock during the cu group declaration
 */
int32_t xrm::system::resUdfCuGroupDeclare(udfCuGroupInformation* udfCuGroupInfo) {
    cuListProperty* cuListProp;
    cuListProperty* udfCuListProp;
    int32_t udfCuGroupIdx = -1;

    if (udfCuGroupInfo == NULL) {
        return (XRM_ERROR_INVALID);
    }
    for (uint32_t cuGroupIdx = 0; cuGroupIdx < m_numUdfCuGroup; cuGroupIdx++) {
        /* compare, 0: equal */
        if (!m_udfCuGroups[cuGroupIdx].udfCuGroupName.compare(udfCuGroupInfo->udfCuGroupName.c_str())) {
            udfCuGroupIdx = (int32_t)cuGroupIdx;
            break;
        }
    }
    if (udfCuGroupIdx != -1) {
        logMsg(XRM_LOG_ERROR, "%s : user defined cu group is already existing\n", __func__);
        return (XRM_ERROR_INVALID);
    }

    m_udfCuGroups[m_numUdfCuGroup].udfCuGroupName = udfCuGroupInfo->udfCuGroupName;
    m_udfCuGroups[m_numUdfCuGroup].optionUdfCuListNum = udfCuGroupInfo->optionUdfCuListNum;
    for (int32_t cuListIdx = 0; cuListIdx < udfCuGroupInfo->optionUdfCuListNum; cuListIdx++) {
        cuListProp = &m_udfCuGroups[m_numUdfCuGroup].optionUdfCuListProps[cuListIdx];
        udfCuListProp = &udfCuGroupInfo->optionUdfCuListProps[cuListIdx];
        memset(cuListProp, 0, sizeof(cuListProperty));
        cuListProp->cuNum = udfCuListProp->cuNum;
        cuListProp->sameDevice = udfCuListProp->sameDevice;
        for (int32_t i = 0; i < cuListProp->cuNum; i++) {
            strncpy(cuListProp->cuProps[i].cuName, udfCuListProp->cuProps[i].cuName, XRM_MAX_NAME_LEN - 1);
            cuListProp->cuProps[i].devExcl = udfCuListProp->cuProps[i].devExcl;
            cuListProp->cuProps[i].requestLoadUnified = udfCuListProp->cuProps[i].requestLoadUnified;
            cuListProp->cuProps[i].requestLoadOriginal = udfCuListProp->cuProps[i].requestLoadOriginal;
        }
    }
    m_numUdfCuGroup++;
    return (XRM_SUCCESS);
}

/*
 * declare one user defined cu group type based on the input information.
 *
 * XRM_SUCCESS: the user defined cu group is added into end of m_udfCuGroupsV2[], no hole
 * Otherwise: failed to declare the user defined cu group
 *
 * Lock: should enter lock during the cu group declaration
 */
int32_t xrm::system::resUdfCuGroupDeclareV2(udfCuGroupInformationV2* udfCuGroupInfoV2) {
    cuListPropertyV2* cuListProp;
    cuListPropertyV2* udfCuListProp;
    int32_t udfCuGroupIdx = -1;

    if (udfCuGroupInfoV2 == NULL) {
        return (XRM_ERROR_INVALID);
    }
    for (uint32_t cuGroupIdx = 0; cuGroupIdx < m_numUdfCuGroupV2; cuGroupIdx++) {
        /* compare, 0: equal */
        if (!m_udfCuGroupsV2[cuGroupIdx].udfCuGroupName.compare(udfCuGroupInfoV2->udfCuGroupName.c_str())) {
            udfCuGroupIdx = (int32_t)cuGroupIdx;
            break;
        }
    }
    if (udfCuGroupIdx != -1) {
        logMsg(XRM_LOG_ERROR, "%s : user defined cu group is already existing\n", __func__);
        return (XRM_ERROR_INVALID);
    }

    m_udfCuGroupsV2[m_numUdfCuGroupV2].udfCuGroupName = udfCuGroupInfoV2->udfCuGroupName;
    m_udfCuGroupsV2[m_numUdfCuGroupV2].optionUdfCuListNum = udfCuGroupInfoV2->optionUdfCuListNum;
    for (int32_t cuListIdx = 0; cuListIdx < udfCuGroupInfoV2->optionUdfCuListNum; cuListIdx++) {
        cuListProp = &m_udfCuGroupsV2[m_numUdfCuGroupV2].optionUdfCuListProps[cuListIdx];
        udfCuListProp = &udfCuGroupInfoV2->optionUdfCuListProps[cuListIdx];
        memset(cuListProp, 0, sizeof(cuListPropertyV2));
        cuListProp->cuNum = udfCuListProp->cuNum;
        for (int32_t i = 0; i < cuListProp->cuNum; i++) {
            strncpy(cuListProp->cuProps[i].cuName, udfCuListProp->cuProps[i].cuName, XRM_MAX_NAME_LEN - 1);
            cuListProp->cuProps[i].devExcl = udfCuListProp->cuProps[i].devExcl;
            cuListProp->cuProps[i].deviceInfo = udfCuListProp->cuProps[i].deviceInfo;
            cuListProp->cuProps[i].memoryInfo = udfCuListProp->cuProps[i].memoryInfo;
            cuListProp->cuProps[i].requestLoadUnified = udfCuListProp->cuProps[i].requestLoadUnified;
            cuListProp->cuProps[i].requestLoadOriginal = udfCuListProp->cuProps[i].requestLoadOriginal;
        }
    }
    m_numUdfCuGroupV2++;
    return (XRM_SUCCESS);
}

/*
 * undeclare one user defined cu group type based on the input information.
 *
 * XRM_SUCCESS: the user defined cu group is removed from m_udfCuGroups[], no hole
 * Otherwise: failed to undeclare the user defined cu group
 *
 * Lock: should enter lock during the cu group undeclaration
 */
int32_t xrm::system::resUdfCuGroupUndeclare(udfCuGroupInformation* udfCuGroupInfo) {
    cuListProperty* desCuListProp;
    cuListProperty* srcCuListProp;
    bool udfCuGroupFound = false;
    uint32_t udfCuGroupIdx;

    if (udfCuGroupInfo == NULL) {
        return (XRM_ERROR_INVALID);
    }
    for (udfCuGroupIdx = 0; udfCuGroupIdx < m_numUdfCuGroup; udfCuGroupIdx++) {
        /* compare, 0: equal */
        if (!m_udfCuGroups[udfCuGroupIdx].udfCuGroupName.compare(udfCuGroupInfo->udfCuGroupName.c_str())) {
            udfCuGroupFound = true;
            break;
        }
    }
    if (!udfCuGroupFound) {
        logMsg(XRM_LOG_ERROR, "%s : user defined cu group is NOT existing\n", __func__);
        return (XRM_ERROR_INVALID);
    }

    /* move slot one step forward */
    for (; udfCuGroupIdx < (m_numUdfCuGroup - 1); udfCuGroupIdx++) {
        m_udfCuGroups[udfCuGroupIdx].udfCuGroupName = m_udfCuGroups[udfCuGroupIdx + 1].udfCuGroupName;
        m_udfCuGroups[udfCuGroupIdx].optionUdfCuListNum = m_udfCuGroups[udfCuGroupIdx + 1].optionUdfCuListNum;
        for (int32_t cuListIdx = 0; cuListIdx < m_udfCuGroups[udfCuGroupIdx].optionUdfCuListNum; cuListIdx++) {
            desCuListProp = &m_udfCuGroups[udfCuGroupIdx].optionUdfCuListProps[cuListIdx];
            srcCuListProp = &m_udfCuGroups[udfCuGroupIdx + 1].optionUdfCuListProps[cuListIdx];
            memset(desCuListProp, 0, sizeof(cuListProperty));
            desCuListProp->cuNum = srcCuListProp->cuNum;
            desCuListProp->sameDevice = srcCuListProp->sameDevice;
            for (int32_t i = 0; i < desCuListProp->cuNum; i++) {
                strncpy(desCuListProp->cuProps[i].cuName, srcCuListProp->cuProps[i].cuName, XRM_MAX_NAME_LEN - 1);
                desCuListProp->cuProps[i].devExcl = srcCuListProp->cuProps[i].devExcl;
                desCuListProp->cuProps[i].requestLoadUnified = srcCuListProp->cuProps[i].requestLoadUnified;
                desCuListProp->cuProps[i].requestLoadOriginal = srcCuListProp->cuProps[i].requestLoadOriginal;
            }
        }
    }

    /* flush the last slot */
    flushUdfCuGroupInfo(udfCuGroupIdx);
    m_numUdfCuGroup--;

    return (XRM_SUCCESS);
}

/*
 * undeclare one user defined cu group type based on the input information.
 *
 * XRM_SUCCESS: the user defined cu group is removed from m_udfCuGroupsV2[], no hole
 * Otherwise: failed to undeclare the user defined cu group
 *
 * Lock: should enter lock during the cu group undeclaration
 */
int32_t xrm::system::resUdfCuGroupUndeclareV2(udfCuGroupInformationV2* udfCuGroupInfoV2) {
    cuListPropertyV2* desCuListProp;
    cuListPropertyV2* srcCuListProp;
    bool udfCuGroupFound = false;
    uint32_t udfCuGroupIdx;

    if (udfCuGroupInfoV2 == NULL) {
        return (XRM_ERROR_INVALID);
    }
    for (udfCuGroupIdx = 0; udfCuGroupIdx < m_numUdfCuGroupV2; udfCuGroupIdx++) {
        /* compare, 0: equal */
        if (!m_udfCuGroupsV2[udfCuGroupIdx].udfCuGroupName.compare(udfCuGroupInfoV2->udfCuGroupName.c_str())) {
            udfCuGroupFound = true;
            break;
        }
    }
    if (!udfCuGroupFound) {
        logMsg(XRM_LOG_ERROR, "%s : user defined cu group is NOT existing\n", __func__);
        return (XRM_ERROR_INVALID);
    }

    /* move slot one step forward */
    for (; udfCuGroupIdx < (m_numUdfCuGroupV2 - 1); udfCuGroupIdx++) {
        m_udfCuGroupsV2[udfCuGroupIdx].udfCuGroupName = m_udfCuGroupsV2[udfCuGroupIdx + 1].udfCuGroupName;
        m_udfCuGroupsV2[udfCuGroupIdx].optionUdfCuListNum = m_udfCuGroupsV2[udfCuGroupIdx + 1].optionUdfCuListNum;
        for (int32_t cuListIdx = 0; cuListIdx < m_udfCuGroupsV2[udfCuGroupIdx].optionUdfCuListNum; cuListIdx++) {
            desCuListProp = &m_udfCuGroupsV2[udfCuGroupIdx].optionUdfCuListProps[cuListIdx];
            srcCuListProp = &m_udfCuGroupsV2[udfCuGroupIdx + 1].optionUdfCuListProps[cuListIdx];
            memset(desCuListProp, 0, sizeof(cuListPropertyV2));
            desCuListProp->cuNum = srcCuListProp->cuNum;
            for (int32_t i = 0; i < desCuListProp->cuNum; i++) {
                strncpy(desCuListProp->cuProps[i].cuName, srcCuListProp->cuProps[i].cuName, XRM_MAX_NAME_LEN - 1);
                desCuListProp->cuProps[i].devExcl = srcCuListProp->cuProps[i].devExcl;
                desCuListProp->cuProps[i].deviceInfo = srcCuListProp->cuProps[i].deviceInfo;
                desCuListProp->cuProps[i].memoryInfo = srcCuListProp->cuProps[i].memoryInfo;
                desCuListProp->cuProps[i].requestLoadUnified = srcCuListProp->cuProps[i].requestLoadUnified;
                desCuListProp->cuProps[i].requestLoadOriginal = srcCuListProp->cuProps[i].requestLoadOriginal;
            }
        }
    }

    /* flush the last slot */
    flushUdfCuGroupInfoV2(udfCuGroupIdx);
    m_numUdfCuGroupV2--;

    return (XRM_SUCCESS);
}

/*
 * Alloc one cu group based on the request property.
 *
 * XRM_SUCCESS: allocated resouce is filled into cuGroupRes
 * Otherwise: failed to allocate the cu group
 *
 * Lock: should enter lock during the cu group allocation
 */
int32_t xrm::system::resAllocCuGroup(cuGroupProperty* cuGroupProp, cuGroupResource* cuGroupRes) {
    int32_t ret = XRM_ERROR;
    cuListProperty cuListProp;
    cuListProperty* udfCuListProp;
    cuProperty* cuProp;
    cuProperty* udfCuProp;
    int32_t udfCuGroupIdx = -1;

    if (cuGroupProp == NULL || cuGroupRes == NULL) {
        return (XRM_ERROR_INVALID);
    }
    for (uint32_t cuGroupIdx = 0; cuGroupIdx < m_numUdfCuGroup; cuGroupIdx++) {
        /* compare, 0: equal */
        if (!m_udfCuGroups[cuGroupIdx].udfCuGroupName.compare(cuGroupProp->udfCuGroupName.c_str())) {
            udfCuGroupIdx = cuGroupIdx;
            break;
        }
    }
    if (udfCuGroupIdx == -1) {
        logMsg(XRM_LOG_ERROR, "%s : user defined cu group is not declared\n", __func__);
        return (XRM_ERROR_INVALID);
    }
    for (int32_t cuListIdx = 0; cuListIdx < m_udfCuGroups[udfCuGroupIdx].optionUdfCuListNum; cuListIdx++) {
        memset(&cuListProp, 0, sizeof(cuListProperty));
        udfCuListProp = &m_udfCuGroups[udfCuGroupIdx].optionUdfCuListProps[cuListIdx];
        cuListProp.cuNum = udfCuListProp->cuNum;
        cuListProp.sameDevice = udfCuListProp->sameDevice;
        for (int32_t i = 0; i < cuListProp.cuNum; i++) {
            cuProp = &cuListProp.cuProps[i];
            udfCuProp = &udfCuListProp->cuProps[i];
            strncpy(cuProp->cuName, udfCuProp->cuName, XRM_MAX_NAME_LEN - 1);
            cuProp->devExcl = udfCuProp->devExcl;
            cuProp->requestLoadUnified = udfCuProp->requestLoadUnified;
            cuProp->requestLoadOriginal = udfCuProp->requestLoadOriginal;
            cuProp->clientId = cuGroupProp->clientId;
            cuProp->clientProcessId = cuGroupProp->clientProcessId;
            cuProp->poolId = cuGroupProp->poolId;
        }
        memset(cuGroupRes, 0, sizeof(cuGroupResource));
        /*
         * NOTE: here since the cuListResource has same member of cuGroupResource, so it's fine to
         * use cuGroupRes to call resAllocCuList() directly, otherwise should use cuListResource instead.
         */
        ret = resAllocCuList(&cuListProp, (cuListResource*)cuGroupRes);
        if (ret == XRM_SUCCESS) break;
    }
    return (ret);
}

/*
 * Alloc one cu group based on the request property.
 *
 * XRM_SUCCESS: allocated resouce is filled into cuGroupRes
 * Otherwise: failed to allocate the cu group
 *
 * Lock: should enter lock during the cu group allocation
 */
int32_t xrm::system::resAllocCuGroupV2(cuGroupPropertyV2* cuGroupPropV2, cuGroupResourceV2* cuGroupResV2) {
    int32_t ret = XRM_ERROR;
    cuListPropertyV2 cuListProp;
    cuListPropertyV2* udfCuListProp;
    cuPropertyV2* cuProp;
    cuPropertyV2* udfCuProp;
    int32_t udfCuGroupIdx = -1;

    if (cuGroupPropV2 == NULL || cuGroupResV2 == NULL) {
        return (XRM_ERROR_INVALID);
    }
    for (uint32_t cuGroupIdx = 0; cuGroupIdx < m_numUdfCuGroupV2; cuGroupIdx++) {
        /* compare, 0: equal */
        if (!m_udfCuGroupsV2[cuGroupIdx].udfCuGroupName.compare(cuGroupPropV2->udfCuGroupName.c_str())) {
            udfCuGroupIdx = cuGroupIdx;
            break;
        }
    }
    if (udfCuGroupIdx == -1) {
        logMsg(XRM_LOG_ERROR, "%s : user defined cu group is not declared\n", __func__);
        return (XRM_ERROR_INVALID);
    }
    for (int32_t cuListIdx = 0; cuListIdx < m_udfCuGroupsV2[udfCuGroupIdx].optionUdfCuListNum; cuListIdx++) {
        memset(&cuListProp, 0, sizeof(cuListPropertyV2));
        udfCuListProp = &m_udfCuGroupsV2[udfCuGroupIdx].optionUdfCuListProps[cuListIdx];
        cuListProp.cuNum = udfCuListProp->cuNum;
        for (int32_t i = 0; i < cuListProp.cuNum; i++) {
            cuProp = &cuListProp.cuProps[i];
            udfCuProp = &udfCuListProp->cuProps[i];
            strncpy(cuProp->cuName, udfCuProp->cuName, XRM_MAX_NAME_LEN - 1);
            cuProp->devExcl = udfCuProp->devExcl;
            cuProp->deviceInfo = udfCuProp->deviceInfo;
            cuProp->memoryInfo = udfCuProp->memoryInfo;
            cuProp->requestLoadUnified = udfCuProp->requestLoadUnified;
            cuProp->requestLoadOriginal = udfCuProp->requestLoadOriginal;
            cuProp->clientId = cuGroupPropV2->clientId;
            cuProp->clientProcessId = cuGroupPropV2->clientProcessId;
            cuProp->poolId = cuGroupPropV2->poolId;
        }
        memset(cuGroupResV2, 0, sizeof(cuGroupResourceV2));
        /*
         * NOTE: here since the cuListResource has same member of cuGroupResource, so it's fine to
         * use cuGroupResV2 to call resAllocCuList() directly, otherwise should use cuListResource instead.
         */
        ret = resAllocCuListV2(&cuListProp, (cuListResourceV2*)cuGroupResV2);
        if (ret == XRM_SUCCESS) break;
    }
    return (ret);
}

/*
 * the interface to free one cu resource
 *
 * lock: need to enter lock to protect the resource pool access during the free process
 */
int32_t xrm::system::resReleaseCu(cuResource* cuRes) {
    if (cuRes == NULL) return (XRM_ERROR_INVALID);

    deviceData* dev;
    cuData* cu;
    int32_t devId, cuId, chanId;
    uint64_t allocServiceId;
    uint64_t clientId = cuRes->clientId;
    int32_t i, reserveIdx, procChanCnt = 0;
    int32_t ret;

    devId = cuRes->deviceId;
    if (devId < 0 || devId > m_numDevice) return (XRM_ERROR_INVALID);
    cuId = cuRes->cuId;
    if (cuId < 0 || cuId > XRM_MAX_XILINX_KERNELS) return (XRM_ERROR_INVALID);
    chanId = cuRes->channelId;
    if (chanId < 0 || chanId > XRM_MAX_KERNEL_CHANNELS) return (XRM_ERROR_INVALID);
    allocServiceId = cuRes->allocServiceId;

    ret = XRM_ERROR;
    dev = &m_devList[devId];
    if (!dev->isLoaded) {
        return (ret);
    }
    if (cuId >= dev->xclbinInfo.numCu) {
        return (ret);
    }
    cu = &dev->xclbinInfo.cuList[cuId];

    for (i = 0; i < XRM_MAX_KERNEL_CHANNELS; i++) {
        if (cu->channels[i].allocServiceId != allocServiceId || cu->channels[i].poolId != cuRes->poolId ||
            cu->channels[i].clientId != clientId || cu->channels[i].channelId != chanId) {
            if (cu->channels[i].clientId == clientId) /* count how many other channel used by the same client */
                procChanCnt++;
            continue;
        }

        /*
         * For the release cu:
         * if not from reserve pool, then return to Big pool
         * if from reserve pool, pool is active, return to pool
         * if from reserve pool, pool is in-active, return to Big pool
         */
        if (cuRes->poolId) {
            /* from the reserve pool */
            reserveIdx = isReservePoolUsingCu(cu, cuRes->poolId);
            if (reserveIdx != -1) {
                if (cu->reserves[reserveIdx].clientIsActive) {
                    /* From reserve pool, reserve client is active, return to reserve pool */
                    cu->reserves[reserveIdx].reserveUsedLoadUnified -= cu->channels[i].channelLoadUnified;
                } else {
                    /* From reserve pool, reserve client is NOT active, return resource to default pool */
                    cu->totalUsedLoadUnified -= cu->channels[i].channelLoadUnified;
                    updateDeviceLoad(cu->deviceId, -cu->channels[i].channelLoadUnified, -1);
                }
            } else {
                /* From reserve pool, reserve client is NOT active, return resource to default pool */
                cu->totalUsedLoadUnified -= cu->channels[i].channelLoadUnified;
                updateDeviceLoad(cu->deviceId, -cu->channels[i].channelLoadUnified, -1);
            }
        } else {
            /* return the resource into default pool */
            cu->totalUsedLoadUnified -= cu->channels[i].channelLoadUnified;
            updateDeviceLoad(cu->deviceId, -cu->channels[i].channelLoadUnified, -1);
        }
        cu->numChanInuse--;
        cu->channels[i].allocServiceId = 0;
        cu->channels[i].poolId = 0;
        cu->channels[i].clientId = 0;
        cu->channels[i].clientProcessId = 0;
        cu->channels[i].channelLoadUnified = 0;
        cu->channels[i].channelLoadOriginal = 0;
        memset(&(cu->channels[i].logicGroupInfo), 0, sizeof(cuResourceLogicGroupInfo));

        /* update dev->clientProcs ref count */
        releaseClientOnDev(devId, clientId);

        ret = XRM_SUCCESS;
    }

    /* If there is NO other channel used by the same client */
    if (!procChanCnt) {
        /* update cu->clients */
        removeClientOnCu(cu, clientId);
    }

    return (ret);
}

/*
 * the interface to free one cu resource
 *
 * lock: need to enter lock to protect the resource pool access during the free process
 */
int32_t xrm::system::resReleaseCuV2(cuResource* cuRes) {
    int32_t ret = resReleaseCu(cuRes);
    return (ret);
}

int32_t xrm::system::resReleaseCuList(cuListResource* cuListRes) {
    int32_t i, ret;
    cuResource* cuRes;

    if (cuListRes == NULL) return (XRM_ERROR_INVALID);
    if (cuListRes->cuNum < 0 || cuListRes->cuNum > XRM_MAX_LIST_CU_NUM) return (XRM_ERROR_INVALID);
    /* No resource need to be released */
    if (cuListRes->cuNum == 0) return (XRM_SUCCESS);

    for (i = 0; i < cuListRes->cuNum; i++) {
        cuRes = &cuListRes->cuResources[i];
        ret = resReleaseCu(cuRes);
        if (ret != XRM_SUCCESS) break;
    }
    return (ret);
}

int32_t xrm::system::resReleaseCuListV2(cuListResourceV2* cuListResV2) {
    int32_t i, ret;
    cuResource* cuRes;

    if (cuListResV2 == NULL) return (XRM_ERROR_INVALID);
    if (cuListResV2->cuNum < 0 || cuListResV2->cuNum > XRM_MAX_LIST_CU_NUM_V2) return (XRM_ERROR_INVALID);
    /* No resource need to be released */
    if (cuListResV2->cuNum == 0) return (XRM_SUCCESS);

    for (i = 0; i < cuListResV2->cuNum; i++) {
        cuRes = &cuListResV2->cuResources[i];
        ret = resReleaseCuV2(cuRes);
        if (ret != XRM_SUCCESS) break;
    }
    return (ret);
}

int32_t xrm::system::resReleaseCuGroup(cuGroupResource* cuGroupRes) {
    int32_t i, ret;
    cuResource* cuRes;

    if (cuGroupRes == NULL) return (XRM_ERROR_INVALID);
    if (cuGroupRes->cuNum < 0 || cuGroupRes->cuNum > XRM_MAX_GROUP_CU_NUM) return (XRM_ERROR_INVALID);
    /* No resource need to be released */
    if (cuGroupRes->cuNum == 0) return (XRM_SUCCESS);

    for (i = 0; i < cuGroupRes->cuNum; i++) {
        cuRes = &cuGroupRes->cuResources[i];
        ret = resReleaseCu(cuRes);
        if (ret != XRM_SUCCESS) break;
    }
    return (ret);
}

int32_t xrm::system::resReleaseCuGroupV2(cuGroupResourceV2* cuGroupRes) {
    int32_t i, ret;
    cuResource* cuRes;

    if (cuGroupRes == NULL) return (XRM_ERROR_INVALID);
    if (cuGroupRes->cuNum < 0 || cuGroupRes->cuNum > XRM_MAX_GROUP_CU_NUM_V2) return (XRM_ERROR_INVALID);
    /* No resource need to be released */
    if (cuGroupRes->cuNum == 0) return (XRM_SUCCESS);

    for (i = 0; i < cuGroupRes->cuNum; i++) {
        cuRes = &cuGroupRes->cuResources[i];
        ret = resReleaseCuV2(cuRes);
        if (ret != XRM_SUCCESS) break;
    }
    return (ret);
}

/*
 * To get next free device for client, but do not allocate client on the device.
 *
 * The dev->clientProcs is NOT updated.
 *
 * Called while holding lock, so not enter lock in the function
 *
 */
int32_t xrm::system::allocDevForClient(int32_t* devId, cuProperty* cuProp) {
    if (cuProp == NULL) return (XRM_ERROR_INVALID);

    while (*devId < m_numDevice) {
        int ret;

        /* the next free device id is updated to "devId" */
        ret = getNextFreeDevForClient(devId, cuProp);
        if (ret < 0) {
            return (ret);
        }
        return (*devId);
    }

    return (XRM_ERROR);
}

/*
 * find next available device for specified client
 *
 * XRM_SUCCESS: find one free/available device, and set the id of this device to *devId
 * XRM_ERROR_NO_DEV: there is no free device
 */
int32_t xrm::system::getNextFreeDevForClient(int32_t* devId, cuProperty* cuProp) {
    if (cuProp == NULL) return (XRM_ERROR_INVALID);

    deviceData* dev = NULL;
    uint64_t clientId = cuProp->clientId;
    int32_t i;

    /* start search from next device: *devId + 1 */
    for (i = *devId >= 0 ? *devId + 1 : 0; i < m_numDevice; i++) {
        dev = &m_devList[i];
        if (!dev->isLoaded) continue;

        if (dev->isExcl) {
            if (dev->clientProcs[0].clientId == clientId) {
                /* the client is still inuse */
                *devId = i;
                return (XRM_SUCCESS);
            }
            continue;
        }
        *devId = i;
        return (XRM_SUCCESS);
    }
    return (XRM_ERROR_NO_DEV);
}

/*
 * To alloc the client resource on specified device
 *
 * The dev->clientProcs ref count is updated, it's used for dev alloc/release.
 */
int32_t xrm::system::allocClientFromDev(int32_t devId, cuProperty* cuProp) {
    if (devId < 0 || devId >= m_numDevice) {
        return (XRM_ERROR_INVALID);
    }
    if (cuProp == NULL) {
        return (XRM_ERROR_INVALID);
    }

    deviceData* deviceList = m_devList;
    uint64_t clientId = cuProp->clientId;
    pid_t clientProcessId = cuProp->clientProcessId;
    bool excl = cuProp->devExcl;
    int32_t pidIdx;
    int32_t ref;

    /* does process already have exclusive access? */
    if (deviceList[devId].isExcl) {
        if (deviceList[devId].clientProcs[0].clientId == clientId) {
            deviceList[devId].clientProcs[0].ref++;
            return (XRM_SUCCESS);
        } else {
            return (XRM_ERROR_NO_DEV);
        }
    }

    /*
     * register client as using exclusive device
     */
    if (excl) {
        ref = 0;
        /* is another client already using this as a non-exclusive device? */
        for (pidIdx = 0; pidIdx < XRM_MAX_DEV_CLIENTS; pidIdx++) {
            if (deviceList[devId].clientProcs[pidIdx].clientId &&
                deviceList[devId].clientProcs[pidIdx].clientId != clientId) {
                return (XRM_ERROR_NO_DEV);
            } else {
                if (deviceList[devId].clientProcs[pidIdx].clientId == clientId)
                    ref = deviceList[devId].clientProcs[pidIdx].ref;
            }
        }
        deviceList[devId].isExcl = true;
        memset(deviceList[devId].clientProcs, 0, sizeof(clientData) * XRM_MAX_DEV_CLIENTS);
        deviceList[devId].clientProcs[0].clientId = clientId;
        deviceList[devId].clientProcs[0].clientProcessId = clientProcessId;
        deviceList[devId].clientProcs[0].ref = ref + 1;
        return (XRM_SUCCESS);
    }

    /*
     * register client as using non-exclusive device
     */
    /* client is already using this non-exclusive device? */
    for (pidIdx = 0; pidIdx < XRM_MAX_DEV_CLIENTS; pidIdx++)
        if (deviceList[devId].clientProcs[pidIdx].clientId == clientId) {
            deviceList[devId].clientProcs[pidIdx].ref++;
            return (XRM_SUCCESS);
        }
    /* There maybe empty slot in dev->clientProcs */
    for (pidIdx = 0; pidIdx < XRM_MAX_DEV_CLIENTS; pidIdx++)
        if (!deviceList[devId].clientProcs[pidIdx].clientId) {
            deviceList[devId].clientProcs[pidIdx].clientId = clientId;
            deviceList[devId].clientProcs[pidIdx].clientProcessId = clientProcessId;
            deviceList[devId].clientProcs[pidIdx].ref = 1;
            return (XRM_SUCCESS);
        }

    return (XRM_ERROR_NO_DEV);
}

/*
 * To check whether the client process pid is still alive.
 * This function will NOT work within container environment.
 *
 * XRM_ERROR: client is not alive
 * XRM_SUCCESS: client is alive
 */
int32_t xrm::system::verifyProcess(pid_t pid) {
    struct stat statBuf;
    int32_t ret;
    char procfsPid[20] = {0};

    snprintf(procfsPid, sizeof(procfsPid), "/proc/%d", pid);
    ret = stat(procfsPid, &statBuf);
    if (ret && errno == ENOENT) {
        return (XRM_ERROR);
    }

    return (XRM_SUCCESS);
}

/*
 * Alloc one cu (compute unit) from specified device(devId) based on the request property.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * call while holding system lock
 */
int32_t xrm::system::allocCuFromDev(int32_t devId, cuProperty* cuProp, cuResource* cuRes) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    int32_t cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    if (devId < 0 || devId >= m_numDevice) {
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp == NULL) || (cuRes == NULL)) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0') && (cuProp->cuName[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "None of kernel name, kernel alias and cu name are presented\n");
        return (XRM_ERROR_INVALID);
    }

    dev = &m_devList[devId];
cu_alloc_again:
    if (!cuAcquired) {
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to re-use existing kernels; else, use a new kernel */
            if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            /* cu name is presented, compare it; otherwise no need to compare */
            cuNameEqual = true;
            if (cuProp->cuName[0] != '\0') {
                if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) continue;
            /* alloc channel and register client id */
            ret = allocChanClientFromCu(cu, cuProp, cuRes);
            if (ret != XRM_SUCCESS) continue;

            cuRes->deviceId = devId;
            cuRes->deviceNumBDF = dev->devNumBDF;
            cuRes->cuId = cuId;
            strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
            strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
            cuAcquired = true;
        }
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_alloc_again;
    }

    if (cuAcquired) {
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * call while holding system lock
 *
 * Alloc channel from cu, update cu resource pool.
 * if reservation id is set, then allocate the resource from reserve pool.
 *
 * allocChanClientFromCu is compatible with previous version (no policy support for allocating cu)
 * with default value of last three parameters
 *
 * NOTE:
 *     If the resource is allocated from default pool (not from reservation pool):
 *        then the allocated resource logic group information will be set
 *        based on the input lgInfor.
 *     If the resource is allocated from reservation pool:
 *        if input lgInfor showed logic group infor is valid
 *           then allocated resource logic group infor will be set based on input lgInfo.
 *        else if pool resource logic group information is valid
 *             then allocated resource will carry logic group information of pool resource.
 *             else allocated resource logic group information will be set as invalid by default.
 *
 * If the policyInfo is set and current cu load is better than precu load, it
 * will update precu to current cu and return XRM_FURTHER_CHECK
 */
int32_t xrm::system::allocChanClientFromCu(
    cuData* cu, cuProperty* cuProp, cuResource* cuRes, cuResourceLogicGroupInfo* lgInfo,
    uint64_t policyInfo, cuData** preCu) {
    uint64_t clientId = cuProp->clientId;
    pid_t clientProcessId = cuProp->clientProcessId;
    int32_t requestLoadUnified = cuProp->requestLoadUnified;
    int32_t requestLoadOriginal = cuProp->requestLoadOriginal;
    int32_t reserveIdx = 0;
    uint64_t reservePoolId = 0;
    int32_t chanId;

    if ((cu == NULL) || (cuProp == NULL) || (cuRes == NULL)) return (XRM_ERROR_INVALID);
    if (requestLoadUnified <= 0) return (XRM_ERROR_INVALID);

    reservePoolId = cuProp->poolId;
    if (reservePoolId) {
        /*
         * To allocate resource from reserve pool
         */
        reserveIdx = isReservePoolUsingCu(cu, reservePoolId);
        if (reserveIdx == -1) {
            return (XRM_ERROR_NO_KERNEL);
        }
        if (!cu->reserves[reserveIdx].clientIsActive) {
            return (XRM_ERROR_NO_KERNEL);
        }
        if (cu->reserves[reserveIdx].reserveUsedLoadUnified + requestLoadUnified >
            cu->reserves[reserveIdx].reserveLoadUnified) {
            return (XRM_ERROR_NO_KERNEL);
        }

        /*
         * Condition of best case (most used case) for target policy, so no need do further check;
         * if no policy, this cu also meets the requirement.
         *
         * Then it will allocate one channel from this cu.
         */
        if ((policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_CU_MOST_USED_FIRST ||
             policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_DEV_MOST_USED_FIRST) &&
            cu->reserves[reserveIdx].reserveUsedLoadUnified + requestLoadUnified !=
                cu->reserves[reserveIdx].reserveLoadUnified) { // most used case
            /*
             * current cu is better then previous one, update record, but will need further check
             * to see if there is better cu later. Real allocation will be done later but not here. 
             */
            if (*preCu == NULL || (cu->reserves[reserveIdx].reserveUsedLoadUnified + requestLoadUnified) >
                                      (*preCu)->reserves[reserveIdx].reserveUsedLoadUnified) {
                *preCu = cu;
            }
            return XRM_FURTHER_CHECK;
        }
        /*
         * Condition of best case (least used case) for target policy, so no need do further check;
         * if no policy, this cu also meets the requirement.
         *
         * Then it will allocate one channel from this cu.
         */
        if ((policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_CU_LEAST_USED_FIRST ||
             policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_DEV_LEAST_USED_FIRST) &&
            cu->reserves[reserveIdx].reserveUsedLoadUnified != 0) { // least used case
            /*
             * current cu is better then previous one, update record, but will need further check
             * to see if there is better cu later. Real allocation will be done later but not here. 
             */
            if (*preCu == NULL || (cu->reserves[reserveIdx].reserveUsedLoadUnified + requestLoadUnified) <
                                      (*preCu)->reserves[reserveIdx].reserveUsedLoadUnified) {
                *preCu = cu;
            }
            return XRM_FURTHER_CHECK;
        }
    } else {
        /*
         * To allocate resource from default (not reserve) pool
         *
         * Similar way as above to handle the allocate policy.
         */
        if (cu->totalUsedLoadUnified + requestLoadUnified > XRM_MAX_CHAN_LOAD_GRANULARITY_1000000) {
            return (XRM_ERROR_NO_KERNEL);
        }
        if ((policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_CU_MOST_USED_FIRST ||
             policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_DEV_MOST_USED_FIRST) &&
            cu->totalUsedLoadUnified + requestLoadUnified != XRM_MAX_CHAN_LOAD_GRANULARITY_1000000) {
            if (*preCu == NULL || (cu->totalUsedLoadUnified > (*preCu)->totalUsedLoadUnified)) {
                *preCu = cu;
            }
            return XRM_FURTHER_CHECK;
        }
        if ((policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_CU_LEAST_USED_FIRST ||
             policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_DEV_LEAST_USED_FIRST) &&
            cu->totalUsedLoadUnified != 0) {
            if (*preCu == NULL || (cu->totalUsedLoadUnified < (*preCu)->totalUsedLoadUnified)) {
                *preCu = cu;
            }
            return XRM_FURTHER_CHECK;
        }
    }
    /*
     * No need to do further check based on the policy result.
     *
     * Continue the the real allcation.
     */
    if (cu->numChanInuse == 0) { /* unused kernel */
        chanId = 0;
        cu->channels[chanId].clientId = clientId;
        cu->channels[chanId].clientProcessId = clientProcessId;
        cu->channels[chanId].channelLoadUnified = requestLoadUnified;
        cu->channels[chanId].channelLoadOriginal = requestLoadOriginal;
        uint64_t allocServiceId = getNextAllocServiceId();
        cu->channels[chanId].allocServiceId = allocServiceId;
        cu->channels[chanId].poolId = reservePoolId;
        if (reservePoolId) {
            // Allocate resource from reservation pool
            cu->reserves[reserveIdx].reserveUsedLoadUnified += requestLoadUnified;
            if ((lgInfo != NULL) && (lgInfo->indicator != XRM_LOGIC_GROUP_INFO_INVALID)) {
                // input logic group information is valid
                /* set the logic group infor for allocated resource */
                copyLogicGroupInfor(&(cuRes->logicGroupInfo), lgInfo);
                /* record the logic group infor in alloc chan */
                copyLogicGroupInfor(&(cu->channels[chanId].logicGroupInfo), lgInfo);
            } else {
                // input logic group information is invalid, so check that from reservation pool
                if (cu->reserves[reserveIdx].logicGroupInfo.indicator != XRM_LOGIC_GROUP_INFO_INVALID) {
                    /*
                     * The logic group information in reservation pool is valid,
                     * pass the logic group information of reservation.
                     */
                    /* set the logic group infor for allocated resource */
                    copyLogicGroupInfor(&(cuRes->logicGroupInfo), &(cu->reserves[reserveIdx].logicGroupInfo));
                    /* record the logic group infor in alloc chan */
                    copyLogicGroupInfor(&(cu->channels[chanId].logicGroupInfo),
                        &(cu->reserves[reserveIdx].logicGroupInfo));
                }
                /* All other case, the logic group information is set as invalid by default. */
            }
        } else {
            // Allocate resource from default pool (not reservation pool)
            cu->totalUsedLoadUnified += requestLoadUnified;
            /*
             * To record the logic group information based on allocation request input.
             */
            if (lgInfo != NULL) {
                /* set the logic group infor for allocated resource */
                copyLogicGroupInfor(&(cuRes->logicGroupInfo), lgInfo);
                /* record the logic group infor in alloc chan */
                copyLogicGroupInfor(&(cu->channels[chanId].logicGroupInfo), lgInfo);
            }
            updateDeviceLoad(cu->deviceId, requestLoadUnified, -1);
        }
        cu->numChanInuse++;
        /* Update cu->clients[], no empty slot in it */
        addClientToCu(cu, clientId);

        strncpy(cuRes->kernelName, cu->kernelName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->instanceName, cu->instanceName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->kernelAlias, cu->kernelAlias.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->cuName, cu->cuName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->kernelPluginFileName, cu->kernelPluginFileName.c_str(), XRM_MAX_NAME_LEN - 1);
        cuRes->cuType = cu->cuType;
        cuRes->channelId = chanId;
        cuRes->clientId = clientId;
        cuRes->channelLoadUnified = requestLoadUnified;
        cuRes->channelLoadOriginal = requestLoadOriginal;
        cuRes->allocServiceId = allocServiceId;
        cuRes->baseAddr = cu->baseAddr;
        cuRes->membankId = cu->membankId;
        cuRes->membankType = cu->membankType;
        cuRes->membankSize = cu->membankSize;
        cuRes->membankBaseAddr = cu->membankBaseAddr;
        cuRes->poolId = reservePoolId;

        return (XRM_SUCCESS);
    } else if (cu->numChanInuse < XRM_MAX_KERNEL_CHANNELS) { /* not maxed the space */
                                                             /* kernel can support request load */
        chanId = cuFindFreeChannelId(cu);
        if (chanId < 0 || chanId > XRM_MAX_KERNEL_CHANNELS) {
            /* no free channel */
            return (XRM_ERROR_NO_KERNEL);
        }

        cu->channels[chanId].clientId = clientId;
        cu->channels[chanId].clientProcessId = clientProcessId;
        cu->channels[chanId].channelLoadUnified = requestLoadUnified;
        cu->channels[chanId].channelLoadOriginal = requestLoadOriginal;
        uint64_t allocServiceId = getNextAllocServiceId();
        cu->channels[chanId].allocServiceId = allocServiceId;
        cu->channels[chanId].poolId = reservePoolId;
        if (reservePoolId) {
            // Allocate resource from reservation pool
            cu->reserves[reserveIdx].reserveUsedLoadUnified += requestLoadUnified;
            if ((lgInfo != NULL) && (lgInfo->indicator != XRM_LOGIC_GROUP_INFO_INVALID)) {
                // input logic group information is valid
                /* set the logic group infor for allocated resource */
                copyLogicGroupInfor(&(cuRes->logicGroupInfo), lgInfo);
                /* record the logic group infor in alloc chan */
                copyLogicGroupInfor(&(cu->channels[chanId].logicGroupInfo), lgInfo);
            } else {
                // input logic group information is invalid, so check that from reservation pool
                if (cu->reserves[reserveIdx].logicGroupInfo.indicator != XRM_LOGIC_GROUP_INFO_INVALID) {
                    /*
                     * The logic group information in reservation pool is valid,
                     * pass the logic group information of reservation.
                     */
                    /* set the logic group infor for allocated resource */
                    copyLogicGroupInfor(&(cuRes->logicGroupInfo), &(cu->reserves[reserveIdx].logicGroupInfo));
                    /* record the logic group infor in alloc chan */
                    copyLogicGroupInfor(&(cu->channels[chanId].logicGroupInfo),
                        &(cu->reserves[reserveIdx].logicGroupInfo));
                }
                /* All other case, the logic group information is set as invalid by default. */
            }
        } else {
            // Allocate resource from default pool (not reservation pool)
            cu->totalUsedLoadUnified += requestLoadUnified;
            /*
             * To record the logic group information based on allocation request input.
             */
            if (lgInfo != NULL) {
                /* set the logic group infor for allocated resource */
                copyLogicGroupInfor(&(cuRes->logicGroupInfo), lgInfo);
                /* record the logic group infor in alloc chan */
                copyLogicGroupInfor(&(cu->channels[chanId].logicGroupInfo), lgInfo);
            }
            updateDeviceLoad(cu->deviceId, requestLoadUnified, -1);
        }
        cu->numChanInuse++;
        /* Update cu->clients[], no empty slot in it */
        addClientToCu(cu, clientId);

        strncpy(cuRes->kernelName, cu->kernelName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->instanceName, cu->instanceName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->kernelAlias, cu->kernelAlias.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->cuName, cu->cuName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->kernelPluginFileName, cu->kernelPluginFileName.c_str(), XRM_MAX_NAME_LEN - 1);
        cuRes->cuType = cu->cuType;
        cuRes->channelId = chanId;
        cuRes->clientId = clientId;
        cuRes->channelLoadUnified = requestLoadUnified;
        cuRes->channelLoadOriginal = requestLoadOriginal;
        cuRes->allocServiceId = allocServiceId;
        cuRes->baseAddr = cu->baseAddr;
        cuRes->membankId = cu->membankId;
        cuRes->membankType = cu->membankType;
        cuRes->membankSize = cu->membankSize;
        cuRes->membankBaseAddr = cu->membankBaseAddr;
        cuRes->poolId = reservePoolId;

        return (XRM_SUCCESS);
    }

    /* no free channel */
    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Add client id at the end of cu->clients[] list. One client uses multiple
 * channel will only be recorded with one slot in cu->clients[], no reference
 * count recorded. So for remove, the caller need to make sure there is no other
 * channels are still used by this client.
 *
 * There is no empty slot from clients[0] to clients[numClient - 1]
 */
void xrm::system::addClientToCu(cuData* cu, uint64_t clientId) {
    int32_t i = -1;

    if (cu == NULL) return;

    i = isClientUsingCu(cu, clientId);
    if (i >= 0) return;

    /* advance clients pointer to next empty slot */
    for (i = 0; cu->clients[i]; i++)
        ;

    cu->clients[i] = clientId;
    cu->numClient++;

    return;
}

/*
 * check client is using the cu or not
 *
 * There is no empty slot from clients[0] to clients[numClient - 1]
 *
 * -1: not using
 * otherwise: return the position of client (clients[i] is using it)
 */
int32_t xrm::system::isClientUsingCu(cuData* cu, uint64_t clientId) {
    int32_t i;

    for (i = 0; i < cu->numClient && i < XRM_MAX_KERNEL_CHANNELS; i++) {
        if (cu->clients[i] != clientId) continue;

        return (i);
    }
    return (-1);
}

/*
 * To release client on specified device
 *
 * The dev->clientProcs ref count is updated.
 */
int32_t xrm::system::releaseClientOnDev(int32_t devId, uint64_t clientId) {
    if (devId < 0 || devId >= m_numDevice) return (XRM_ERROR_INVALID);
    deviceData* deviceList = m_devList;

    if (!deviceList[devId].isLoaded) return (XRM_ERROR_NO_DEV);

    if (deviceList[devId].isExcl) {
        if (deviceList[devId].clientProcs[0].clientId != clientId) {
            return (XRM_ERROR_INVALID);
        }
        deviceList[devId].clientProcs[0].ref--;
        if (deviceList[devId].clientProcs[0].ref == 0) {
            deviceList[devId].isExcl = false;
            deviceList[devId].clientProcs[0].clientId = 0;
            deviceList[devId].clientProcs[0].clientProcessId = 0;
        }
        return (XRM_SUCCESS);
    } else {
        for (int32_t pidIdx = 0; pidIdx < XRM_MAX_DEV_CLIENTS; pidIdx++)
            if (deviceList[devId].clientProcs[pidIdx].clientId == clientId) {
                deviceList[devId].clientProcs[pidIdx].ref--;
                if (deviceList[devId].clientProcs[pidIdx].ref == 0) {
                    deviceList[devId].clientProcs[pidIdx].clientId = 0;
                    deviceList[devId].clientProcs[pidIdx].clientProcessId = 0;
                }
                return (XRM_SUCCESS);
            }
    }
    return (XRM_ERROR_INVALID);
}

/*
 * Release all cu channels on specified device with specified client id
 * The cu->channels and cu->clients are updated
 */
void xrm::system::releaseAllCuChanClientOnDev(deviceData* dev, uint64_t clientId) {
    int32_t i, j;

    if (dev == NULL) return;

    for (i = 0; i < XRM_MAX_XILINX_KERNELS && i < dev->xclbinInfo.numCu; i++) {
        cuData* cu = &dev->xclbinInfo.cuList[i];

        /*
         * for each CU:
         * 1) resource NOT allocated from reserve pool: return to Big pool (default pool)
         * 2) resource allocated from reserve pool:
         *    2.1) reserve pool is active: return to reserver pool
         *    2.2) reserve pool is in-active: return to Big pool (default pool)
         */

        /* Determine if client is using this cu */
        if (clientId && (isClientUsingCu(cu, clientId) < 0)) continue;

        /* Update cu->clients, no empty slot in cu->clients */
        removeClientOnCu(cu, clientId);

        /* clear the channel entries, there maybe empty slot in cu->channels */
        for (j = 0; j < XRM_MAX_KERNEL_CHANNELS; j++) {
            uint64_t cu_client = cu->channels[j].clientId;

            /* clientId is 0, the channelLoadUnified are also 0 */
            if (!cu_client || cu_client != clientId) continue;

            /*
             * if not from reserve pool, then return to Big pool
             * if from reserve pool, pool is active, return to pool
             * if from reserve pool, pool is in-active, return to Big pool
             */
            if (cu->channels[j].poolId == 0) {
                /* Not allocated from reserve pool, then return to Big pool */
                cu->totalUsedLoadUnified -= cu->channels[j].channelLoadUnified;
                updateDeviceLoad(cu->deviceId, -cu->channels[j].channelLoadUnified, -1);
                cu->numChanInuse--;
                cu->channels[j].clientId = 0;
                cu->channels[j].clientProcessId = 0;
                cu->channels[j].channelLoadUnified = 0;
                cu->channels[j].channelLoadOriginal = 0;
                cu->channels[j].allocServiceId = 0;
                cu->channels[j].poolId = 0;
            } else {
                /* Allocated from reserve pool */
                int32_t reservePoolIdx = -1;
                for (int32_t reserveIdx = 0; reserveIdx < cu->numReserve; reserveIdx++) {
                    if (!cu->reserves[reserveIdx].clientIsActive) continue;
                    if (cu->reserves[reserveIdx].reservePoolId == cu->channels[j].poolId) {
                        reservePoolIdx = reserveIdx;
                        break;
                    }
                }

                if (reservePoolIdx != -1) {
                    /* from reserve pool, reserve pool is active, return to reserve pool */
                    cu->reserves[reservePoolIdx].reserveUsedLoadUnified -= cu->channels[j].channelLoadUnified;
                    cu->numChanInuse--;
                    cu->channels[j].clientId = 0;
                    cu->channels[j].clientProcessId = 0;
                    cu->channels[j].channelLoadUnified = 0;
                    cu->channels[j].channelLoadOriginal = 0;
                    cu->channels[j].allocServiceId = 0;
                    cu->channels[j].poolId = 0;
                } else {
                    /* from reserve pool, reserve pool is in-active, return to Big pool */
                    cu->totalUsedLoadUnified -= cu->channels[j].channelLoadUnified;
                    updateDeviceLoad(cu->deviceId, -cu->channels[j].channelLoadUnified, -1);
                    cu->numChanInuse--;
                    cu->channels[j].clientId = 0;
                    cu->channels[j].clientProcessId = 0;
                    cu->channels[j].channelLoadUnified = 0;
                    cu->channels[j].channelLoadOriginal = 0;
                    cu->channels[j].allocServiceId = 0;
                    cu->channels[j].poolId = 0;
                }
            } // end from reserve pool
        }     /* end channel clearing loop */
    }         /* end cu loop */
}

/*
 * Check whether client is using the cu or not, if yes, remove it from
 * cu->clients[] list and fill the hole. One client uses multiple
 * channel will only be recorded with one slot in cu->clients[], no reference
 * count recorded. So for remove, the caller need to make sure there is no other
 * channels are still used by this client.
 *
 * There is no empty slot from clients[0] to clients[numClient - 1]
 */
void xrm::system::removeClientOnCu(cuData* cu, uint64_t clientId) {
    if (cu == NULL) return;

    int32_t i = isClientUsingCu(cu, clientId);

    if (i < 0) {
        return;
    }

    /* Remove client and defragment list */
    for (; i < cu->numClient - 1 && i < XRM_MAX_KERNEL_CHANNELS - 1; i++) cu->clients[i] = cu->clients[i + 1];

    /* Zero the last item on the old list after defrag */
    cu->clients[i] = 0;
    cu->numClient--;

    return;
}

/*
 * check cu stat
 *
 * -1: not existing cu
 * otherwise: store the stat of the cu
 */
int32_t xrm::system::checkCuStat(cuResource* cuRes, cuStatus* cuStat) {
    if (cuRes == NULL || cuStat == NULL) return (XRM_ERROR_INVALID);

    int32_t devId, cuId;
    deviceData* dev;
    cuData* cu;

    devId = cuRes->deviceId;
    if (devId < 0 || devId > m_numDevice) return (XRM_ERROR_INVALID);

    cuId = cuRes->cuId;
    if (cuId < 0 || cuId > XRM_MAX_XILINX_KERNELS) return (XRM_ERROR_INVALID);

    dev = &m_devList[devId];
    cu = &dev->xclbinInfo.cuList[cuId];

    /* TODO:: need to get xrt/xma support to get the hardware stat */
    if (cu->totalUsedLoadUnified > 0)
        cuStat->isBusy = true;
    else
        cuStat->isBusy = false;
    cuStat->usedLoadUnified = cu->totalUsedLoadUnified;
    cuStat->usedLoadOriginal = (cu->totalUsedLoadUnified << 8);

    return (XRM_SUCCESS);
}

/*
 * increase number of concurrent client
 */
bool xrm::system::incNumConcurrentClient() {
    if (m_numConcurrentClient < m_limitConcurrentClient) {
        m_numConcurrentClient++;
        return (true);
    } else {
        logMsg(XRM_LOG_DEBUG, "Reach limit of concurrent client (%d)\n", m_limitConcurrentClient);
        return (false);
    }
}

/*
 * decrease number of concurrent client
 */
bool xrm::system::decNumConcurrentClient() {
    if (m_numConcurrentClient > 0) {
        m_numConcurrentClient--;
        return (true);
    } else {
        return (false);
    }
}

/*
 * get number of concurrent client
 *
 * Need to be proected by lock.
 */
uint32_t xrm::system::getNumConcurrentClient() {
    uint32_t numConcurrentClient;
    numConcurrentClient = m_numConcurrentClient;
    return (numConcurrentClient);
}

/*
 * get next allocation service id (0x1 - 0xFFFFFFFFFFFFFFFF)
 *
 * return: next allocation service id
 */
uint64_t xrm::system::getNextAllocServiceId() {
    if (m_allocServiceId == 0xFFFFFFFFFFFFFFFF)
        return (1);
    else
        return (m_allocServiceId + 1);
}

/*
 * update current allocation service id(0x1 - 0xFFFFFFFFFFFFFFFF)
 *
 */
void xrm::system::updateAllocServiceId() {
    if (m_allocServiceId == 0xFFFFFFFFFFFFFFFF)
        m_allocServiceId = 1;
    else
        m_allocServiceId++;
}

/*
 * Query allocated cu resource from system based on the query property.
 *
 * XRM_SUCCESS: queried resouces are filled into cuListRes
 * Otherwise: failed to query the allocated resource
 *
 * Lock: should enter lock during the resource allocation query
 */
int32_t xrm::system::resAllocationQuery(allocationQueryInfo* allocQuery, cuListResource* cuListRes) {
    int32_t devId, cuId, chanId;
    uint64_t allocServiceId;
    bool kernelNameEqual, kernelAliasEqual;
    deviceData* dev = NULL;
    cuData* cu = NULL;
    channelData* chan = NULL;
    cuResource* cuRes = NULL;
    int32_t cuNum;

    if (allocQuery == NULL || cuListRes == NULL) {
        return (XRM_ERROR_INVALID);
    }
    allocServiceId = allocQuery->allocServiceId;
    if (allocServiceId == 0) {
        return (XRM_ERROR_INVALID);
    }

    cuNum = 0;
    memset(cuListRes, 0, sizeof(cuListResource));
    /* Check all the devices */
    for (devId = 0; devId < m_numDevice; devId++) {
        dev = &m_devList[devId];
        /* Check all the cu on this device */
        for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];
            /* kernel name is presented, compare it; otherwise no need to compare */
            kernelNameEqual = true;
            if (allocQuery->kernelName[0] != '\0') {
                if (cu->kernelName.compare(allocQuery->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (allocQuery->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(allocQuery->kernelAlias)) kernelAliasEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual)) continue;
            /* Check all the channels on this cu */
            for (chanId = 0; chanId < XRM_MAX_KERNEL_CHANNELS; chanId++) {
                chan = &cu->channels[chanId];
                if (chan->allocServiceId == allocServiceId) {
                    /* out of cu list limitation */
                    if (cuNum >= XRM_MAX_LIST_CU_NUM) {
                        memset(cuListRes, 0, sizeof(cuListResource));
                        return (XRM_ERROR);
                    }
                    cuRes = &cuListRes->cuResources[cuNum];
                    strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelPluginFileName, cu->kernelPluginFileName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelName, cu->kernelName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->instanceName, cu->instanceName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelAlias, cu->kernelAlias.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->cuName, cu->cuName.c_str(), XRM_MAX_NAME_LEN - 1);
                    cuRes->cuType = cu->cuType;
                    cuRes->baseAddr = cu->baseAddr;
                    cuRes->membankId = cu->membankId;
                    cuRes->membankType = cu->membankType;
                    cuRes->membankSize = cu->membankSize;
                    cuRes->membankBaseAddr = cu->membankBaseAddr;
                    cuRes->deviceId = devId;
                    cuRes->deviceNumBDF = dev->devNumBDF;
                    cuRes->cuId = cuId;
                    cuRes->channelId = chanId;
                    cuRes->allocServiceId = allocServiceId;
                    cuRes->poolId = chan->poolId;
                    cuRes->clientId = chan->clientId;
                    cuRes->channelLoadUnified = chan->channelLoadUnified;
                    cuRes->channelLoadOriginal = chan->channelLoadOriginal;
                    cuNum++;
                }
            }
        }
    }
    cuListRes->cuNum = cuNum;
    return (XRM_SUCCESS);
}

/*
 * Query allocated cu resource from system based on the query property.
 *
 * XRM_SUCCESS: queried resouces are filled into cuListRes
 * Otherwise: failed to query the allocated resource
 *
 * Lock: should enter lock during the resource allocation query
 */
int32_t xrm::system::resAllocationQueryV2(allocationQueryInfoV2* allocQueryV2, cuListResourceV2* cuListResV2) {
    int32_t devId, cuId, chanId;
    uint64_t allocServiceId;
    bool kernelNameEqual, kernelAliasEqual;
    deviceData* dev = NULL;
    cuData* cu = NULL;
    channelData* chan = NULL;
    cuResource* cuRes = NULL;
    int32_t cuNum;

    if (allocQueryV2 == NULL || cuListResV2 == NULL) {
        return (XRM_ERROR_INVALID);
    }
    allocServiceId = allocQueryV2->allocServiceId;
    if (allocServiceId == 0) {
        return (XRM_ERROR_INVALID);
    }

    cuNum = 0;
    memset(cuListResV2, 0, sizeof(cuListResourceV2));
    /* Check all the devices */
    for (devId = 0; devId < m_numDevice; devId++) {
        dev = &m_devList[devId];
        /* Check all the cu on this device */
        for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];
            /* kernel name is presented, compare it; otherwise no need to compare */
            kernelNameEqual = true;
            if (allocQueryV2->kernelName[0] != '\0') {
                if (cu->kernelName.compare(allocQueryV2->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (allocQueryV2->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(allocQueryV2->kernelAlias)) kernelAliasEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual)) continue;
            /* Check all the channels on this cu */
            for (chanId = 0; chanId < XRM_MAX_KERNEL_CHANNELS; chanId++) {
                chan = &cu->channels[chanId];
                if (chan->allocServiceId == allocServiceId) {
                    /* out of cu list limitation */
                    if (cuNum >= XRM_MAX_LIST_CU_NUM_V2) {
                        memset(cuListResV2, 0, sizeof(cuListResource));
                        return (XRM_ERROR);
                    }
                    cuRes = &cuListResV2->cuResources[cuNum];
                    strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelPluginFileName, cu->kernelPluginFileName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelName, cu->kernelName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->instanceName, cu->instanceName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelAlias, cu->kernelAlias.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->cuName, cu->cuName.c_str(), XRM_MAX_NAME_LEN - 1);
                    cuRes->cuType = cu->cuType;
                    cuRes->baseAddr = cu->baseAddr;
                    cuRes->membankId = cu->membankId;
                    cuRes->membankType = cu->membankType;
                    cuRes->membankSize = cu->membankSize;
                    cuRes->membankBaseAddr = cu->membankBaseAddr;
                    cuRes->deviceId = devId;
                    cuRes->deviceNumBDF = dev->devNumBDF;
                    cuRes->cuId = cuId;
                    cuRes->channelId = chanId;
                    cuRes->allocServiceId = allocServiceId;
                    cuRes->poolId = chan->poolId;
                    cuRes->clientId = chan->clientId;
                    cuRes->channelLoadUnified = chan->channelLoadUnified;
                    cuRes->channelLoadOriginal = chan->channelLoadOriginal;
                    cuNum++;
                }
            }
        }
    }
    cuListResV2->cuNum = cuNum;
    return (XRM_SUCCESS);
}

/*
 * Recycle all resource from client whose connection is broken.
 */
void xrm::system::recycleResource(uint64_t clientId) {
    deviceData* dev = NULL;
    cuData* cu = NULL;
    int32_t devId;
    int64_t tmp_sum = 0;

    /* Check all the devices */
    for (devId = 0; devId < m_numDevice; devId++) {
        dev = &m_devList[devId];
        if (!dev->isLoaded) continue;

        /* relinquish all reserved resource from the client */
        /*
         * for each CU:
         * 1) no resource still allocated from pool of this client: reserved resource back to
         *    Big pool (default pool), de-active the client and remove reserve pool
         * 2) resource allocated from pool of this client: reserved but not allocated back
         *    to Big pool (default pool), de-active the client and remove reserve pool
         */
        for (int32_t cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];
            for (int32_t reserveIdx = 0; reserveIdx < cu->numReserve; reserveIdx++) {
                if (!cu->reserves[reserveIdx].clientIsActive) continue;
                if (cu->reserves[reserveIdx].clientId == clientId) {
                    int64_t tmp =
                        cu->reserves[reserveIdx].reserveLoadUnified - cu->reserves[reserveIdx].reserveUsedLoadUnified;
                    cu->totalUsedLoadUnified -= tmp;
                    tmp_sum -= tmp;
                    cu->totalReservedLoadUnified -= cu->reserves[reserveIdx].reserveLoadUnified;
                    /*
                     * To de-active the client and remove reserve pool.
                     * The allocated resource from this reserve pool will be directly return
                     * to Big pool in future (either during release or recycle) since the pool is de-active.
                     */
                    cu->reserves[reserveIdx].clientIsActive = false;
                    cu->reserves[reserveIdx].reserveLoadUnified = 0;

                    /* switch the slot until the last one */
                    int32_t i;
                    for (i = reserveIdx; i < cu->numReserve - 1; i++) {
                        cu->reserves[i].reserveLoadUnified = cu->reserves[i + 1].reserveLoadUnified;
                        cu->reserves[i].reserveUsedLoadUnified = cu->reserves[i + 1].reserveUsedLoadUnified;
                        cu->reserves[i].reservePoolId = cu->reserves[i + 1].reservePoolId;
                        cu->reserves[i].clientIsActive = cu->reserves[i + 1].clientIsActive;
                        cu->reserves[i].clientId = cu->reserves[i + 1].clientId;
                        cu->reserves[i].clientProcessId = cu->reserves[i + 1].clientProcessId;
                    }
                    /* empty the last slot */
                    cu->reserves[i].reserveLoadUnified = 0;
                    cu->reserves[i].reserveUsedLoadUnified = 0;
                    cu->reserves[i].reservePoolId = 0;
                    cu->reserves[i].clientIsActive = false;
                    cu->reserves[i].clientId = 0;
                    cu->reserves[i].clientProcessId = 0;
                    cu->numReserve--;
                    /* removed one slot, so set the reserveIdx to the right one */
                    reserveIdx--;
                }
            }
        } // endif relinquish

        updateDeviceLoad(devId, tmp_sum, 0);
        /* release all allocated resource from the client */
        /*
         * for each CU:
         * 1) resource NOT allocated from reserve pool: return to Big pool (default pool)
         * 2) resource allocated from reserve pool:
         *    2.1) reserve pool is active: return to reserver pool
         *    2.2) reserve pool is in-active: return to Big pool (default pool)
         */
        if (dev->isExcl) {
            if (dev->clientProcs[0].ref == 0) break;
            /* recycle the resource from the client */
            if (dev->clientProcs[0].clientId == clientId) {
                releaseAllCuChanClientOnDev(dev, clientId);
                dev->isExcl = false;
                memset(dev->clientProcs, 0, sizeof(clientData) * XRM_MAX_DEV_CLIENTS);
            }
        } else {
            for (int32_t i = 0; i < XRM_MAX_DEV_CLIENTS; i++) {
                if (dev->clientProcs[i].ref == 0) continue;
                /* recycle the resource from the client */
                if (dev->clientProcs[i].clientId == clientId) {
                    releaseAllCuChanClientOnDev(dev, clientId);
                    dev->clientProcs[i].clientId = 0;
                    dev->clientProcs[i].clientProcessId = 0;
                    dev->clientProcs[i].ref = 0;
                }
            }
        } // end of release
    }

    decNumConcurrentClient();
    /* The save() function is time cost operation, so decide to not it here. */
    // save();
}

/* the implementation of resource reserve and relinquish */

/*
 * get next reserve pool id (0x1 - 0xFFFFFFFFFFFFFFFF)
 *
 * return: next reservation id
 */
uint64_t xrm::system::getNextReservePoolId() {
    if (m_reservePoolId == 0xFFFFFFFFFFFFFFFF)
        return (1);
    else
        return (m_reservePoolId + 1);
}

/*
 * update current reservation id(0x1 - 0xFFFFFFFFFFFFFFFF)
 *
 */
void xrm::system::updateReservePoolId() {
    if (m_reservePoolId == 0xFFFFFFFFFFFFFFFF)
        m_reservePoolId = 1;
    else
        m_reservePoolId++;
}

/*
 * check reservation pool id is using the cu or not
 *
 * There is no empty slot from reserves[0] to reserves[numReserve - 1]
 *
 * The input reservePoolId > 0
 *
 * -1: not using
 * otherwise: return the index of reserve (reserves[i] is using it)
 */
int32_t xrm::system::isReservePoolUsingCu(cuData* cu, uint64_t reservePoolId) {
    int32_t i;

    if (reservePoolId < 1) return (-1);
    for (i = 0; i < cu->numReserve && i < XRM_MAX_KERNEL_RESERVES; i++) {
        if (cu->reserves[i].reservePoolId != reservePoolId) continue;

        return (i);
    }
    return (-1);
}

/*
 * call while holding system lock
 *
 * Reserve request load from on specified cu, update cu resource pool.
 *
 * If input lgInfor is set, then record the logic group information for reserved resource pool.
 *
 */
int32_t xrm::system::reserveLoadFromCu(uint64_t reservePoolId, cuData* cu, cuProperty* cuProp,
    cuResourceLogicGroupInfo* lgInfo) {
    int32_t requestLoadUnified = cuProp->requestLoadUnified;
    int32_t reserveIdx;

    if ((cu == NULL) || (cuProp == NULL)) return (XRM_ERROR_INVALID);
    if (reservePoolId < 1) return (XRM_ERROR_INVALID);
    if (requestLoadUnified <= 0 || requestLoadUnified > XRM_MAX_CHAN_LOAD_GRANULARITY_1000000)
        return (XRM_ERROR_INVALID);

    if (((cu->totalUsedLoadUnified + requestLoadUnified) <= XRM_MAX_CHAN_LOAD_GRANULARITY_1000000) &&
        ((cu->totalReservedLoadUnified + requestLoadUnified) <= XRM_MAX_CHAN_LOAD_GRANULARITY_1000000)) {
        reserveIdx = isReservePoolUsingCu(cu, reservePoolId);
        if (reserveIdx != -1) {
            /* in use, update the existing reserve slot to record the information */
            cu->totalUsedLoadUnified += requestLoadUnified;
            updateDeviceLoad(cu->deviceId, requestLoadUnified, -1);
            cu->totalReservedLoadUnified += requestLoadUnified;
            cu->reserves[reserveIdx].reserveLoadUnified += requestLoadUnified;
            // If input lgInfo is set, then record the logic group information for reserved resource pool.
            if (lgInfo != NULL) {
                copyLogicGroupInfor(&(cu->reserves[reserveIdx].logicGroupInfo), lgInfo);
            }
            return (XRM_SUCCESS);
        } else if (cu->numReserve < XRM_MAX_KERNEL_RESERVES) {
            /* not in use, get a new reserve slot to record the information */
            cu->totalUsedLoadUnified += requestLoadUnified;
            updateDeviceLoad(cu->deviceId, requestLoadUnified, -1);
            cu->totalReservedLoadUnified += requestLoadUnified;
            cu->reserves[cu->numReserve].reserveLoadUnified = requestLoadUnified;
            cu->reserves[cu->numReserve].reservePoolId = reservePoolId;
            cu->reserves[cu->numReserve].clientIsActive = true;
            cu->reserves[cu->numReserve].clientId = cuProp->clientId;
            cu->reserves[cu->numReserve].clientProcessId = cuProp->clientProcessId;
            // If input lgInfo is set, then record the logic group information for reserved resource pool.
            if (lgInfo != NULL) {
                copyLogicGroupInfor(&(cu->reserves[cu->numReserve].logicGroupInfo), lgInfo);
            }
            cu->numReserve++;
            return (XRM_SUCCESS);
        }
    }

    /* over load */
    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Reserve one CU (Compute Unit) based on the request property.
 *
 * Please be noted that the reservation will record the client information
 * (clientId) in the cu->reserves[].
 *
 * return:
 *   XRM_SUCCESS: reserve the cu resouce and associate it with the reserve pool id
 *   Otherwise: failed to reserve the cu resource with specified pool id
 *
 * Lock: should already in lock protection during the cu reservation
 */
int32_t xrm::system::resReserveCu(uint64_t reservePoolId, cuProperty* cuProp, uint64_t* fromDevId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual;
    int32_t devId, cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    if (cuProp == NULL) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "neither kernel name nor alias are presented\n");
        return (XRM_ERROR_INVALID);
    }

cu_reserve_loop:
    for (devId = -1; !cuAcquired && (devId < m_numDevice);) {
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            break;
        }

        dev = &m_devList[devId];
        *fromDevId = (uint64_t)devId;
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to re-use existing kernels; else, use a new kernel */
            if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual)) continue;
            /* Reserve request load from the specified cu */
            ret = reserveLoadFromCu(reservePoolId, cu, cuProp);
            if (ret != XRM_SUCCESS) {
                continue;
            }

            cuAcquired = true;
        }
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_reserve_loop;
    }

    if (cuAcquired) {
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Reserve one CU (Compute Unit) from specified device(devId) based on the request property.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 * return:
 *   XRM_SUCCESS: reserve the cu resouce and associate it with the reserve pool id
 *   Otherwise: failed to reserve the cu resource with specified reserve pool id
 *
 * Lock: should already in lock protection during the cu reservation
 */
int32_t xrm::system::reserveCuFromDev(uint64_t reservePoolId, int32_t devId, cuProperty* cuProp) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual;
    int32_t cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    if (devId < 0 || devId >= m_numDevice) {
        return (XRM_ERROR_INVALID);
    }
    if (cuProp == NULL) {
        logMsg(XRM_LOG_ERROR, "cuProp or cuRes is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "neither kernel name nor alias are presented\n");
        return (XRM_ERROR_INVALID);
    }

    dev = &m_devList[devId];
cu_reserve_again:
    if (!cuAcquired) {
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to re-use existing kernels; else, use a new kernel */
            if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual)) continue;
            /* Reserve request load from the specified cu */
            ret = reserveLoadFromCu(reservePoolId, cu, cuProp);
            if (ret != XRM_SUCCESS) {
                continue;
            }
            cuAcquired = true;
        }
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_reserve_again;
    }

    if (cuAcquired) {
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Reserve one cu list based on the request property.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 * return:
 *   XRM_SUCCESS: reserve the cu list and associate it with the reserve pool id
 *   Otherwise: failed to reserve the cu list with specified reserve pool id
 *
 * Lock: should already in lock protection during the cu list reservation.
 */
int32_t xrm::system::resReserveCuList(uint64_t reservePoolId, cuListProperty* cuListProp) {
    int32_t i, devId, ret;
    cuProperty* cuProp = NULL;

    if (cuListProp == NULL) {
        return (XRM_ERROR_INVALID);
    }
    if (cuListProp->cuNum <= 0 || cuListProp->cuNum > XRM_MAX_LIST_CU_NUM) {
        return (XRM_ERROR_INVALID);
    }

    ret = XRM_ERROR;
    if (cuListProp->sameDevice) {
        /*
         * To reserve the cu list from same device
         */
        for (devId = -1; devId < m_numDevice;) {
            /* find one available device for first cu */
            cuProp = &cuListProp->cuProps[0];
            devId = allocDevForClient(&devId, cuProp);
            if (devId < 0) {
                return (ret);
            }
            /* reserve the cu list from this device */
            for (i = 0; i < cuListProp->cuNum; i++) {
                cuProp = &cuListProp->cuProps[i];
                ret = reserveCuFromDev(reservePoolId, devId, cuProp);
                if (ret != XRM_SUCCESS) {
                    resRelinquishCuList(reservePoolId);
                    break;
                }
            }
            if (ret == XRM_SUCCESS) {
                return (ret);
            }
        }
        return (ret);
    } else {
        /*
         * To reserve the cu list from any device
         */
        uint64_t fromDevId;
        for (i = 0; i < cuListProp->cuNum; i++) {
            cuProp = &cuListProp->cuProps[i];
            ret = resReserveCu(reservePoolId, cuProp, &fromDevId);
            if (ret != XRM_SUCCESS) {
                resRelinquishCuList(reservePoolId);
                return (ret);
            }
        }
        return (ret);
    }
}

/*
 * Allocate cu sub list from:
 *          Target device
 *          Target logic group (since it's from same device, so it's same group)
 */
int32_t xrm::system::resReserveCuSubListFromTargetDeviceTargetLogicGroup(uint64_t reservePoolId, int32_t devId,
    int32_t lgId, logicGroupConstraintSubListTypeV2* lgSubList,
    cuListPropertyV2* cuListPropV2, itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        ret = resReserveCuFromTargetDeviceTargetLogicGroup(reservePoolId, devId, lgId, cuPropV2);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
            itemFlags[cuId].devId = devId;
            cuListResInforV2->cuResInfor[cuId].deviceId = devId;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /*
         * Fail to reserve cu of this sub list, relinquish the allocated resource
         * for this sub-list based on these input and record.
         */
        resRelinquishCuLgSubListV2(reservePoolId, lgSubList, cuListPropV2, itemFlags);
    }
    return (ret);
}

/*
 * Relinquish the reserved resource for the logic group based sub-list,
 * and update the record in itemFlags which indicate resource allocated or not.
 */
void xrm::system::resRelinquishCuLgSubListV2(uint64_t reservePoolId, logicGroupConstraintSubListTypeV2* lgSubList,
    cuListPropertyV2* cuListPropV2, itemFlag* itemFlags) {
    int32_t lgCuIndex, cuId, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2;

    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request logic group based sub-list
        cuId = lgSubList->cuId[lgCuIndex];
        if (itemFlags[cuId].allocated) {
            cuPropV2 = &cuListPropV2->cuProps[cuId];
            devId = itemFlags[cuId].devId;
            cuPropertyCopyFromV2(cuProp, cuPropV2);
            relinquishCuFromDev(reservePoolId, devId, cuProp);
            itemFlags[cuId].allocated = false;
        }
    }
}

/*
 * Reserve one cu (compute unit) from target device's target logic group.
 * It's used for cu list reserve, no allocate policy applied.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 * return:
 *   XRM_SUCCESS: reserve the cu resouce and associate it with the reserve pool id
 *   Otherwise: failed to reserve the cu resource with specified reserve pool id
 *
 * Lock: should already in lock protection during the cu reservation
 */
int32_t xrm::system::resReserveCuFromTargetDeviceTargetLogicGroup(uint64_t reservePoolId, int32_t devId,
    int32_t lgId, cuPropertyV2* cuPropV2) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual;
    int32_t cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;

    // no need to check cuPropV2 or cuRes as it should be checked already
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    dev = &m_devList[devId];
cu_reserve_loop:
    /*
     * Now check whether matching cu is in target device target logic group
     */
    if (lgId >= dev->xclbinInfo.numLogicGroup) {
        return (XRM_ERROR);
    }
    logicGroupData* logicGroup = &(dev->xclbinInfo.logicGroups[lgId]);
    for (int32_t lgCuIndex = 0; lgCuIndex < logicGroup->numCu && !cuAcquired; lgCuIndex++) {
        cuId = logicGroup->cuId[lgCuIndex];
        cu = &dev->xclbinInfo.cuList[cuId];

        /* first attempt to re-use existing kernels; else, use a new kernel */
        if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual)) continue;
        /* get the logic group infor for resource reservation */
        cuResourceLogicGroupInfo lgInfo;
        lgInfo.indicator = XRM_LOGIC_GROUP_INFO_ID_TYPE_VALID;
        lgInfo.groupId = lgId;
        strncpy(lgInfo.groupType, logicGroup->groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1);
        lgInfo.groupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1] = '\0';
        /* Reserve request load from the specified cu with logic group information*/
        ret = reserveLoadFromCu(reservePoolId, cu, cuProp, &lgInfo);
        if (ret != XRM_SUCCESS) {
            continue;
        }
        cuAcquired = true;
    } // cu in the logic group

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_reserve_loop;
    }

    if (cuAcquired) {
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Reserve cu sub list from:
 *          Target device
 *          Target type logic group (can be different group)
 */
int32_t xrm::system::resReserveCuSubListFromTargetDeviceTargetTypeLogicGroup(uint64_t reservePoolId,
    int32_t devId, logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];

        ret = resReserveCuFromTargetDeviceTargetTypeLogicGroup(reservePoolId, devId, cuPropV2);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
            itemFlags[cuId].devId = devId;
            cuListResInforV2->cuResInfor[cuId].deviceId = devId;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /*
         * Fail to allocate cu of this sub list, relinquish the allocated resource
         * for this sub-list based on these input and record.
         */
        resRelinquishCuLgSubListV2(reservePoolId, lgSubList, cuListPropV2, itemFlags);
    }
    return (ret);
}

/*
 * Reserve single cu from:
 *          Target device
 *          Target type logic group
 */
int32_t xrm::system::resReserveCuFromTargetDeviceTargetTypeLogicGroup(uint64_t reservePoolId, int32_t devId,
    cuPropertyV2* cuPropV2) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // find logic group index with target type match
        if (strncmp(cuPropV2->logicGroupInfo.groupType, dev->xclbinInfo.logicGroups[lgId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) == 0 )
            ret = resReserveCuFromTargetDeviceTargetLogicGroup(reservePoolId, devId, lgId, cuPropV2);
        if (ret == XRM_SUCCESS)
            break;
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Target device
 *          Same logic group, and the group is different from that in used group list
 */
int32_t xrm::system::resReserveCuSubListFromTargetDeviceSameLogicGroupNotUsed(uint64_t reservePoolId, int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList,
    cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // find logic group index (with device id pair) not in used list
        if (isInLogicGroupIdDeviceIdPairList(usedLgIdDevIdPairList, lgId, devId))
            // logic group is used, go to next one
            continue;
        ret = resReserveCuSubListFromTargetDeviceTargetLogicGroup(reservePoolId, devId, lgId, lgSubList,
            cuListPropV2, itemFlags, cuListResInforV2);
        if (ret == XRM_SUCCESS) {
            // update new logic group id / device id pair into used list
            usedLgIdDevIdPairList->pairGroupIds[usedLgIdDevIdPairList->pairNum] = lgId;
            usedLgIdDevIdPairList->pairDevIds[usedLgIdDevIdPairList->pairNum] = devId;
            usedLgIdDevIdPairList->pairNum++;
            break;
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Target device
 *          Same logic group, no limit the group to be different from that in used group list
 */
int32_t xrm::system::resReserveCuSubListFromTargetDeviceSameLogicGroupNoLimitUsedOrNot(uint64_t reservePoolId, int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // Can be any logic group
        ret = resReserveCuSubListFromTargetDeviceTargetLogicGroup(reservePoolId, devId, lgId, lgSubList,
            cuListPropV2, itemFlags, cuListResInforV2);
        if (ret == XRM_SUCCESS) {
            // no need to record logic group id / device id pair into used list
            break;
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Target device
 *          Target type logic group
 *          Same logic group, and the group is different from that in used group list
 */
int32_t xrm::system::resReserveCuSubListFromTargetDeviceTargetTypeSameLogicGroupNotUsed(uint64_t reservePoolId,
    int32_t devId, logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList,
    cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // find logic group id / device id pair not in used list
        if (isInLogicGroupIdDeviceIdPairList(usedLgIdDevIdPairList, lgId, devId))
            // logic group is used, go to next one
            continue;
        if (strncmp(lgSubList->groupType,
            dev->xclbinInfo.logicGroups[lgId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) != 0 )
            // type not match, go to next one
            continue;
        ret = resReserveCuSubListFromTargetDeviceTargetLogicGroup(reservePoolId, devId, lgId, lgSubList,
            cuListPropV2, itemFlags, cuListResInforV2);
        if (ret == XRM_SUCCESS) {
            // update new logic group id / device id pair into used list
            usedLgIdDevIdPairList->pairGroupIds[usedLgIdDevIdPairList->pairNum] = lgId;
            usedLgIdDevIdPairList->pairDevIds[usedLgIdDevIdPairList->pairNum] = devId;
            usedLgIdDevIdPairList->pairNum++;
            break;
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Target device
 *          Target type logic group
 *          Same logic group, no limit the group to be different from that in used group list
 */
int32_t xrm::system::resReserveCuSubListFromTargetDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(uint64_t reservePoolId,
    int32_t devId, logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgId;
    deviceData* dev = NULL;

    dev = &m_devList[devId];
    // loop all the logic group of this device
    ret = XRM_ERROR;
    for (lgId = 0; lgId < dev->xclbinInfo.numLogicGroup; lgId++) {
        // no limit, can be any logic group
        if (strncmp(lgSubList->groupType,
            dev->xclbinInfo.logicGroups[lgId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) != 0 )
            // type not match, go to next one
            continue;
        ret = resReserveCuSubListFromTargetDeviceTargetLogicGroup(reservePoolId, devId, lgId, lgSubList,
            cuListPropV2, itemFlags, cuListResInforV2);
        if (ret == XRM_SUCCESS) {
            // no need to record the logic group id / device id pair into used list
            break;
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Target device
 *          Any logic group
 * NOTE: there is no logic group constraint.
 */
int32_t xrm::system::resReserveCuSubListFromTargetDeviceAnyLogicGroup(uint64_t reservePoolId, int32_t devId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        ret = resReserveCuFromTargetDeviceAnyLogicGroup(reservePoolId, devId, cuPropV2);
        if (ret == XRM_SUCCESS) {
            itemFlags[cuId].allocated = true;
            itemFlags[cuId].devId = devId;
            cuListResInforV2->cuResInfor[cuId].deviceId = devId;
        } else {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /*
         * Fail to allocate cu of this sub list, relinquish the allocated resource
         * for this sub-list based on these input and record.
         */
        resRelinquishCuLgSubListV2(reservePoolId, lgSubList, cuListPropV2, itemFlags);
    }
    return (ret);
}

/*
 * Reserve one cu (compute unit) from target device's any logic group.
 * It's used for cu list allocation, no allocate policy applied.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 * return:
 *   XRM_SUCCESS: reserve the cu resouce and associate it with the reserve pool id
 *   Otherwise: failed to reserve the cu resource with specified reserve pool id
 *
 * Lock: should already in lock protection during the cu reservation
 */
int32_t xrm::system::resReserveCuFromTargetDeviceAnyLogicGroup(uint64_t reservePoolId,
        int32_t devId, cuPropertyV2* cuPropV2) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual;
    int32_t cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    dev = &m_devList[devId];
cu_reserve_again:
    if (!cuAcquired) {
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        for (cuId = 0; cuId < XRM_MAX_XILINX_KERNELS && cuId < dev->xclbinInfo.numCu && !cuAcquired; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to re-use existing kernels; else, use a new kernel */
            if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) continue;
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual)) continue;
            /*
             * Reserve request load from the specified cu, logic group information will be set
             * as invalid by default.
             */
            ret = reserveLoadFromCu(reservePoolId, cu, cuProp);
            if (ret != XRM_SUCCESS) {
                continue;
            }
            cuAcquired = true;
        }
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_reserve_again;
    }

    if (cuAcquired) {
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Reserve multiple cu sub lists from target device:
 *
 * These different cu sub lists are generated based on logic group constraint information.
 */
int32_t xrm::system::resReserveCuSubListsFromTargetDevice(uint64_t reservePoolId, int32_t devId,
    int32_t numLgConstraintSubList, logicGroupConstraintSubListTypeV2* lgConstraintSubLists,
    cuListPropertyV2* cuListPropV2, itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret;
    logicGroupConstraintSubListTypeV2* lgSubList = NULL;
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList = NULL;
    int32_t lgSubListIndex, lgSubListType, lgId;

    /*
     * use to record used logic group id/device id pair for sub-list which request from
     * different pseudo logic group index (different pair means different logic group)
     */
    usedLgIdDevIdPairList = (logicGroupIdDeviceIdPairListPropertyV2*)malloc(
        sizeof(logicGroupIdDeviceIdPairListPropertyV2));
    memset(usedLgIdDevIdPairList, 0, sizeof(logicGroupIdDeviceIdPairListPropertyV2));

    // loop all the Logic group constraint based sub-Lists
    for (lgSubListIndex = 0; lgSubListIndex < numLgConstraintSubList; lgSubListIndex++) {
        lgSubList = &(lgConstraintSubLists[lgSubListIndex]);
        lgSubListType = lgSubList->type;
        switch (lgSubListType) {
            case LG_CONSTRAINT_SUBLIST_TYPE_REAL_GROUP_ID:
                /* Reserve this logic group sub-list from target logic group of this device */
                // get the target logic group id for this sub-list
                lgId = lgSubList->groupId;
                ret = resReserveCuSubListFromTargetDeviceTargetLogicGroup(reservePoolId, devId, lgId, lgSubList,
                    cuListPropV2, itemFlags, cuListResInforV2);
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE:
                /*
                 * Reserve this logic group sub-list from target type logic group of this device.
                 * Can from different logic group.
                 */
                ret = resReserveCuSubListFromTargetDeviceTargetTypeLogicGroup(reservePoolId, devId, lgSubList,
                    cuListPropV2, itemFlags, cuListResInforV2);
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_PSEUDO_GROUP_ID:
                if (lgSubList->groupId >= 0) {
                    /*
                     * Reserve this logic group sub-list from same logic group of this device.
                     * groupId >= 0, means this logic group need to be different from previous used groups.
                     */
                    ret = resReserveCuSubListFromTargetDeviceSameLogicGroupNotUsed(reservePoolId, devId, lgSubList,
                        cuListPropV2, itemFlags, usedLgIdDevIdPairList, cuListResInforV2);
                } else {
                    /*
                     * Reserve this logic group sub-list from same logic group of this device.
                     * groupId < 0, means no limit this logic group to be different from previous used groups.
                     */
                    ret = resReserveCuSubListFromTargetDeviceSameLogicGroupNoLimitUsedOrNot(reservePoolId, devId, lgSubList,
                        cuListPropV2, itemFlags, cuListResInforV2);
                }
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE_AND_PSEUDO_GROUP_ID:
                if (lgSubList->groupId >= 0) {
                    /*
                     * Reserve this logic group sub-list from target type same logic group of this device.
                     * groupId >= 0, means this logic group need to be different from previous used groups.
                     */
                    ret = resReserveCuSubListFromTargetDeviceTargetTypeSameLogicGroupNotUsed(reservePoolId, devId, lgSubList,
                        cuListPropV2, itemFlags, usedLgIdDevIdPairList, cuListResInforV2);
                } else {
                    /*
                     * Reserve this logic group sub-list from target type same logic group of this device.
                     * groupId < 0, means no limit this logic group to be different from previous used groups.
                     */
                    ret = resReserveCuSubListFromTargetDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(reservePoolId, devId, lgSubList,
                        cuListPropV2, itemFlags, cuListResInforV2);
                }
                break;
            case DEV_CONSTRAINT_SUBLIST_TYPE_NULL:
                ret = resReserveCuSubListFromTargetDeviceAnyLogicGroup(reservePoolId, devId, lgSubList,
                    cuListPropV2, itemFlags, cuListResInforV2);
                break;
            default:
                ret = XRM_ERROR_INVALID;
                break;
        }
        if (ret != XRM_SUCCESS) {
            break;
        }
    } // loop lgConstraintSubLists

    if (ret != XRM_SUCCESS) {
        /*
         * Fail to reserve cu of this sub list, relinquish the reserved resource
         * for all sub-lists
         */
        for (lgSubListIndex = 0; lgSubListIndex < numLgConstraintSubList; lgSubListIndex++) {
            lgSubList = &(lgConstraintSubLists[lgSubListIndex]);
            resRelinquishCuLgSubListV2(reservePoolId, lgSubList, cuListPropV2, itemFlags);
        }
    }
    if (usedLgIdDevIdPairList != NULL)
        free(usedLgIdDevIdPairList);
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Any device
 *          Target logic group
 * So the cu in this sub list can from group on differernt device.
 */
int32_t xrm::system::resReserveCuSubListFromAnyDeviceTargetLogicGroup(uint64_t reservePoolId,
    int32_t lgId, logicGroupConstraintSubListTypeV2* lgSubList,
    cuListPropertyV2* cuListPropV2, itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];
        // itemFlags record is udpated in reserve call since devId is unknown here
        ret = resReserveCuFromAnyDeviceTargetLogicGroup(reservePoolId, lgId, cuPropV2, itemFlags,
            cuId, cuListResInforV2);
        if (ret != XRM_SUCCESS) {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /*
         * Fail to reserve cu of this sub list, relinquish the allocated resource
         * for this sub-list based on these input and record.
         */
        resRelinquishCuLgSubListV2(reservePoolId, lgSubList, cuListPropV2, itemFlags);
    }
    return (ret);
}

/*
 * Reserve cu from:
 *          Any device
 *          Target logic group
 * So the cu can from group on differernt device.
 */
int32_t xrm::system::resReserveCuFromAnyDeviceTargetLogicGroup(uint64_t reservePoolId, int32_t lgId,
    cuPropertyV2* cuPropV2, itemFlag* itemFlags, int32_t itemIndex, cuListResInforV2* cuListResInforV2) {
    int32_t ret = XRM_ERROR;
    int32_t devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }

        ret = resReserveCuFromTargetDeviceTargetLogicGroup(reservePoolId, devId, lgId, cuPropV2);
        if (ret == XRM_SUCCESS) {
            itemFlags[itemIndex].allocated = true;
            itemFlags[itemIndex].devId = devId;
            cuListResInforV2->cuResInfor[itemIndex].deviceId = devId;
            break;
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Any device
 *          Target type logic group (can be different group), can also be
 *          group from different device.
 */
int32_t xrm::system::resReserveCuSubListFromAnyDeviceTargetTypeLogicGroup(uint64_t reservePoolId, 
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];

        // itemFlags record is updated in the reserve call since device id is unknown here.
        ret = resReserveCuFromAnyDeviceTargetTypeLogicGroup(reservePoolId, cuPropV2, itemFlags,
            cuId, cuListResInforV2);
        if (ret != XRM_SUCCESS) {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /*
         * Fail to allocate cu of this sub list, relinquish the allocated resource
         * for this sub-list based on these input and record.
         */
        resRelinquishCuLgSubListV2(reservePoolId, lgSubList, cuListPropV2, itemFlags);
    }
    return (ret);
}

/*
 * Reserve cu from:
 *          Any device
 *          Target type logic group
 * So the cu can from group on differernt device.
 */
int32_t xrm::system::resReserveCuFromAnyDeviceTargetTypeLogicGroup(uint64_t reservePoolId,
    cuPropertyV2* cuPropV2, itemFlag* itemFlags, int32_t itemIndex, cuListResInforV2* cuListResInforV2) {
    int32_t ret = XRM_ERROR;
    int32_t devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }

        ret = resReserveCuFromTargetDeviceTargetTypeLogicGroup(reservePoolId, devId, cuPropV2);
        if (ret == XRM_SUCCESS) {
            itemFlags[itemIndex].allocated = true;
            itemFlags[itemIndex].devId = devId;
            cuListResInforV2->cuResInfor[itemIndex].deviceId = devId;
            break;
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Any device
 *          Same logic group, and the group is different from that in used group list
 */
int32_t xrm::system::resReserveCuSubListFromAnyDeviceSameLogicGroupNotUsed(uint64_t reservePoolId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags,
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList, cuListResInforV2* cuListResInforV2) {
    int32_t ret, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = &(cuListPropV2->cuProps[lgSubList->cuId[0]]);
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    ret = XRM_ERROR;
    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }
        ret = resReserveCuSubListFromTargetDeviceSameLogicGroupNotUsed(reservePoolId, devId, lgSubList,
                cuListPropV2, itemFlags, usedLgIdDevIdPairList, cuListResInforV2);
        if (ret == XRM_SUCCESS) {
            return (ret);
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Any device
 *          Same logic group, no limit the group to be different from that in used group list
 */
int32_t xrm::system::resReserveCuSubListFromAnyDeviceSameLogicGroupNoLimitUsedOrNot(uint64_t reservePoolId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = &(cuListPropV2->cuProps[lgSubList->cuId[0]]);
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    ret = XRM_ERROR;
    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }
        ret = resReserveCuSubListFromTargetDeviceSameLogicGroupNoLimitUsedOrNot(reservePoolId, devId, lgSubList,
                cuListPropV2, itemFlags, cuListResInforV2);
        if (ret == XRM_SUCCESS) {
            return (ret);
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Any device
 *          Target type logic grou
 *          Same logic group, and the group is different from that in used group list
 */
int32_t xrm::system::resReserveCuSubListFromAnyDeviceTargetTypeSameLogicGroupNotUsed(uint64_t reservePoolId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags,
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList, cuListResInforV2* cuListResInforV2) {
    int32_t ret, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = &(cuListPropV2->cuProps[lgSubList->cuId[0]]);
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    ret = XRM_ERROR;
    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }
        ret = resReserveCuSubListFromTargetDeviceTargetTypeSameLogicGroupNotUsed(reservePoolId, devId, lgSubList,
                cuListPropV2, itemFlags, usedLgIdDevIdPairList, cuListResInforV2);
        if (ret == XRM_SUCCESS) {
            return (ret);
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Any device
 *          Target type logic grou
 *          Same logic group, no limit the group to be different from that in used group list
 */
int32_t xrm::system::resReserveCuSubListFromAnyDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(uint64_t reservePoolId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = &(cuListPropV2->cuProps[lgSubList->cuId[0]]);
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    ret = XRM_ERROR;
    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }
        ret = resReserveCuSubListFromTargetDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(reservePoolId, devId, lgSubList,
                cuListPropV2, itemFlags, cuListResInforV2);
        if (ret == XRM_SUCCESS) {
            return (ret);
        }
    }
    return (ret);
}

/*
 * Reserve cu sub list from:
 *          Any device
 *          Any logic group
 * NOTE: there is no logic group constraint.
 */
int32_t xrm::system::resReserveCuSubListFromAnyDeviceAnyLogicGroup(uint64_t reservePoolId,
    logicGroupConstraintSubListTypeV2* lgSubList, cuListPropertyV2* cuListPropV2,
    itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret, lgCuIndex, cuId;
    cuPropertyV2* cuPropV2 = NULL;

    // loop request cu in sub list
    for (lgCuIndex = 0; lgCuIndex < lgSubList->numCu; lgCuIndex++) {
        // find cu index in the request list
        cuId = lgSubList->cuId[lgCuIndex];
        cuPropV2 = &cuListPropV2->cuProps[cuId];

        // itemFlags record is updated in reserve call since devId is unknown here.
        ret = resReserveCuFromAnyDeviceAnyLogicGroup(reservePoolId, cuPropV2, itemFlags, cuId, cuListResInforV2);
        if (ret != XRM_SUCCESS) {
            break;
        }
    }
    if (ret != XRM_SUCCESS) {
        /*
         * Fail to reserve cu of this sub list, relinquish the allocated resource
         * for this sub-list based on these input and record.
         */
        resRelinquishCuLgSubListV2(reservePoolId, lgSubList, cuListPropV2, itemFlags);
    }
    return (ret);
}

/*
 * Reserve cu from:
 *          Any device
 *          Any logic group
 * So the cu can from group on differernt device.
 */
int32_t xrm::system::resReserveCuFromAnyDeviceAnyLogicGroup(uint64_t reservePoolId,
    cuPropertyV2* cuPropV2, itemFlag* itemFlags, int32_t itemIndex, cuListResInforV2* cuListResInforV2) {
    int32_t ret = XRM_ERROR;
    int32_t devId;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    // loop all devices
    for (devId = -1; devId < m_numDevice;) {
        /* find one available device for this cu*/
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            return (XRM_ERROR_NO_KERNEL);
        }

        ret = resReserveCuFromTargetDeviceAnyLogicGroup(reservePoolId, devId, cuPropV2);
        if (ret == XRM_SUCCESS) {
            itemFlags[itemIndex].allocated = true;
            itemFlags[itemIndex].devId = devId;
            cuListResInforV2->cuResInfor[itemIndex].deviceId = devId;
            break;
        }
    }
    return (ret);
}

/*
 * Reserve multiple cu sub lists from any device:
 *
 * These different cu sub lists are generated based on logic group constraint information.
 */
int32_t xrm::system::resReserveCuSubListsFromAnyDevice(uint64_t reservePoolId, int32_t numLgConstraintSubList,
    logicGroupConstraintSubListTypeV2* lgConstraintSubLists,
    cuListPropertyV2* cuListPropV2, itemFlag* itemFlags, cuListResInforV2* cuListResInforV2) {
    int32_t ret;
    logicGroupConstraintSubListTypeV2* lgSubList = NULL;
    logicGroupIdDeviceIdPairListPropertyV2* usedLgIdDevIdPairList = NULL;
    int32_t lgSubListIndex, lgSubListType, lgId;

    /*
     * use to record used logic group id/device id pair for sub-list which request from
     * different pseudo logic group index (different pair means different logic group)
     */
    usedLgIdDevIdPairList = (logicGroupIdDeviceIdPairListPropertyV2*)malloc(
        sizeof(logicGroupIdDeviceIdPairListPropertyV2));
    memset(usedLgIdDevIdPairList, 0, sizeof(logicGroupIdDeviceIdPairListPropertyV2));

    // loop all the Logic group constraint based sub-Lists
    for (lgSubListIndex = 0; lgSubListIndex < numLgConstraintSubList; lgSubListIndex++) {
        lgSubList = &(lgConstraintSubLists[lgSubListIndex]);
        lgSubListType = lgSubList->type;
        switch (lgSubListType) {
            case LG_CONSTRAINT_SUBLIST_TYPE_REAL_GROUP_ID:
                /* 
                 * Reserve this logic group sub-list from target logic group of any device.
                 * CU can from different device.
                 */
                lgId = lgSubList->groupId;
                ret = resReserveCuSubListFromAnyDeviceTargetLogicGroup(reservePoolId, lgId, lgSubList,
                    cuListPropV2, itemFlags, cuListResInforV2);
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE:
                /*
                 * Reserve this logic group sub-list from target type logic group of any device.
                 * CU can from different logic group.
                 */
                ret = resReserveCuSubListFromAnyDeviceTargetTypeLogicGroup(reservePoolId, lgSubList,
                    cuListPropV2, itemFlags, cuListResInforV2);
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_PSEUDO_GROUP_ID:
                if (lgSubList->groupId >= 0) {
                    /*
                     * Reserve this logic group sub-list from same logic group of any device.
                     * groupId >= 0, means this logic group need to be different from previous used groups.
                     */
                    ret = resReserveCuSubListFromAnyDeviceSameLogicGroupNotUsed(reservePoolId, lgSubList,
                        cuListPropV2, itemFlags, usedLgIdDevIdPairList, cuListResInforV2);
                } else {
                    /*
                     * Reserve this logic group sub-list from same logic group of any device.
                     * groupId < 0, means no limit he logic group to be different from previous used groups.
                     */
                    ret = resReserveCuSubListFromAnyDeviceSameLogicGroupNoLimitUsedOrNot(reservePoolId, lgSubList,
                        cuListPropV2, itemFlags, cuListResInforV2);
                }
                break;
            case LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE_AND_PSEUDO_GROUP_ID:
                if (lgSubList->groupId >= 0) {
                    /*
                     * Reserve this logic group sub-list from target type same logic group of any device.
                     * groupId >= 0, means this logic group need to be different from previous used groups.
                     */
                    ret = resReserveCuSubListFromAnyDeviceTargetTypeSameLogicGroupNotUsed(reservePoolId, lgSubList,
                        cuListPropV2, itemFlags, usedLgIdDevIdPairList, cuListResInforV2);
                } else {
                    /*
                     * Reserve this logic group sub-list from target type same logic group of any device.
                     * groupId < 0, means no limit this logic group to be different from previous used groups.
                     */
                    ret = resReserveCuSubListFromAnyDeviceTargetTypeSameLogicGroupNoLimitUsedOrNot(reservePoolId, lgSubList,
                        cuListPropV2, itemFlags, cuListResInforV2);
                }
                break;
            case DEV_CONSTRAINT_SUBLIST_TYPE_NULL:
                ret = resReserveCuSubListFromAnyDeviceAnyLogicGroup(reservePoolId, lgSubList,
                    cuListPropV2, itemFlags, cuListResInforV2);
                break;
            default:
                ret = XRM_ERROR_INVALID;
                break;
        }
        if (ret != XRM_SUCCESS) {
            break;
        }
    } // loop lgConstraintSubLists

    if (ret != XRM_SUCCESS) {
        /*
         * Fail to reserve cu of this sub list, relinquish the reserved resource
         * for all sub-lists
         */
        for (lgSubListIndex = 0; lgSubListIndex < numLgConstraintSubList; lgSubListIndex++) {
            lgSubList = &(lgConstraintSubLists[lgSubListIndex]);
            resRelinquishCuLgSubListV2(reservePoolId, lgSubList, cuListPropV2, itemFlags);
        }
    }

    if (usedLgIdDevIdPairList != NULL)
        free(usedLgIdDevIdPairList);
    return (ret);
}

/*
 * Reserve one cu list based on the request property.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 * 0. Allocation policy is not applied for cu list reservation, so no need to handle it.
 *
 * 1. Device constraint is priority 1, it's handled first.
 *
 *    To handle the device constraint information, CU list will be put into differernt
 *    sub-list:
 *    a) one or multiple sub-lists with same hw device index, different index into
 *       different sub-list. CU in such sub-list will be reserved from target device later.
 *    b) one or multiple sub-lists with same virtual device index, different index
 *       into different sub-list. cu in same sub list will be reserved from same device.
 *       different sub-list will be reserved from different device later.
 *    c) one sub-list for cu without device constraint. cu in this sub-list will be
 *       reserved from any device later.
 *
 * 2. Logic group constraint is priority 2, it's handled based on device constraint
 *    process result.
 *
 *    Above device constraint based sub-lists willbe further put into logic group
 *    constraint based small sub-lists:
 *    a) one or multiple sub-list with same real group id.
 *    b) one or multiple sub-list with same group type.
 *    c) one or multiple sub-list with same pseudo group id, different list
 *       from different logic group.
 *    d) one or multiple sub-list with same pseudo group id and type, different list
 *       from different logic group.
 *    e) one sub-list for cu without logic group constraint.
 *
 *    Then these sub logic group based on sub-lists will be reserved accordingly:
 *    1) if logic group constraint not set: reserve it from any logic group.
 *    2) if logic group constraint set with real group id: reserve it from target logic group.
 *    3) if logic group constraint set with group type: reserve it from logic group with the type.
 *    4) if logic group constraint set with pseudo group id: find all others with same
 *       pseudo group id, put them into one sub-list, and reserve requested CU
 *       from same logic group, the logic group id should be different from that
 *       for previous sub-lists request from pseudo group id.
 *    5) if logic group constraint set with pseudo group id and type: find all others with same
 *       pseudo group id, put them into one sub-list, and check the group type information. If
 *       group type of same sub-list is different, whole reservation will fail. Then reserve
 *       requested CU from same logic group, the logic group id should be different from that
 *       for previous sub-lists request from pseudo group id.
 *
 * XRM_SUCCESS: reserved resouce is associated with the reservationId
 * Otherwise: failed to reserve the cu list
 *
 * Lock: should enter lock during the cu list reservation
 */
int32_t xrm::system::resReserveCuListV2(uint64_t reservePoolId,
                                        cuListPropertyV2* cuListPropV2,
                                        cuListResInforV2* cuListResInforV2) {
    int32_t index, ret;
    deviceData* dev = NULL;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    cuPropertyV2* cuPropV2 = NULL;
    uint64_t clientId;

    itemFlag* itemFlags = NULL;
    deviceIdListPropertyV2* usedDevIdList = NULL;

    deviceConstraintSubListTypeV2* deviceConstraintSubLists = NULL;
    logicGroupConstraintSubListTypeV2* lgConstraintSubLists = NULL;

    deviceConstraintSubListTypeV2* devSubList = NULL;
    int32_t devSubListIndex, devSubListNumCu, devCuIndex, cuId, devSubListType, devId;
    int32_t numLgConstraintSubList;

    if (cuListPropV2 == NULL) {
        logMsg(XRM_LOG_ERROR, "%s : cuListPropV2 is null\n", __func__);
        return (XRM_ERROR_INVALID);
    }
    if (cuListPropV2->cuNum <= 0 || cuListPropV2->cuNum > XRM_MAX_LIST_CU_NUM_V2) {
        logMsg(XRM_LOG_ERROR, "%s : cuNum out of rage %d\n", __func__, cuListPropV2->cuNum);
        return (XRM_ERROR_INVALID);
    }

    // client id of same list is same
    clientId = cuListPropV2->cuProps[0].clientId;

    /* use to set the flag to record whether the item is handled */
    itemFlags = (itemFlag*)malloc(sizeof(itemFlag) * XRM_MAX_LIST_CU_NUM_V2);
    memset(itemFlags, 0, sizeof(itemFlag) * XRM_MAX_LIST_CU_NUM_V2);

    /* use to record used device id for sub-list which request from different virtual device index */
    usedDevIdList = (deviceIdListPropertyV2*)malloc(sizeof(deviceIdListPropertyV2));
    memset(usedDevIdList, 0, sizeof(deviceIdListPropertyV2));

    ret = XRM_ERROR;

    /*
     * 1. handle device constraint information
     *
     * put cu into differernt type deviceConstraintSubList:
     *     DEV_CONSTRAINT_SUBLIST_TYPE_HARDWARE_DEVICE_ID
     *     DEV_CONSTRAINT_SUBLIST_TYPE_VIRTUAL_DEVICE_ID
     *     DEV_CONSTRAINT_SUBLIST_TYPE_NULL
     */
    deviceConstraintSubLists = (deviceConstraintSubListTypeV2*)malloc(
        sizeof(deviceConstraintSubListTypeV2) * XRM_MAX_LIST_CU_NUM_V2);
    memset(deviceConstraintSubLists, 0, sizeof(deviceConstraintSubListTypeV2) * XRM_MAX_LIST_CU_NUM_V2);
    int32_t numDeviceConstraintSubList = 0;
    for (index = 0; index < cuListPropV2->cuNum; index++) {
        cuPropV2 = &cuListPropV2->cuProps[index];
        if ((cuPropV2->policyInfo & XRM_POLICY_INFO_CONSTRAINT_TYPE_MASK) != XRM_POLICY_INFO_CONSTRAINT_TYPE_NULL) {
            logMsg(XRM_LOG_ERROR, "%s : cu [%d] policy should not set for CU List\n", __func__, index);
            ret = XRM_ERROR_INVALID;
            break;
        }
        ret = putCuIntoDeviceConstraintSubLists(cuPropV2, index, &numDeviceConstraintSubList,
            deviceConstraintSubLists);
        if (ret != XRM_SUCCESS)
            break;
    }
    if (ret != XRM_SUCCESS) {
        goto resReserveCuListV2Exit;
    }

    /*
     * 2. handle logic group constraint information
     *
     * For each deviceConstraintSubList, put cu into differernt type logicGroupConstraintSubList:
     *     LG_CONSTRAINT_SUBLIST_TYPE_REAL_GROUP_ID
     *     LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE
     *     LG_CONSTRAINT_SUBLIST_TYPE_PSEUDO_GROUP_ID
     *     LG_CONSTRAINT_SUBLIST_TYPE_GROUP_TYPE_AND_PSEUDO_GROUP_ID
     *     DEV_CONSTRAINT_SUBLIST_TYPE_NULL
     * Then reserve cu of all logicGroupConstraintSubList (from this deviceConstraintSubList).
     */
    lgConstraintSubLists = (logicGroupConstraintSubListTypeV2*)malloc(sizeof
        (logicGroupConstraintSubListTypeV2) * XRM_MAX_LIST_CU_NUM_V2);

    for (devSubListIndex = 0; devSubListIndex < numDeviceConstraintSubList; devSubListIndex++) {
        /* handle each devSubList */

        /* init logic group related data for each devSubList */
        memset(lgConstraintSubLists, 0, sizeof(logicGroupConstraintSubListTypeV2) * XRM_MAX_LIST_CU_NUM_V2);
        numLgConstraintSubList = 0;

        devSubList = &(deviceConstraintSubLists[devSubListIndex]);
        devSubListNumCu = devSubList->numCu;

        /* put cu into differernt type logicGroupConstraintSubList */
        for (devCuIndex = 0; devCuIndex < devSubListNumCu; devCuIndex++) {
            /* get the cuId in cuListPropV2 */
            cuId = devSubList->cuId[devCuIndex];
            cuPropV2 = &cuListPropV2->cuProps[cuId];
            ret = putCuIntoLgConstraintSubLists(cuPropV2, cuId, &numLgConstraintSubList, lgConstraintSubLists);
            if (ret != XRM_SUCCESS) {
                goto resReserveCuListV2Exit;
            }
        } // put cu into differernt type logicGroupConstraintSubList

        /* reserve cu of each lgConstraintSubList */
        devSubListType = devSubList->type;
        switch (devSubListType) {
            case DEV_CONSTRAINT_SUBLIST_TYPE_HARDWARE_DEVICE_ID:
                /* handle the sub list with hardware device id */
                devId = devSubList->deviceId;

                dev = &m_devList[devId];
                if (!dev->isLoaded) {
                    ret = XRM_ERROR_NO_DEV;
                    goto resReserveCuListV2Exit;
                }
                if ((dev->isExcl) && (dev->clientProcs[0].clientId != clientId)) {
                    /* the device is used by other client exclusively */
                    ret = XRM_ERROR_NO_DEV;
                    goto resReserveCuListV2Exit;
                }

                // reserve resource for the logic based sub-lists (device id is set)
                ret = resReserveCuSubListsFromTargetDevice(reservePoolId, devId, numLgConstraintSubList,
                    lgConstraintSubLists, cuListPropV2, itemFlags, cuListResInforV2);
                if (ret != XRM_SUCCESS) {
                    goto resReserveCuListV2Exit;
                }
                break;
            case DEV_CONSTRAINT_SUBLIST_TYPE_VIRTUAL_DEVICE_ID:
                /*
                 * To handle the sub list with virtual device id, Different sub list with differernt
                 * virtual device id.
                 *
                 * Reserve resource for the logic group based sub-lists (virtual device id is set):
                 * CU in one sub-list from same device, the device is not in used list to make sure
                 * CU in different sub-list will come from different device.
                 */
                for (devId = -1; devId < m_numDevice;) {
                    /* find one available device for first cu */
                    cuPropV2 = &(cuListPropV2->cuProps[lgConstraintSubLists[0].cuId[0]]);
                    cuPropertyCopyFromV2(cuProp, cuPropV2);
                alloc_device:
                    devId = allocDevForClient(&devId, cuProp);
                    if (devId < 0) {
                        ret = XRM_ERROR_NO_DEV;
                        goto resReserveCuListV2Exit;
                    }
                    /* check whether the device is in used devId list, if yes, then find another one */
                    if (isInDeviceIdList(usedDevIdList, devId)) {
                        goto alloc_device;
                    }

                    ret = resReserveCuSubListsFromTargetDevice(reservePoolId, devId, numLgConstraintSubList,
                        lgConstraintSubLists, cuListPropV2, itemFlags, cuListResInforV2);
                    if (ret == XRM_SUCCESS) {
                        // update used device id list: add this new one in
                        usedDevIdList->deviceIds[usedDevIdList->deviceNum] = devId;
                        usedDevIdList->deviceNum++;
                        break;
                    }
                } // loop all the devices
                if (ret != XRM_SUCCESS) {
                    goto resReserveCuListV2Exit;
                }
                break;
            case DEV_CONSTRAINT_SUBLIST_TYPE_NULL:
                /*
                 * To handle the sub list without device constraint: can from any device:
                 * To reserve resource for the logic based sub-lists (device id is NOT set)
                 */
                ret = resReserveCuSubListsFromAnyDevice(reservePoolId, numLgConstraintSubList, lgConstraintSubLists,
                    cuListPropV2, itemFlags, cuListResInforV2);
                if (ret != XRM_SUCCESS) {
                    goto resReserveCuListV2Exit;
                }
                break;
            default:
                logMsg(XRM_LOG_ERROR, "invalid devSubList [%d] constraint type\n", devSubListIndex);
                ret = XRM_ERROR_INVALID;
                break;
        } // reserve cu of each lgConstraintSubList in deviceConstraintSubList
    } // handle each devSubList

resReserveCuListV2Exit:
    if (ret != XRM_SUCCESS) {
        /*
         * fail to reserve whole list, then the whole pool reserve will fail,
         * so release resource of this whole pool associated with reservePoolId.
         */
        resRelinquishCuPoolV2(reservePoolId);
    }
    if (deviceConstraintSubLists != NULL)
        free(deviceConstraintSubLists);
    if (lgConstraintSubLists != NULL)
        free(lgConstraintSubLists);
    if (itemFlags != NULL)
        free(itemFlags);
    if (usedDevIdList != NULL)
        free(usedDevIdList);
    return (ret);
}

/*
 * relinquish one CU (Compute Unit) from specified device(devId) based on the request property.
 *
 * lock: should already in lock protection during the cu relinquish process
 */
int32_t xrm::system::relinquishCuFromDev(uint64_t reservePoolId, int32_t devId, cuProperty* cuProp) {
    deviceData* dev;
    cuData* cu;
    bool kernelNameEqual, kernelAliasEqual;
    int32_t cuId;
    int32_t ret = XRM_ERROR;

    if (reservePoolId == 0) {
        logMsg(XRM_LOG_ERROR, "reservePoolId is invalid\n");
        return (XRM_ERROR_INVALID);
    }
    if (devId < 0 || devId >= m_numDevice) {
        logMsg(XRM_LOG_ERROR, "devId is invalid\n");
        return (XRM_ERROR_INVALID);
    }
    if (cuProp == NULL) {
        logMsg(XRM_LOG_ERROR, "cuProp is NULL\n");
        return (XRM_ERROR_INVALID);
    }
    if ((cuProp->kernelName[0] == '\0') && (cuProp->kernelAlias[0] == '\0')) {
        logMsg(XRM_LOG_ERROR, "neither kernel name nor alias are presented\n");
        return (XRM_ERROR_INVALID);
    }

    dev = &m_devList[devId];
    if (!dev->isLoaded) {
        logMsg(XRM_LOG_ERROR, "device [%d] is not loaded\n", devId);
        return (ret);
    }
    for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual)) continue;

        for (int32_t i = 0; i < cu->numReserve; i++) {
            if ((cu->reserves[i].reservePoolId == reservePoolId) &&
                (cu->reserves[i].reserveLoadUnified == cuProp->requestLoadUnified)) {
                /* there maybe same cu reserved, try others */
                if (cu->reserves[i].reserveUsedLoadUnified) {
                    continue;
                }
                cu->totalUsedLoadUnified -= cu->reserves[i].reserveLoadUnified;
                updateDeviceLoad(devId, -cu->reserves[i].reserveLoadUnified, -1);
                cu->totalReservedLoadUnified -= cu->reserves[i].reserveLoadUnified;
                /* switch the slot until the last one */
                for (; i < cu->numReserve - 1; i++) {
                    cu->reserves[i].reserveLoadUnified = cu->reserves[i + 1].reserveLoadUnified;
                    cu->reserves[i].reserveUsedLoadUnified = cu->reserves[i + 1].reserveUsedLoadUnified;
                    cu->reserves[i].reservePoolId = cu->reserves[i + 1].reservePoolId;
                    cu->reserves[i].clientIsActive = cu->reserves[i + 1].clientIsActive;
                    cu->reserves[i].clientId = cu->reserves[i + 1].clientId;
                    cu->reserves[i].clientProcessId = cu->reserves[i + 1].clientProcessId;
                    copyLogicGroupInfor(&(cu->reserves[i].logicGroupInfo),
                        &(cu->reserves[i + 1].logicGroupInfo));
                }
                /* empty the last slot */
                cu->reserves[i].reserveLoadUnified = 0;
                cu->reserves[i].reserveUsedLoadUnified = 0;
                cu->reserves[i].reservePoolId = 0;
                cu->reserves[i].clientIsActive = false;
                cu->reserves[i].clientId = 0;
                cu->reserves[i].clientProcessId = 0;
                memset(&(cu->reserves[i].logicGroupInfo), 0, sizeof(cuResourceLogicGroupInfo));
                cu->numReserve--;
                /* relinquished the cu successfully */
                ret = XRM_SUCCESS;
                return (ret);
            }
        }
    }
    return (ret);
}

/*
 * the interface to relinquish cu list resource based on the reserve pool id,
 * it will actually relinquish whole cu pool resource.
 *
 * lock: should already in lock protection during the cu list relinquish process
 */
int32_t xrm::system::resRelinquishCuList(uint64_t reservePoolId) {
    if (reservePoolId == 0) return (XRM_ERROR_INVALID);

    deviceData* dev;
    cuData* cu;
    int32_t devId, cuId;

    for (devId = 0; devId < m_numDevice; devId++) {
        dev = &m_devList[devId];
        if (!dev->isLoaded) continue;
        for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];
            for (int32_t i = 0; i < cu->numReserve; i++) {
                if (cu->reserves[i].reservePoolId == reservePoolId) {
                    if (cu->reserves[i].reserveUsedLoadUnified) {
                        return (XRM_ERROR);
                    }
                    cu->totalUsedLoadUnified -= cu->reserves[i].reserveLoadUnified;
                    updateDeviceLoad(devId, -cu->reserves[i].reserveLoadUnified, -1);
                    cu->totalReservedLoadUnified -= cu->reserves[i].reserveLoadUnified;
                    /* switch the slot until the last one */
                    for (; i < cu->numReserve - 1; i++) {
                        cu->reserves[i].reserveLoadUnified = cu->reserves[i + 1].reserveLoadUnified;
                        cu->reserves[i].reserveUsedLoadUnified = cu->reserves[i + 1].reserveUsedLoadUnified;
                        cu->reserves[i].reservePoolId = cu->reserves[i + 1].reservePoolId;
                        cu->reserves[i].clientIsActive = cu->reserves[i + 1].clientIsActive;
                        cu->reserves[i].clientId = cu->reserves[i + 1].clientId;
                        cu->reserves[i].clientProcessId = cu->reserves[i + 1].clientProcessId;
                        copyLogicGroupInfor(&(cu->reserves[i].logicGroupInfo),
                            &(cu->reserves[i + 1].logicGroupInfo));
                    }
                    /* empty the last slot */
                    cu->reserves[i].reserveLoadUnified = 0;
                    cu->reserves[i].reserveUsedLoadUnified = 0;
                    cu->reserves[i].reservePoolId = 0;
                    cu->reserves[i].clientIsActive = false;
                    cu->reserves[i].clientId = 0;
                    cu->reserves[i].clientProcessId = 0;
                    memset(&(cu->reserves[i].logicGroupInfo), 0, sizeof(cuResourceLogicGroupInfo));
                    cu->numReserve--;
                    break;
                }
            }
        }
    }
    return (XRM_SUCCESS);
}

/*
 * the interface to relinquish cu list resource based on the reserve pool id,
 * it will actually relinquish whole cu pool resource.
 *
 * lock: should already in lock protection during the cu list relinquish process
 */
int32_t xrm::system::resRelinquishCuListV2(uint64_t reservePoolId) {
    int32_t ret = resRelinquishCuList(reservePoolId);
    return (ret);
}

/*
 * Reserve CU (Compute Unit) of one whole xclbin based on the request property.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 * return:
 *   XRM_SUCCESS: reserve the cu resouce and associate it with the reserve pool id
 *   Otherwise: failed to reserve the cu resource with specified reserve pool id
 *
 * Lock: should already in lock protection during the cu reservation
 */
int32_t xrm::system::resReserveCuOfXclbin(
    uint64_t reservePoolId, uuid_t uuid, uint64_t clientId, pid_t clientProcessId, uint64_t* fromDevId) {
    deviceData* dev;
    cuData* cu;
    int32_t cuId;

    if (reservePoolId == 0) {
        logMsg(XRM_LOG_ERROR, "%s(), reserve pool is is 0\n", __func__);
        return (XRM_ERROR_INVALID);
    }

    for (int32_t devId = 0; devId < m_numDevice; devId++) {
        dev = &m_devList[devId];
        if (!dev->isLoaded) continue;
        /* To check whether the device is loaded with xclbin of specified uuid */
        if (!isDeviceLoadedXclbinWithUuid(devId, uuid)) continue;
        /* Whether device is already in used */
        if (isDeviceBusy(devId)) continue;
        /* reserve the all the cu on this device */
        for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];
            cu->totalUsedLoadUnified = XRM_MAX_CU_LOAD_GRANULARITY_1000000; // will update dev load at end of loop
            cu->totalReservedLoadUnified = XRM_MAX_CU_LOAD_GRANULARITY_1000000;
            cu->reserves[0].reserveLoadUnified = XRM_MAX_CU_LOAD_GRANULARITY_1000000;
            cu->reserves[0].reserveUsedLoadUnified = 0;
            cu->reserves[0].reservePoolId = reservePoolId;
            cu->reserves[0].clientIsActive = true;
            cu->reserves[0].clientId = clientId;
            cu->reserves[0].clientProcessId = clientProcessId;
            cu->numReserve = 1;
        }
        updateDeviceLoad(devId, 0, dev->xclbinInfo.numCu * XRM_MAX_CU_LOAD_GRANULARITY_1000000);
        *fromDevId = devId;
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Reserve CU (Compute Unit) of one whole device based on the request property.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 * return:
 *   XRM_SUCCESS: reserve the cu resouce and associate it with the reserve pool id
 *   Otherwise: failed to reserve the cu resource with specified reserve pool id
 *
 * Lock: should already in lock protection during the cu reservation
 */
int32_t xrm::system::resReserveDevice(uint64_t reservePoolId,
                                      uint64_t deviceId,
                                      uint64_t clientId,
                                      pid_t clientProcessId) {
    deviceData* dev;
    int32_t devId = (int32_t)(deviceId & 0xFFFFFFFF);
    cuData* cu;
    int32_t cuId;

    if (reservePoolId == 0) {
        logMsg(XRM_LOG_ERROR, "%s(), reserve pool is is 0\n", __func__);
        return (XRM_ERROR_INVALID);
    }

    if (devId < 0 || devId > (XRM_MAX_XILINX_DEVICES - 1)) {
        logMsg(XRM_LOG_ERROR, "%s(), device id %d is out of range\n", __func__, devId);
        return (XRM_ERROR_INVALID);
    }

    dev = &m_devList[devId];
    if (!dev->isLoaded) return (XRM_ERROR_NO_KERNEL);
    /* Whether device is already in used */
    if (isDeviceBusy(devId)) return (XRM_ERROR_NO_KERNEL);
    /* reserve the all the cu on this device */
    for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
        cu = &dev->xclbinInfo.cuList[cuId];
        cu->totalUsedLoadUnified = XRM_MAX_CU_LOAD_GRANULARITY_1000000; // end of loop for dev load
        cu->totalReservedLoadUnified = XRM_MAX_CU_LOAD_GRANULARITY_1000000;
        cu->reserves[0].reserveLoadUnified = XRM_MAX_CU_LOAD_GRANULARITY_1000000;
        cu->reserves[0].reserveUsedLoadUnified = 0;
        cu->reserves[0].reservePoolId = reservePoolId;
        cu->reserves[0].clientIsActive = true;
        cu->reserves[0].clientId = clientId;
        cu->reserves[0].clientProcessId = clientProcessId;
        cu->numReserve = 1;
    }
    updateDeviceLoad(devId, 0, dev->xclbinInfo.numCu * XRM_MAX_CU_LOAD_GRANULARITY_1000000);
    return (XRM_SUCCESS);
}

/*
 * Reserve one cu pool based on the request property.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 *   reservePoolId: reserve cu pool successed
 *   0: failed to reserve the cu pool resource
 *
 * Lock: should enter lock during the cu pool reservation process
 */
uint64_t xrm::system::resReserveCuPool(cuPoolProperty* cuPoolProp) {
    int32_t ret;
    uint64_t reservePoolId = 0;
    cuListProperty* cuListProp = NULL;
    int32_t reserveCuListNum = 0;
    int32_t reserveXclbinNum = 0;

    if (cuPoolProp == NULL) {
        return (reservePoolId);
    }
    if (cuPoolProp->cuListNum < 0 || cuPoolProp->xclbinNum < 0) {
        return (reservePoolId);
    }
    cuListProp = &(cuPoolProp->cuListProp);
    if (cuListProp->cuNum < 0 || cuListProp->cuNum > XRM_MAX_LIST_CU_NUM) {
        return (reservePoolId);
    }

    ret = XRM_ERROR;
    reservePoolId = getNextReservePoolId();
    uint64_t fromDevId;
    /*
     * to reserve whole device at first, then reserve the cu list in case cu list
     * resource available on multiple different devices.
     */
    while (reserveXclbinNum < cuPoolProp->xclbinNum) {
        /*
         * To reserve all the cu resources of one xclbin
         */
        ret = resReserveCuOfXclbin(reservePoolId, cuPoolProp->xclbinUuid, cuPoolProp->clientId,
                                   cuPoolProp->clientProcessId, &fromDevId);
        if (ret != XRM_SUCCESS) {
            resRelinquishCuPool(reservePoolId);
            return (0);
        }
        reserveXclbinNum++;
    }
    while (reserveCuListNum < cuPoolProp->cuListNum) {
        /*
         * To reserve the cu list
         */
        ret = resReserveCuList(reservePoolId, cuListProp);
        if (ret != XRM_SUCCESS) {
            resRelinquishCuPool(reservePoolId);
            return (0);
        }
        reserveCuListNum++;
    }

    updateReservePoolId();
    return (reservePoolId);
}

/*
 * Reserve one cu pool based on the request property.
 *
 * Please be noted that the reservation will NOT record the client information
 * (clientId etc).
 *
 *   reservePoolId: reserve cu pool successed
 *   0: failed to reserve the cu pool resource
 *
 * Lock: should enter lock during the cu pool reservation process
 */
uint64_t xrm::system::resReserveCuPoolV2(cuPoolPropertyV2* cuPoolProp, cuPoolResInforV2* cuPoolResInfor) {
    int32_t ret;
    uint64_t reservePoolId = 0;
    cuListPropertyV2* cuListProp = NULL;
    deviceIdListPropertyV2* deviceIdListProp = NULL;
    int32_t reserveCuListNum = 0;
    int32_t reserveXclbinNum = 0;
    cuListResInforV2* cuListResInfor = NULL;

    if (cuPoolProp == NULL) {
        return (reservePoolId);
    }
    if (cuPoolProp->cuListNum < 0 || cuPoolProp->xclbinNum < 0 || cuPoolProp->deviceIdListProp.deviceNum < 0 ||
        cuPoolProp->deviceIdListProp.deviceNum > XRM_MAX_XILINX_DEVICES) {
        return (reservePoolId);
    }
    cuListProp = &(cuPoolProp->cuListProp);
    if (cuListProp->cuNum < 0 || cuListProp->cuNum > XRM_MAX_LIST_CU_NUM_V2) {
        return (reservePoolId);
    }

    ret = XRM_ERROR;
    reservePoolId = getNextReservePoolId();
    uint64_t fromDevId;
    /*
     * first to reserve specified device (hardware device id);
     * second to reserve whole device (xclbin uuid);
     * third to reserve the cu list in case cu list
     * resource available on multiple different devices.
     */
    deviceIdListProp = &(cuPoolProp->deviceIdListProp);
    for (int32_t i = 0; i < deviceIdListProp->deviceNum; i++) {
        /*
         * To reserve all the cu resources of one device
         */
        ret = resReserveDevice(reservePoolId, deviceIdListProp->deviceIds[i], cuPoolProp->clientId,
                               cuPoolProp->clientProcessId);
        if (ret != XRM_SUCCESS) {
            resRelinquishCuPoolV2(reservePoolId);
            return (0);
        }
    }
    while (reserveXclbinNum < cuPoolProp->xclbinNum) {
        /*
         * To reserve all the cu resources of one xclbin
         */
        ret = resReserveCuOfXclbin(reservePoolId, cuPoolProp->xclbinUuid, cuPoolProp->clientId,
                                   cuPoolProp->clientProcessId, &fromDevId);
        if (ret != XRM_SUCCESS) {
            resRelinquishCuPoolV2(reservePoolId);
            return (0);
        }
        cuPoolResInfor->xclbinResInfor[reserveXclbinNum].deviceId = fromDevId;
        reserveXclbinNum++;
    }
    cuPoolResInfor->xclbinNum = cuPoolProp->xclbinNum;
    while (reserveCuListNum < cuPoolProp->cuListNum) {
        /*
         * To reserve the cu list
         */
        cuListResInfor = &(cuPoolResInfor->cuListResInfor[reserveCuListNum]);
        ret = resReserveCuListV2(reservePoolId, cuListProp, cuListResInfor);
        if (ret != XRM_SUCCESS) {
            resRelinquishCuPoolV2(reservePoolId);
            return (0);
        }
        cuListResInfor->cuNum = cuListProp->cuNum;
        reserveCuListNum++;
    }
    cuPoolResInfor->cuListNum = cuPoolProp->cuListNum;

    updateReservePoolId();
    return (reservePoolId);
}

/*
 * the interface to relinquish cu pool resource based on the reservation id
 *
 * lock: need to enter lock to protect the resource pool access during the relinquish process
 */
int32_t xrm::system::resRelinquishCuPool(uint64_t reservePoolId) {
    if (reservePoolId == 0) return (XRM_ERROR_INVALID);

    deviceData* dev;
    cuData* cu;
    int32_t devId, cuId;

    for (devId = 0; devId < m_numDevice; devId++) {
        dev = &m_devList[devId];
        if (!dev->isLoaded) continue;
        for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];
            for (int32_t i = 0; i < cu->numReserve; i++) {
                if (cu->reserves[i].reservePoolId == reservePoolId) {
                    if (cu->reserves[i].reserveUsedLoadUnified) {
                        return (XRM_ERROR);
                    }
                    cu->totalUsedLoadUnified -= cu->reserves[i].reserveLoadUnified;
                    updateDeviceLoad(devId, -cu->reserves[i].reserveLoadUnified, -1);
                    cu->totalReservedLoadUnified -= cu->reserves[i].reserveLoadUnified;
                    /* switch the slot until the last one */
                    for (; i < cu->numReserve - 1; i++) {
                        cu->reserves[i].reserveLoadUnified = cu->reserves[i + 1].reserveLoadUnified;
                        cu->reserves[i].reserveUsedLoadUnified = cu->reserves[i + 1].reserveUsedLoadUnified;
                        cu->reserves[i].reservePoolId = cu->reserves[i + 1].reservePoolId;
                        cu->reserves[i].clientIsActive = cu->reserves[i + 1].clientIsActive;
                        cu->reserves[i].clientId = cu->reserves[i + 1].clientId;
                        cu->reserves[i].clientProcessId = cu->reserves[i + 1].clientProcessId;
                        copyLogicGroupInfor(&(cu->reserves[i].logicGroupInfo),
                            &(cu->reserves[i + 1].logicGroupInfo));
                    }
                    /* empty the last slot */
                    cu->reserves[i].reserveLoadUnified = 0;
                    cu->reserves[i].reserveUsedLoadUnified = 0;
                    cu->reserves[i].reservePoolId = 0;
                    cu->reserves[i].clientIsActive = false;
                    cu->reserves[i].clientId = 0;
                    cu->reserves[i].clientProcessId = 0;
                    memset(&(cu->reserves[i].logicGroupInfo), 0, sizeof(cuResourceLogicGroupInfo));
                    cu->numReserve--;
                    break;
                }
            }
        }
    }
    return (XRM_SUCCESS);
}

/*
 * the interface to relinquish cu pool resource based on the reservation id
 *
 * lock: need to enter lock to protect the resource pool access during the relinquish process
 */
int32_t xrm::system::resRelinquishCuPoolV2(uint64_t reservePoolId) {
    int32_t ret = resRelinquishCuPool(reservePoolId);
    return (ret);
}

/*
 * Query reserve cu resource from system based on the reserve pool id.
 *
 * XRM_SUCCESS: queried resouces are filled into cuPoolRes
 * Otherwise: failed to query the reserve resource
 *
 * Lock: should enter lock during the resource allocation query
 */
int32_t xrm::system::resReservationQuery(uint64_t reservePoolId, cuPoolResource* cuPoolRes) {
    int32_t devId, cuId, reserveIdx;
    deviceData* dev = NULL;
    cuData* cu = NULL;
    reserveData* reserve = NULL;
    cuResource* cuRes = NULL;
    int32_t cuNum;

    if (cuPoolRes == NULL) {
        return (XRM_ERROR_INVALID);
    }
    if (reservePoolId == 0) {
        return (XRM_ERROR_INVALID);
    }

    cuNum = 0;
    memset(cuPoolRes, 0, sizeof(cuPoolResource));
    /* Check all the devices */
    for (devId = 0; devId < m_numDevice; devId++) {
        dev = &m_devList[devId];
        /* Check all the cu on this device */
        for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];
            /* Check all the reservations on this cu */
            for (reserveIdx = 0; reserveIdx < cu->numReserve; reserveIdx++) {
                reserve = &cu->reserves[reserveIdx];
                if (reserve->reservePoolId == reservePoolId) {
                    /* out of cu pool limitation */
                    if (cuNum >= XRM_MAX_POOL_CU_NUM) {
                        break;
                        /* to just return max number of cu, instead of return error.
                         * // memset(cuPoolRes, 0, sizeof(cuPoolResource));
                         * // return (XRM_ERROR);
                         */
                    }
                    cuRes = &cuPoolRes->cuResources[cuNum];
                    strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelPluginFileName, cu->kernelPluginFileName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelName, cu->kernelName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->instanceName, cu->instanceName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelAlias, cu->kernelAlias.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->cuName, cu->cuName.c_str(), XRM_MAX_NAME_LEN - 1);
                    cuRes->cuType = cu->cuType;
                    cuRes->baseAddr = cu->baseAddr;
                    cuRes->membankId = cu->membankId;
                    cuRes->membankType = cu->membankType;
                    cuRes->membankSize = cu->membankSize;
                    cuRes->membankBaseAddr = cu->membankBaseAddr;
                    cuRes->deviceId = devId;
                    cuRes->deviceNumBDF = dev->devNumBDF;
                    cuRes->cuId = cuId;
                    cuRes->poolId = reservePoolId;
                    cuNum++;
                }
            }
        }
    }
    cuPoolRes->cuNum = cuNum;
    return (XRM_SUCCESS);
}

/*
 * Query reserve cu resource from system based on the reserve pool id.
 *
 * XRM_SUCCESS: queried resouces are filled into cuPoolRes
 * Otherwise: failed to query the reserve resource
 *
 * Lock: should enter lock during the resource allocation query
 */
int32_t xrm::system::resReservationQueryV2(reservationQueryInfoV2* reserveQueryInfo, cuPoolResourceV2* cuPoolRes) {
    int32_t devId, cuId, reserveIdx;
    deviceData* dev = NULL;
    cuData* cu = NULL;
    bool kernelNameEqual, kernelAliasEqual;
    reserveData* reserve = NULL;
    cuResource* cuRes = NULL;
    int32_t cuNum;

    if (reserveQueryInfo == NULL) {
        return (XRM_ERROR_INVALID);
    }
    if (cuPoolRes == NULL) {
        return (XRM_ERROR_INVALID);
    }
    if (reserveQueryInfo->poolId == 0) {
        return (XRM_ERROR_INVALID);
    }

    cuNum = 0;
    memset(cuPoolRes, 0, sizeof(cuPoolResource));
    /* Check all the devices */
    for (devId = 0; devId < m_numDevice; devId++) {
        dev = &m_devList[devId];
        /* Check all the cu on this device */
        for (cuId = 0; cuId < dev->xclbinInfo.numCu; cuId++) {
            cu = &dev->xclbinInfo.cuList[cuId];
            /* kernel name is presented, compare it; otherwise no need to compare */
            kernelNameEqual = true;
            if (reserveQueryInfo->kernelName[0] != '\0') {
                if (cu->kernelName.compare(reserveQueryInfo->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (reserveQueryInfo->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(reserveQueryInfo->kernelAlias)) kernelAliasEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual)) continue;
            /* Check all the reservations on this cu */
            for (reserveIdx = 0; reserveIdx < cu->numReserve; reserveIdx++) {
                reserve = &cu->reserves[reserveIdx];
                if (reserve->reservePoolId == reserveQueryInfo->poolId) {
                    /* out of cu pool limitation */
                    if (cuNum >= (XRM_MAX_POOL_CU_NUM_V2 - 1)) {
                        break;
                        /* to just return max number of cu, instead of return error.
                         * // memset(cuPoolRes, 0, sizeof(cuPoolResource));
                         * // return (XRM_ERROR);
                         */
                    }
                    cuRes = &cuPoolRes->cuResources[cuNum];
                    strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelPluginFileName, cu->kernelPluginFileName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelName, cu->kernelName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->instanceName, cu->instanceName.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->kernelAlias, cu->kernelAlias.c_str(), XRM_MAX_NAME_LEN - 1);
                    strncpy(cuRes->cuName, cu->cuName.c_str(), XRM_MAX_NAME_LEN - 1);
                    cuRes->cuType = cu->cuType;
                    cuRes->baseAddr = cu->baseAddr;
                    cuRes->membankId = cu->membankId;
                    cuRes->membankType = cu->membankType;
                    cuRes->membankSize = cu->membankSize;
                    cuRes->membankBaseAddr = cu->membankBaseAddr;
                    cuRes->deviceId = devId;
                    cuRes->deviceNumBDF = dev->devNumBDF;
                    cuRes->cuId = cuId;
                    cuRes->poolId = reserveQueryInfo->poolId;
                    cuNum++;
                }
            }
        }
    }
    cuPoolRes->cuNum = cuNum;
    return (XRM_SUCCESS);
}

/*
 * flush the information data of xrm plugin
 */
void xrm::system::flushXrmPluginInfo(uint32_t xrmPluginId) {
    m_xrmPluginList[xrmPluginId].xrmPluginFileName = "";
    m_xrmPluginList[xrmPluginId].xrmPluginName = "";
    memset(&m_xrmPluginList[xrmPluginId].pluginData, 0, sizeof(xrmPluginData));
}

/*
 * Load one xrm plugin
 *
 * Return:
 *   XRM_SUCCESS: successed to load one xrm plugin
 *   XRM_ERROR: failed to load one xrm plugin
 *
 */
int32_t xrm::system::loadOneXrmPlugin(std::string& xrmPluginFileName, std::string& xrmPluginName, std::string& errmsg) {
    int32_t ret = XRM_ERROR;
    char* errorMsg;
    uint32_t pluginVersion = 0;

    /* check whether plugin loading reaches limitation */
    if (m_numXrmPlugin >= XRM_MAX_PLUGIN_NUM) {
        logMsg(XRM_LOG_ERROR, "%s: Plugin loading has reached the limitation (%d)\n", __func__, XRM_MAX_PLUGIN_NUM);
        errmsg += "Plugin loading has reached the limitation (" + std::to_string(XRM_MAX_PLUGIN_NUM) + ")";
        return (ret);
    }

    /* check whether plugin is loaded */
    for (uint32_t pluginId = 0; pluginId < m_numXrmPlugin; pluginId++) {
        if (m_xrmPluginList[pluginId].xrmPluginName.compare(xrmPluginName.c_str()) == 0) {
            pluginVersion = m_xrmPluginList[pluginId].pluginData.pluginVersion();
            logMsg(XRM_LOG_ERROR, "%s: plugin %s (version %d) is already loaded\n", __func__, xrmPluginName.c_str(),
                   pluginVersion);
            errmsg += "plugin " + xrmPluginName + " (version " + std::to_string(pluginVersion) + ") is already loaded";
            return (ret);
        }
    }

    /* load plugin library file */
    void* pluginHandle = dlopen(xrmPluginFileName.c_str(), (RTLD_GLOBAL | RTLD_NOW));
    if (!pluginHandle) {
        errorMsg = dlerror();
        logMsg(XRM_LOG_ERROR, "%s: Failed to open plugin %s\n Error msg: %s\n", __func__, xrmPluginFileName.c_str(),
               errorMsg);
        errmsg += errorMsg;
        return (ret);
    }
    xrmPluginData* plugin = (xrmPluginData*)dlsym(pluginHandle, xrmPluginName.c_str());
    if ((errorMsg = dlerror()) != NULL) {
        logMsg(XRM_LOG_ERROR, "%s: Failed to get plugin named %s\n Error msg: %s\n", __func__, xrmPluginName.c_str(),
               errorMsg);
        errmsg += errorMsg;
        return (ret);
    }

    /* check the api version */
    if (plugin->apiVersion == NULL) {
        logMsg(XRM_LOG_ERROR, "%s: Plugin library must provide api version function\n", __func__);
        errmsg += "There is No API version function in plugin";
        return (ret);
    }
    int32_t apiVersion = plugin->apiVersion();
    if (apiVersion != XRM_API_VERSION_1) {
        logMsg(XRM_LOG_ERROR, "%s: Plugin api version is not matching: %d\n", __func__, apiVersion);
        errmsg += "Plugin API version not match";
        return (ret);
    }

    /* check the plugin version */
    if (plugin->pluginVersion == NULL) {
        logMsg(XRM_LOG_ERROR, "%s: Plugin library must provide version function\n", __func__);
        errmsg += "There is No plugin version function in plugin";
        return (ret);
    }
    m_xrmPluginList[m_numXrmPlugin].xrmPluginFileName = xrmPluginFileName;
    m_xrmPluginList[m_numXrmPlugin].xrmPluginName = xrmPluginName;
    memcpy(&m_xrmPluginList[m_numXrmPlugin].pluginData, plugin, sizeof(xrmPluginData));
    m_numXrmPlugin++;
    ret = XRM_SUCCESS;
    return (ret);
}

/*
 * Load xrm plugins - loaded plugin recored in m_xrmPluginList[], starting from [0], no hole
 *
 * Return:
 *   XRM_SUCCESS: successed to load all xrm plugins
 *   XRM_ERROR: failed to load all xrm plugins
 *
 * the load process need to be protected by lock
 */
int32_t xrm::system::loadXrmPlugins(pt::ptree& loadPluginsTree, std::string& errmsg) {
    int32_t ret = XRM_ERROR;
    std::string xrmPluginFileName, xrmPluginName;
    std::string loadErrmsg;

    for (auto it : loadPluginsTree) {
        auto kv = it.second;
        xrmPluginFileName = kv.get<std::string>("xrmPluginFileName", "");
        if (xrmPluginFileName.c_str()[0] == '\0') {
            errmsg += "request parameters xrm plugin file name is not provided";
            ret = XRM_ERROR_INVALID;
            break;
        }

        xrmPluginName = kv.get<std::string>("xrmPluginName", "");
        if (xrmPluginName.c_str()[0] == '\0') {
            errmsg += "request parameters xrm plugin name is not provided";
            ret = XRM_ERROR_INVALID;
            break;
        }

        ret = loadOneXrmPlugin(xrmPluginFileName, xrmPluginName, loadErrmsg);
        if (ret != XRM_SUCCESS) {
            errmsg += loadErrmsg;
            break;
        }
    }
    return (ret);
}

/*
 * Unload one xrm plugin
 *
 * Return:
 *   XRM_SUCCESS: successed to unload one plugin
 *   XRM_ERROR: failed to unload one plugin
 *
 */
int32_t xrm::system::unloadOneXrmPlugin(std::string& xrmPluginName, std::string& errmsg) {
    int32_t ret = XRM_ERROR;
    bool foundPlugin = false;
    uint32_t pluginId = 0;

    /* find the plugin */
    for (pluginId = 0; pluginId < m_numXrmPlugin; pluginId++) {
        if (m_xrmPluginList[pluginId].xrmPluginName.compare(xrmPluginName.c_str()) == 0) {
            foundPlugin = true;
            break;
        }
    }
    if (!foundPlugin) {
        errmsg += "XRM Plugin not loaded";
        return (ret);
    }

    /* move the slot forward */
    for (; pluginId < (m_numXrmPlugin - 1); pluginId++) {
        memcpy(&m_xrmPluginList[pluginId].pluginData, &m_xrmPluginList[pluginId + 1].pluginData, sizeof(xrmPluginData));
        m_xrmPluginList[pluginId].xrmPluginFileName = m_xrmPluginList[pluginId + 1].xrmPluginFileName;
        m_xrmPluginList[pluginId].xrmPluginName = m_xrmPluginList[pluginId + 1].xrmPluginName;
    }

    /* clear the last slot */
    flushXrmPluginInfo(pluginId);
    m_numXrmPlugin--;
    ret = XRM_SUCCESS;
    return (ret);
}

/*
 * Unload xrm plugins - unloaded xrm plugin removed from m_xrmPluginList[], no hole
 *
 * Return:
 *   XRM_SUCCESS: successed to unload all xrm plugins
 *   XRM_ERROR: failed to unload all xrm plugins
 *
 * the unload process need to be protected by lock
 */
int32_t xrm::system::unloadXrmPlugins(pt::ptree& unloadPluginsTree, std::string& errmsg) {
    int32_t ret = XRM_ERROR;
    std::string xrmPluginName;
    std::string unloadErrmsg;

    for (auto it : unloadPluginsTree) {
        auto kv = it.second;
        xrmPluginName = kv.get<std::string>("xrmPluginName", "");
        if (xrmPluginName.c_str()[0] == '\0') {
            errmsg += "request parameters xrm plugin name is not provided";
            ret = XRM_ERROR_INVALID;
            break;
        }

        ret = unloadOneXrmPlugin(xrmPluginName, unloadErrmsg);
        if (ret != XRM_SUCCESS) {
            errmsg += unloadErrmsg;
            break;
        }
    }
    return (ret);
}

void* xrm::system::listXrmPlugin(int32_t pluginId) {
    return ((void*)&m_xrmPluginList[pluginId]);
}

int32_t xrm::system::getXrmPluginNumber() {
    return (m_numXrmPlugin);
}

/*
 * Execute the xrm plugin function
 *
 * Return:
 *   XRM_SUCCESS: successed to execute the xrm plugin function
 *   XRM_ERROR: failed to execute the xrm plugin function
 *
 */
int32_t xrm::system::execXrmPluginFunc(const char* xrmPluginName,
                                       uint32_t funcId,
                                       xrmPluginFuncParam* param,
                                       std::string& errmsg) {
    int32_t ret = XRM_ERROR_INVALID;
    uint32_t pluginId = 0;

    if (xrmPluginName[0] == '\0') {
        logMsg(XRM_LOG_ERROR, "%s: XRM Plugin name is not set\n", __func__);
        errmsg += "XRM Plugin name is not set";
        return (ret);
    }
    if (funcId > XRM_MAX_PLUGIN_FUNC_ID) {
        logMsg(XRM_LOG_ERROR, "%s: Plugin function id is out of range: %d\n", __func__, funcId);
        errmsg += "Plugin function id out of range";
        return (ret);
    }

    bool foundPlugin = false;
    for (pluginId = 0; pluginId < m_numXrmPlugin; pluginId++) {
        if (m_xrmPluginList[pluginId].xrmPluginName.compare(xrmPluginName) == 0) {
            foundPlugin = true;
            break;
        }
    }
    if (!foundPlugin) {
        logMsg(XRM_LOG_ERROR, "%s: Plugin %s is not loaded\n", __func__, xrmPluginName);
        errmsg += "Plugin is not loaded";
        return (ret);
    }
    if (m_xrmPluginList[pluginId].pluginData.pluginFunc[funcId] == NULL) {
        logMsg(XRM_LOG_ERROR, "%s: Plugin Function %d is not implemented\n", __func__, funcId);
        errmsg += "Function is not implemented";
        return (ret);
    }
    ret = m_xrmPluginList[pluginId].pluginData.pluginFunc[funcId](param);

    return (ret);
}

/*
 * Function wrapper
 * older version:
 *     NULL
 * 2019.2 and 2019.2_PU1
 *     int xclCuName2Index(xclDeviceHandle handle, const char *cu_name, uint32_t *cu_index)
 * 2019.2_PU2:
 *     int xclIPName2Index(xclDeviceHandle handle, const char *ipName, uint32_t *ipIndex);
 * 2020.1:
 *     int xclIPName2Index(xclDeviceHandle handle, const char *ipName);
 */
#ifdef __SUPPORT_XRT__
int32_t xrm::system::wrapIPName2Index(xclDeviceHandle handle, const char* ipName) {
    if (m_libVersionDepFuncs.libVersionDepFuncXclCuName2Index != NULL) {
        // xrt 2019.2 and 2019.2_PU1 API
        uint32_t cuId;
        int32_t ret;
        ret = m_libVersionDepFuncs.libVersionDepFuncXclCuName2Index(handle, ipName, &cuId);
        if (ret == 0)
            return ((int32_t)cuId);
        else
            return (ret);
    } else if (m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index25 != NULL) {
        // xrt 2019.2_PU2 API
        uint32_t cuId;
        int32_t ret;
        ret = m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index25(handle, ipName, &cuId);
        if (ret == 0)
            return ((int32_t)cuId);
        else
            return (ret);
    } else if (m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index26 != NULL) {
        // xrt 2020.1 API
        return (m_libVersionDepFuncs.libVersionDepFuncXclIPName2Index26(handle, ipName));
    } else {
        // older version, no API
        return (XRM_ERROR_INVALID);
    }
}
#endif

/*
 * Function wrapper
 * 2020.1 and older version
 * int xclLockDevice(xclDeviceHandle handle);
 */
#ifdef __SUPPORT_XRT__
int32_t xrm::system::wrapLockDevice(xclDeviceHandle handle) {
    if (m_libVersionDepFuncs.libVersionDepFuncXclLockDevice != NULL) {
        // xrt 2020.1 and older version, API implemented
        int32_t ret;
        ret = m_libVersionDepFuncs.libVersionDepFuncXclLockDevice(handle);
        return (ret);
    } else {
        // newer version, not support this API
        return (0);
    }
}
#endif

/*
 * Function wrapper
 * 2020.1 and older version
 * int xclUnlockDevice(xclDeviceHandle handle);
 */
#ifdef __SUPPORT_XRT__
int32_t xrm::system::wrapUnlockDevice(xclDeviceHandle handle) {
    if (m_libVersionDepFuncs.libVersionDepFuncXclUnlockDevice != NULL) {
        // xrt 2020.1 and older version, API implemented
        int32_t ret;
        ret = m_libVersionDepFuncs.libVersionDepFuncXclUnlockDevice(handle);
        return (ret);
    } else {
        // newer version, not support this API
        return (0);
    }
}
#endif

/*
 * To get the next cu based on logic group constraint information for coming allocation.
 *
 * startCuInforV2 will be the start position, the next cu meet requirement will
 * be the one after this start position:
 *        cuId: -1 is the start of whole device
 *        groupId: 0 is the start of whole device
 *        cuInGroupIndex: -1 is the start of whole device
 *
 * return the cuId if such cu is available, the cu information (cuId, groupId, cuInGroupIndex) will 
 *        be updated to startCuInforv2.
 * return cuId as -1 if no such cu.
 *
 * NOTE!! This function is NOT directly used for cu list allocation with
 *           XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_PSEUDO_LOGIC_GROUP_ID
 *           or
 *           XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE_AND_PSEUDO_LOGIC_GROUP_ID
 *        These two type logic group constraint will be handled in cu list allocation related functions.
 */
void xrm::system::getNextCuFromDev(int32_t devId, cuPropertyV2* cuPropV2, singleCuResInforV2* startCuInforV2) {
    deviceData* dev;
    int32_t cuId = -1;
    int32_t groupId;
    int32_t cuInGroupIndex;

    dev = &m_devList[devId];
    uint32_t constraintInfo = (cuPropV2->logicGroupInfo.constraintInfo & XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_MASK);

    switch (constraintInfo) {
    case XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_NULL:
        /*
         * no logic group constraint:
         *  - Only cuId of startCuInforV2 will be used to
         *    find next available cu.
         *  - groupId and cuInGroupIndex will NOT be used to find next available cu.
         */
        // next in order cu will be the one
        cuId = startCuInforV2->cuId + 1;
        if (cuId >= XRM_MAX_XILINX_KERNELS || cuId >= dev->xclbinInfo.numCu) {
            // reach the end of device, no cu for the requirement
            cuId = -1;
        }
        startCuInforV2->cuId = cuId;
        // record the logic group infor as invalid
        setLogicGroupInforInvalid(&(startCuInforV2->logicGroupInfo));
        break;
    case XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID:
        /*
         * logic group constraint is real logic group id:
         *  - The groupId will be that in cuPropV2.
         *  - Only cuInGroupIndex of startCuInforV2 will be used to
         *    find next available cu.
         *  - cuId will NOT be used to find next available cu.
         */
        groupId = cuPropV2->logicGroupInfo.groupId;
        // next cu from the group
        cuInGroupIndex = startCuInforV2->cuInGroupIndex + 1;

        if (cuInGroupIndex < dev->xclbinInfo.logicGroups[groupId].numCu) {
            // get next cu in logic group
            cuId = dev->xclbinInfo.logicGroups[groupId].cuId[cuInGroupIndex];
        } else {
            // reach the end of logic group, no cu for the request
            cuId = -1;
        }
        /* update result */
        startCuInforV2->cuId = cuId;
        startCuInforV2->cuInGroupIndex = cuInGroupIndex;
        // including the logic group infor
        startCuInforV2->logicGroupInfo.indicator = XRM_LOGIC_GROUP_INFO_ID_TYPE_VALID;
        startCuInforV2->logicGroupInfo.groupId = groupId;
        strncpy(startCuInforV2->logicGroupInfo.groupType,
            dev->xclbinInfo.logicGroups[groupId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1);
        startCuInforV2->logicGroupInfo.groupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1] = '\0';
        break;
    case XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE:
        /*
         * logic group constraint is real logic group id:
         *  - groupType in cuPropV2 will be used to check type match.
         *  - Only groupId and cuInGroupIndex of startCuInforV2 will be used to
         *    find next available cu.
         *  - cuId will NOT be used to find next available cu.
         */
        /* start position */
        groupId = startCuInforV2->logicGroupInfo.groupId;
        cuInGroupIndex = startCuInforV2->cuInGroupIndex;
        if (strncmp(cuPropV2->logicGroupInfo.groupType,
            dev->xclbinInfo.logicGroups[groupId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) == 0 ) {
            // Current (start) group type match request type, get the next cu in this group
            cuInGroupIndex++;
            if (cuInGroupIndex >= dev->xclbinInfo.logicGroups[groupId].numCu) {
                // Reach end of current logic group. Go to next group
                cuInGroupIndex = -1;
            }
        } else {
            // Current (start) group type NOT match request type. Go to next group.
            cuInGroupIndex = -1;
        }
        if (cuInGroupIndex == -1) {
            /*
             * To find next group match target type and numCu > 0
             */
            groupId++;
            for (; groupId < dev->xclbinInfo.numLogicGroup; groupId++) {
                if ((strncmp(cuPropV2->logicGroupInfo.groupType,
                    dev->xclbinInfo.logicGroups[groupId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN) == 0 ) &&
                    (dev->xclbinInfo.logicGroups[groupId].numCu > 0)) {
                    /*
                     * Group type equal to request type, there is cu in this group
                     * To use the first cu in this group.
                     */
                    cuInGroupIndex = 0;
                    break;
                }
            }
        }
        if (cuInGroupIndex != -1) {
            // found one cu
            cuId = dev->xclbinInfo.logicGroups[groupId].cuId[cuInGroupIndex];
        } else {
            // no cu for the request
            cuId = -1;
        }
        /* update result */
        startCuInforV2->cuId = cuId;
        startCuInforV2->cuInGroupIndex = cuInGroupIndex;
        // record the logic group infor
        startCuInforV2->logicGroupInfo.indicator = XRM_LOGIC_GROUP_INFO_ID_TYPE_VALID;
        startCuInforV2->logicGroupInfo.groupId = groupId;
        strncpy(startCuInforV2->logicGroupInfo.groupType,
            dev->xclbinInfo.logicGroups[groupId].groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1);
        startCuInforV2->logicGroupInfo.groupType[XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1] = '\0';
        break;
    default:
        startCuInforV2->cuId = -1;
        memset(&(startCuInforV2->logicGroupInfo), 0 , sizeof(cuResourceLogicGroupInfo));
        break;
    }
}

/*
 * copy logic group information from src to des.
 */
void xrm::system::copyLogicGroupInfor(cuResourceLogicGroupInfo* desInfo, cuResourceLogicGroupInfo* srcInfo) {
    if (desInfo != NULL && srcInfo != NULL) {
        desInfo->indicator = srcInfo->indicator;
        desInfo->groupId = srcInfo->groupId;
        strncpy(desInfo->groupType, srcInfo->groupType, XRM_MAX_LOGIC_GROUP_TYPE_LEN);
    }
}

/*
 * set logic group information to invalid result.
 */
void xrm::system::setLogicGroupInforInvalid(cuResourceLogicGroupInfo* logicGroupInfo) {
    if (logicGroupInfo != NULL) {
        logicGroupInfo->indicator = XRM_LOGIC_GROUP_INFO_INVALID;
        logicGroupInfo->groupId = 0;
        strncpy(logicGroupInfo->groupType, "", XRM_MAX_LOGIC_GROUP_TYPE_LEN - 1);
    }
}

/*
 * Alloc one cu (compute unit) based on the request property version 2.
 * e.g for dev most used first policy, it tries to find cu by the order
 * of most used dev load id to least used dev load id
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocCuByDevLoad(cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    deviceData* dev;
    cuData* cu;
    cuData* preCu = NULL;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuPropV2->clientId;
    int32_t devId, cuId;
    int32_t preDevId = -1, preCuId = -1;
    bool cuAcquired = false;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;
    // no need to check cuPropV2 or cuRes as it should be checked already
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    cuResourceLogicGroupInfo preLogicGroupInfo;
    memset(&preLogicGroupInfo, 0, sizeof(cuResourceLogicGroupInfo));

    singleCuResInforV2 startCuInfor;

    std::vector<xrm::deviceLoadInfo> devLoadArray(m_numDevice);
    for (int i = 0; i < m_numDevice; i++) {
        devLoadArray[i] = m_devList[i].devLoadInfo;
    }
    std::sort(devLoadArray.begin(), devLoadArray.end());

    for (int i = m_numDevice - 1; i >= 0 && !cuAcquired; i--) {
        /*
         * find the device based on policy
         */
        devId = devLoadArray[i].deviceId;
        if (cuPropV2->policyInfo == XRM_POLICY_INFO_CONSTRAINT_TYPE_DEV_LEAST_USED_FIRST)
            devId = devLoadArray[m_numDevice - 1 - i].deviceId;
        ret = allocClientFromDev(devId, cuProp);
        if (ret < 0) {
            continue;
        }
        dev = &m_devList[devId];

        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        memset(&startCuInfor, 0, sizeof(singleCuResInforV2));
        startCuInfor.cuId = -1;
        startCuInfor.logicGroupInfo.groupId = 0;
        startCuInfor.cuInGroupIndex = -1;
        /* get the next cu based on logic group constraint information */
        getNextCuFromDev(devId, cuPropV2, &startCuInfor);
        cuId = startCuInfor.cuId;
        while ((cuId != -1) && !cuAcquired) {
            cu = &dev->xclbinInfo.cuList[cuId];

            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            /* cu name is presented, compare it; otherwise no need to compare */
            cuNameEqual = true;
            if (cuProp->cuName[0] != '\0') {
                if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) {
                /* go to next cu */
                getNextCuFromDev(devId, cuPropV2, &startCuInfor);
                cuId = startCuInfor.cuId;
                continue;
            }
            /*
             * alloc channel and register client id
             *
             * if policy is not set, then alloc chan from this cu (either fail or success).
             * if policy is set, then just check if this cu is better, the better cu
             * information will be recorded with preCu, preDevId, preCuId.
             */
            cuData* tmpCu = preCu;
            /*
             * logic group information will be set in allocChanClientFromCu call based
             * on input logic group information or reservation pool logic group information.
             */
            ret = allocChanClientFromCu(cu, cuProp, cuRes, &(startCuInfor.logicGroupInfo),
                cuPropV2->policyInfo, &preCu);
            if (ret != XRM_SUCCESS) {
                if (tmpCu != preCu) {
                    /* find better cu, record the information */
                    preDevId = devId;
                    preCuId = cuId;
                    copyLogicGroupInfor(&preLogicGroupInfo, &(startCuInfor.logicGroupInfo));
                }
                /* go to next cu */
                getNextCuFromDev(devId, cuPropV2, &startCuInfor);
                cuId = startCuInfor.cuId;
                continue;
            }

            /*
             * XRM_SUCCESS, channel allocation is done. To set allocated resource information.
             */
            preCu = NULL;
            cuRes->deviceId = devId;
            cuRes->deviceNumBDF = dev->devNumBDF;
            cuRes->cuId = cuId;
            strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
            strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
            cuAcquired = true;
        } // cu loop

        if (preCu) {
            /*
             * policy is set, so real chan allocation will be done here.
             */
            dev = &m_devList[preDevId];
            cu = &dev->xclbinInfo.cuList[preCuId];
            /*
             * logic group information will be set in allocChanClientFromCu call based
             * on input logic group information or reservation pool logic group information.
             */
            ret = allocChanClientFromCu(cu, cuProp, cuRes, &(startCuInfor.logicGroupInfo));
            cuRes->deviceId = preDevId;
            cuRes->deviceNumBDF = dev->devNumBDF;
            cuRes->cuId = preCuId;
            strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
            strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
            cuAcquired = true;
        }
        if (cuAcquired) {
            if (updateId) updateAllocServiceId();
            return (XRM_SUCCESS);
        }

        if (!cuAcquired) {
            releaseClientOnDev(devId, clientId);
        }
    }
    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Alloc one cu (compute unit) based on the request property version 2.
 * e.g for cu load most used first policy, it tries to find cu by the order
 * of most used cu load to least used cu load
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocCuByCuLoad(cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    deviceData* dev;
    cuData* cu;
    cuData* preCu = NULL;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuPropV2->clientId;
    int32_t devId, cuId;
    int32_t preDevId = -1, preCuId = -1;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;

    cuResourceLogicGroupInfo preLogicGroupInfo;
    memset(&preLogicGroupInfo, 0, sizeof(cuResourceLogicGroupInfo));

    singleCuResInforV2 startCuInfor;

    // no need to check cuPropV2 or cuRes as it should be checked already
    cuPropertyCopyFromV2(cuProp, cuPropV2);
cu_alloc_loop:
    for (devId = -1; !cuAcquired && (devId < m_numDevice);) {
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            break;
        }

        ret = allocClientFromDev(devId, cuProp);
        if (ret < 0) {
            continue;
        }
        dev = &m_devList[devId];
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        memset(&startCuInfor, 0, sizeof(singleCuResInforV2));
        startCuInfor.cuId = -1;
        startCuInfor.logicGroupInfo.groupId = 0;
        startCuInfor.cuInGroupIndex = -1;
        /* get the next cu based on logic group constraint information */
        getNextCuFromDev(devId, cuPropV2, &startCuInfor);
        cuId = startCuInfor.cuId;
        while ((cuId != -1) && !cuAcquired) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to re-use existing kernels; else, use a new kernel */
            if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) {
                /* go to next cu */
                getNextCuFromDev(devId, cuPropV2, &startCuInfor);
                cuId = startCuInfor.cuId;
                continue;
            }
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            /* cu name is presented, compare it; otherwise no need to compare */
            cuNameEqual = true;
            if (cuProp->cuName[0] != '\0') {
                if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) {
                /* go to next cu */
                getNextCuFromDev(devId, cuPropV2, &startCuInfor);
                cuId = startCuInfor.cuId;
                continue;
            }
            /*
             * alloc channel and register client id
             *
             * if policy is not set, then alloc chan from this cu (either fail or success).
             * if policy is set, then just check if this cu is better, the better cu
             * information will be recorded with preCu, preDevId, preCuId.
             */
            cuData* tmpCu = preCu;
            /*
             * logic group information will be set in allocChanClientFromCu call based
             * on input logic group information or reservation pool logic group information.
             */
            ret = allocChanClientFromCu(cu, cuProp, cuRes, &(startCuInfor.logicGroupInfo),
                cuPropV2->policyInfo, &preCu);
            if (ret != XRM_SUCCESS) {
                if (tmpCu != preCu) {
                    /* find better cu, record the information */
                    preDevId = devId;
                    preCuId = cuId;
                    copyLogicGroupInfor(&preLogicGroupInfo, &(startCuInfor.logicGroupInfo));
                }
                /* go to next cu */
                getNextCuFromDev(devId, cuPropV2, &startCuInfor);
                cuId = startCuInfor.cuId;
                continue;
            }

            /*
             * XRM_SUCCESS, channel allocation is done. To set allocated resource information.
             */
            preCu = NULL;
            cuRes->deviceId = devId;
            cuRes->deviceNumBDF = dev->devNumBDF;
            cuRes->cuId = cuId;
            strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
            strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
            cuAcquired = true;
        }

        if (!cuAcquired) {
            releaseClientOnDev(devId, clientId);
        }
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_alloc_loop;
    }
    if (preCu) {
        /*
         * policy is set, so real chan allocation will be done here.
         */
        dev = &m_devList[preDevId];
        cu = &dev->xclbinInfo.cuList[preCuId];
        /*
         * logic group information will be set in allocChanClientFromCu call based
         * on input logic group information or reservation pool logic group information.
         */
        ret = allocChanClientFromCu(cu, cuProp, cuRes, &(preLogicGroupInfo));
        cuRes->deviceId = preDevId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = preCuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    }
    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

/*
 * Alloc one cu (compute unit) based on the request property, no allocate policy.
 *
 * XRM_SUCCESS: allocated resouce is recorded by cuRes
 * Otherwise: failed to allocate the cu resource
 *
 * Lock: should enter lock during the cu allocation
 */
int32_t xrm::system::resAllocCuWithoutPolicy(cuPropertyV2* cuPropV2, cuResource* cuRes, bool updateId) {
    deviceData* dev;
    cuData* cu;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuPropV2->clientId;
    int32_t devId, cuId;
    bool cuAcquired = false;
    /* First pass will look for kernels already in-use by proc */
    bool cuAffinityPass = true;

    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;

    singleCuResInforV2 startCuInfor;

    // no need to check cuPropV2 or cuRes as it should be checked already
    cuPropertyCopyFromV2(cuProp, cuPropV2);

cu_alloc_loop:
    for (devId = -1; !cuAcquired && (devId < m_numDevice);) {
        devId = allocDevForClient(&devId, cuProp);
        if (devId < 0) {
            break;
        }

        ret = allocClientFromDev(devId, cuProp);
        if (ret < 0) {
            continue;
        }
        dev = &m_devList[devId];
        /*
         * Now check whether matching cu is on allocated device
         * if not, free device, increment dev count, re-loop
         */
        memset(&startCuInfor, 0, sizeof(singleCuResInforV2));
        startCuInfor.cuId = -1;
        startCuInfor.logicGroupInfo.groupId = 0;
        startCuInfor.cuInGroupIndex = -1;
        /* get the next cu based on logic group constraint information */
        getNextCuFromDev(devId, cuPropV2, &startCuInfor);
        cuId = startCuInfor.cuId;
        while ((cuId != -1) && !cuAcquired) {
            cu = &dev->xclbinInfo.cuList[cuId];

            /* first attempt to re-use existing kernels; else, use a new kernel */
            if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) {
                /* go to next cu */
                getNextCuFromDev(devId, cuPropV2, &startCuInfor);
                cuId = startCuInfor.cuId;
                continue;
            }
            /*
             * compare, 0: equal
             *
             * kernel name is presented, compare it; otherwise no need to compare
             */
            kernelNameEqual = true;
            if (cuProp->kernelName[0] != '\0') {
                if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
            }
            /* kernel alias is presented, compare it; otherwise no need to compare */
            kernelAliasEqual = true;
            if (cuProp->kernelAlias[0] != '\0') {
                if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
            }
            /* cu name is presented, compare it; otherwise no need to compare */
            cuNameEqual = true;
            if (cuProp->cuName[0] != '\0') {
                if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
            }
            if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) {
                /* go to next cu */
                getNextCuFromDev(devId, cuPropV2, &startCuInfor);
                cuId = startCuInfor.cuId;
                continue;
            }
            /* alloc channel and register client id */
            /*
             * logic group information will be set in allocChanClientFromCu call based
             * on input logic group information or reservation pool logic group information.
             */
            ret = allocChanClientFromCu(cu, cuProp, cuRes, &(startCuInfor.logicGroupInfo));
            if (ret != XRM_SUCCESS) {
                /* go to next cu */
                getNextCuFromDev(devId, cuPropV2, &startCuInfor);
                cuId = startCuInfor.cuId;
                continue;
            }

            /*
             * XRM_SUCCESS, channel allocation is done. To set allocated resource information.
             */
            cuRes->deviceId = devId;
            cuRes->deviceNumBDF = dev->devNumBDF;
            cuRes->cuId = cuId;
            strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
            strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
            cuAcquired = true;
        }

        if (!cuAcquired) {
            releaseClientOnDev(devId, clientId);
        }
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_alloc_loop;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    }

    return (XRM_ERROR_NO_KERNEL);
}

int32_t xrm::system::resAllocCuFromDevByCuLoad(int32_t deviceId,
                                               cuPropertyV2* cuPropV2,
                                               cuResource* cuRes,
                                               bool updateId) {
    deviceData* dev;
    cuData* cu;
    cuData* preCu = NULL;
    int32_t ret = 0;
    bool kernelNameEqual, kernelAliasEqual, cuNameEqual;
    uint64_t clientId = cuPropV2->clientId;
    int32_t cuId;
    int32_t preCuId = -1;
    bool cuAcquired = false;
    // First pass will look for kernels already in-use by proc
    bool cuAffinityPass = true;
    cuProperty tmpProp;
    cuProperty* cuProp = &tmpProp;

    cuResourceLogicGroupInfo preLogicGroupInfo;
    memset(&preLogicGroupInfo, 0, sizeof(cuResourceLogicGroupInfo));

    singleCuResInforV2 startCuInfor;

    if ((deviceId < 0) || (deviceId > (m_numDevice - 1))) {
        logMsg(XRM_LOG_ERROR, "deviceId (%d) is out of range\n", deviceId);
        return (XRM_ERROR_INVALID);
    }
    // no need to check cuPropV2 or cuRes as it should be checked already
    cuPropertyCopyFromV2(cuProp, cuPropV2);

    dev = &m_devList[deviceId];
    if (!dev->isLoaded) {
        return (XRM_ERROR_NO_DEV);
    }
    if ((dev->isExcl) && (dev->clientProcs[0].clientId != clientId)) {
        // the device is used by other client exclusively
        return (XRM_ERROR_NO_DEV);
    }
    ret = allocClientFromDev(deviceId, cuProp);
    if (ret < 0) {
        return (XRM_ERROR_NO_DEV);
    }

cu_alloc_from_dev_loop:
    /*
     * Now check whether matching cu is on target device
     * and try to allocate requested cu.
     */
    memset(&startCuInfor, 0, sizeof(singleCuResInforV2));
    startCuInfor.cuId = -1;
    startCuInfor.logicGroupInfo.groupId = 0;
    startCuInfor.cuInGroupIndex = -1;
    /* get the next cu based on logic group constraint information */
    getNextCuFromDev(deviceId, cuPropV2, &startCuInfor);
    cuId = startCuInfor.cuId;
    while ((cuId != -1) && !cuAcquired) {
        cu = &dev->xclbinInfo.cuList[cuId];

        /* first attempt to re-use existing kernels; else, use a new kernel */
        if ((cuAffinityPass && cu->numClient == 0) || (!cuAffinityPass && cu->numClient > 0)) {
            /* go to next cu */
            getNextCuFromDev(deviceId, cuPropV2, &startCuInfor);
            cuId = startCuInfor.cuId;
            continue;
        }
        /*
         * compare, 0: equal
         *
         * kernel name is presented, compare it; otherwise no need to compare
         */
        kernelNameEqual = true;
        if (cuProp->kernelName[0] != '\0') {
            if (cu->kernelName.compare(cuProp->kernelName)) kernelNameEqual = false;
        }
        /* kernel alias is presented, compare it; otherwise no need to compare */
        kernelAliasEqual = true;
        if (cuProp->kernelAlias[0] != '\0') {
            if (cu->kernelAlias.compare(cuProp->kernelAlias)) kernelAliasEqual = false;
        }
        /* cu name is presented, compare it; otherwise no need to compare */
        cuNameEqual = true;
        if (cuProp->cuName[0] != '\0') {
            if (cu->cuName.compare(cuProp->cuName)) cuNameEqual = false;
        }
        if (!(kernelNameEqual && kernelAliasEqual && cuNameEqual)) {
            /* go to next cu */
            getNextCuFromDev(deviceId, cuPropV2, &startCuInfor);
            cuId = startCuInfor.cuId;
            continue;
        }
        /*
         * alloc channel and register client id
         *
         * if policy is not set, then alloc chan from this cu (either fail or success).
         * if policy is set, then just check if this cu is better, the better cu
         * information will be recorded with preCu, preCuId.
         */
        cuData* tmpCu = preCu;
        /*
         * logic group information will be set in allocChanClientFromCu call based
         * on input logic group information or reservation pool logic group information.
         */
        ret = allocChanClientFromCu(cu, cuProp, cuRes, &(startCuInfor.logicGroupInfo),
            cuPropV2->policyInfo, &preCu);
        if (ret != XRM_SUCCESS) {
            if (tmpCu != preCu) {
                /* find better cu, record the information */
                preCuId = cuId;
                copyLogicGroupInfor(&preLogicGroupInfo, &(startCuInfor.logicGroupInfo));
            }
            /* go to next cu */
            getNextCuFromDev(deviceId, cuPropV2, &startCuInfor);
            cuId = startCuInfor.cuId;
            continue;
        }

        /*
         * XRM_SUCCESS, channel allocation is done. To set allocated resource information.
         */
        preCu = NULL;
        cuRes->deviceId = deviceId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = cuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    }

    if (!cuAcquired && cuAffinityPass) {
        cuAffinityPass = false; /* open up search to all kernels */
        goto cu_alloc_from_dev_loop;
    }
    if (preCu) {
        /*
         * policy is set, so real chan allocation will be done here.
         */
        cu = &dev->xclbinInfo.cuList[preCuId];
        /*
         * logic group information will be set in allocChanClientFromCu call based
         * on input logic group information or reservation pool logic group information.
         */
        ret = allocChanClientFromCu(cu, cuProp, cuRes, &(preLogicGroupInfo));
        cuRes->deviceId = deviceId;
        cuRes->deviceNumBDF = dev->devNumBDF;
        cuRes->cuId = preCuId;
        strncpy(cuRes->xclbinFileName, dev->xclbinName.c_str(), XRM_MAX_NAME_LEN - 1);
        strncpy(cuRes->uuidStr, dev->xclbinInfo.uuidStr.c_str(), XRM_MAX_NAME_LEN - 1);
        cuAcquired = true;
    }

    if (cuAcquired) {
        if (updateId) updateAllocServiceId();
        return (XRM_SUCCESS);
    } else {
        releaseClientOnDev(deviceId, clientId);
        return (XRM_ERROR_NO_KERNEL);
    }
}
